#  @file
#  @author Christian Diddens <c.diddens@utwente.nl>
#  @author Duarte Rocha <d.rocha@utwente.nl>
#  
#  @section LICENSE
# 
#  pyoomph - a multi-physics finite element framework based on oomph-lib and GiNaC 
#  Copyright (C) 2021-2025  Christian Diddens & Duarte Rocha
# 
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
# 
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>. 
#
#  The authors may be contacted at c.diddens@utwente.nl and d.rocha@utwente.nl
#
# ========================================================================


##########################################################################################
# These parameter stem from the published data of the modified UNIFAC model (Dortmund)
#		https://www.ddbst.com/PublishedParametersUNIFACDO.html
#
# Please cite the relevant publications when publishing results based on the modified UNIFAC model (Dortmund)
#		see: https://www.ddbst.com/PublishedParametersUNIFACDO.html
#
# Also note that these are the published parameters. The UNIFAC Consortium offers updated
# parameters, which will lead to more accurate results. You can obtain these parameters
# at:
#
#		https://www.ddbst.com/unifac-consortium.html
#
# and register the updated parameter by following the template below.
#
# These updated parameters, however, are subject to the *terms of use* 
# of the UNIFAC Consortium.
##########################################################################################


from ..activity import UNIFACLikeActivityModel,ActivityModel

@ActivityModel.register_activity_model()
class DortmundUNIFAC(UNIFACLikeActivityModel):
    """
    Modified (Dortmund) UNIFAC model for activity coefficient calculations.
    
    The published parameters were obtained with kind permisson from the DDBST website (https://www.ddbst.com/PublishedParametersUNIFACDO.html).
    
    Please cite the relevant publications when publishing results based on this model (https://www.ddbst.com/PublishedParametersUNIFACDO.html).
    
    Note that these are the published parameters. The UNIFAC Consortium offers updated and revised parameters, which will lead to more accurate results.     
    You can obtain these parameters at: https://www.ddbst.com/unifac-consortium.html 
    
    These updated parameters, however, are subject to the *terms of use* of the UNIFAC Consortium.
    
    Furthermore, you must implement these revised parameters in the code yourself.
    """

    name = "Dortmund"
    def __init__(self):
        super(DortmundUNIFAC, self).__init__()
        self.modified_volume_fraction_exponent=3/4
        self.define_groups()
        self.define_interaction_table()

    def define_groups(self):
        with self.define_main_group("CH2", index=1):
            self.define_sub_group("CH3", R=0.6325, Q=1.0608, index=1)
            self.define_sub_group("CH2", R=0.6325, Q=0.7081, index=2)
            self.define_sub_group("CH", R=0.6325, Q=0.3554, index=3)
            self.define_sub_group("C", R=0.6325, Q=0.0, index=4)
        with self.define_main_group("C=C", index=2):
            self.define_sub_group("CH2=CH", R=1.2832, Q=1.6016, index=5)
            self.define_sub_group("CH=CH", R=1.2832, Q=1.2489, index=6)
            self.define_sub_group("CH2=C", R=1.2832, Q=1.2489, index=7)
            self.define_sub_group("CH=C", R=1.2832, Q=0.8962, index=8)
            self.define_sub_group("C=C", R=1.2832, Q=0.4582, index=70)
        with self.define_main_group("ACH", index=3):
            self.define_sub_group("ACH", R=0.3763, Q=0.4321, index=9)
            self.define_sub_group("AC", R=0.3763, Q=0.2113, index=10)
        with self.define_main_group("ACCH2", index=4):
            self.define_sub_group("ACCH3", R=0.91, Q=0.949, index=11)
            self.define_sub_group("ACCH2", R=0.91, Q=0.7962, index=12)
            self.define_sub_group("ACCH", R=0.91, Q=0.3769, index=13)
        with self.define_main_group("OH", index=5):
            self.define_sub_group("OH (P)", R=1.2302, Q=0.8927, index=14)
            self.define_sub_group("OH (S)", R=1.063, Q=0.8663, index=81)
            self.define_sub_group("OH (T)", R=0.6895, Q=0.8345, index=82)
        with self.define_main_group("CH3OH", index=6):
            self.define_sub_group("CH3OH", R=0.8585, Q=0.9938, index=15)
        with self.define_main_group("H2O", index=7):
            self.define_sub_group("H2O", R=1.7334, Q=2.4561, index=16)
        with self.define_main_group("ACOH", index=8):
            self.define_sub_group("ACOH", R=1.08, Q=0.975, index=17)
        with self.define_main_group("CH2CO", index=9):
            self.define_sub_group("CH3CO", R=1.7048, Q=1.67, index=18)
            self.define_sub_group("CH2CO", R=1.7048, Q=1.5542, index=19)
        with self.define_main_group("HCO", index=10):
            self.define_sub_group("HCO", R=0.7173, Q=0.771, index=20)
        with self.define_main_group("CCOO", index=11):
            self.define_sub_group("CH3COO", R=1.27, Q=1.6286, index=21)
            self.define_sub_group("CH2COO", R=1.27, Q=1.4228, index=22)
        with self.define_main_group("HCOO", index=12):
            self.define_sub_group("HCOO", R=1.9, Q=1.8, index=23)
        with self.define_main_group("CH2O", index=13):
            self.define_sub_group("CH3O", R=1.1434, Q=1.6022, index=24)
            self.define_sub_group("CH2O", R=1.1434, Q=1.2495, index=25)
            self.define_sub_group("CHO", R=1.1434, Q=0.8968, index=26)
        with self.define_main_group("CY-CH2O", index=43):
            self.define_sub_group("THF", R=1.7023, Q=1.8784, index=27)
            self.define_sub_group("CY-CH2O", R=1.4046, Q=1.4, index=83)
            self.define_sub_group("TRIOXAN", R=1.0413, Q=1.0116, index=84)
        with self.define_main_group("CH2NH2", index=14):
            self.define_sub_group("CH3NH2", R=1.6607, Q=1.6904, index=28)
            self.define_sub_group("CH2NH2", R=1.6607, Q=1.3377, index=29)
            self.define_sub_group("CHNH2", R=1.6607, Q=0.985, index=30)
            self.define_sub_group("CNH2", R=1.6607, Q=0.985, index=85)
        with self.define_main_group("CH2NH", index=15):
            self.define_sub_group("CH3NH", R=1.368, Q=1.4332, index=31)
            self.define_sub_group("CH2NH", R=1.368, Q=1.0805, index=32)
            self.define_sub_group("CHNH", R=1.368, Q=0.7278, index=33)
        with self.define_main_group("(C)3N", index=16):
            self.define_sub_group("CH3N", R=1.0746, Q=1.176, index=34)
            self.define_sub_group("CH2N", R=1.0746, Q=0.824, index=35)
        with self.define_main_group("ACNH2", index=17):
            self.define_sub_group("ACNH2", R=1.1849, Q=0.8067, index=36)
        with self.define_main_group("PYRIDINE", index=18):
            self.define_sub_group("AC2H2N", R=1.4578, Q=0.9022, index=37)
            self.define_sub_group("AC2HN", R=1.2393, Q=0.633, index=38)
            self.define_sub_group("AC2N", R=1.0731, Q=0.353, index=39)
        with self.define_main_group("CH2CN", index=19):
            self.define_sub_group("CH3CN", R=1.5575, Q=1.5193, index=40)
            self.define_sub_group("CH2CN", R=1.5575, Q=1.1666, index=41)
        with self.define_main_group("COOH", index=20):
            self.define_sub_group("COOH", R=0.8, Q=0.9215, index=42)
        with self.define_main_group("HCOOH", index=44):
            self.define_sub_group("HCOOH", R=0.8, Q=1.2742, index=43)
        with self.define_main_group("CCL", index=21):
            self.define_sub_group("CH2CL", R=0.9919, Q=1.3654, index=44)
            self.define_sub_group("CHCL", R=0.9919, Q=1.0127, index=45)
            self.define_sub_group("CCL", R=0.9919, Q=0.66, index=46)
        with self.define_main_group("CCL2", index=22):
            self.define_sub_group("CH2CL2", R=1.8, Q=2.5, index=47)
            self.define_sub_group("CHCL2", R=1.8, Q=2.1473, index=48)
            self.define_sub_group("CCL2", R=1.8, Q=1.7946, index=49)
        with self.define_main_group("CHCL3", index=45):
            self.define_sub_group("CHCL3", R=2.45, Q=2.8912, index=50)
        with self.define_main_group("CCL3", index=23):
            self.define_sub_group("CCL3", R=2.65, Q=2.3778, index=51)
        with self.define_main_group("CCL4", index=24):
            self.define_sub_group("CCL4", R=2.618, Q=3.1836, index=52)
        with self.define_main_group("ACCL", index=25):
            self.define_sub_group("ACCL", R=0.5365, Q=0.3177, index=53)
        with self.define_main_group("CNO2", index=26):
            self.define_sub_group("CH3NO2", R=2.644, Q=2.5, index=54)
            self.define_sub_group("CH2NO2", R=2.5, Q=2.304, index=55)
            self.define_sub_group("CHNO2", R=2.887, Q=2.241, index=56)
        with self.define_main_group("ACNO2", index=27):
            self.define_sub_group("ACNO2", R=0.4656, Q=0.3589, index=57)
        with self.define_main_group("CS2", index=28):
            self.define_sub_group("CS2", R=1.24, Q=1.068, index=58)
        with self.define_main_group("CH3SH", index=29):
            self.define_sub_group("CH3SH", R=1.289, Q=1.762, index=59)
            self.define_sub_group("CH2SH", R=1.535, Q=1.316, index=60)
        with self.define_main_group("FURFURAL", index=30):
            self.define_sub_group("FURFURAL", R=1.299, Q=1.289, index=61)
        with self.define_main_group("DOH", index=31):
            self.define_sub_group("DOH", R=2.088, Q=2.4, index=62)
        with self.define_main_group("I", index=32):
            self.define_sub_group("I", R=1.076, Q=0.9169, index=63)
        with self.define_main_group("BR", index=33):
            self.define_sub_group("BR", R=1.209, Q=1.4, index=64)
        with self.define_main_group("C=-C", index=34):
            self.define_sub_group("CH=-C", R=0.9214, Q=1.3, index=65)
            self.define_sub_group("C=-C", R=1.303, Q=1.132, index=66)
        with self.define_main_group("DMSO", index=35):
            self.define_sub_group("DMSO", R=3.6, Q=2.692, index=67)
        with self.define_main_group("ACRY", index=36):
            self.define_sub_group("ACRY", R=1.0, Q=0.92, index=68)
        with self.define_main_group("CLCC", index=37):
            self.define_sub_group("CL-(C=C)", R=0.5229, Q=0.7391, index=69)
        with self.define_main_group("ACF", index=38):
            self.define_sub_group("ACF", R=0.8814, Q=0.7269, index=71)
        with self.define_main_group("DMF", index=39):
            self.define_sub_group("DMF", R=2.0, Q=2.093, index=72)
            self.define_sub_group("HCON(..", R=2.381, Q=1.522, index=73)
        with self.define_main_group("CF2", index=40):
            self.define_sub_group("CF3", R=1.284, Q=1.266, index=74)
            self.define_sub_group("CF2", R=1.284, Q=1.098, index=75)
            self.define_sub_group("CF", R=0.8215, Q=0.5135, index=76)
        with self.define_main_group("COO", index=41):
            self.define_sub_group("COO", R=1.6, Q=0.9, index=77)
        with self.define_main_group("CY-CH2", index=42):
            self.define_sub_group("CY-CH2", R=0.7136, Q=0.8635, index=78)
            self.define_sub_group("CY-CH", R=0.3479, Q=0.1071, index=79)
            self.define_sub_group("CY-C", R=0.347, Q=0.0, index=80)
        with self.define_main_group("CY-CONC", index=46):
            self.define_sub_group("NMP", R=3.981, Q=3.2, index=86)
            self.define_sub_group("NEP", R=3.7543, Q=2.892, index=87)
            self.define_sub_group("NIPP", R=3.5268, Q=2.58, index=88)
            self.define_sub_group("NTBP", R=3.2994, Q=2.352, index=89)
        with self.define_main_group("CONR", index=47):
            self.define_sub_group("CONH2", R=1.4515, Q=1.248, index=91)
            self.define_sub_group("CONHCH3", R=1.5, Q=1.08, index=92)
            self.define_sub_group("CONHCH2", R=1.5, Q=1.08, index=100)
        with self.define_main_group("HCONR", index=49):
            self.define_sub_group("HCONHCH3", R=2.4617, Q=2.192, index=93)
            self.define_sub_group("HCONHCH2", R=2.4617, Q=1.842, index=94)
        with self.define_main_group("CONR2", index=48):
            self.define_sub_group("AM(CH3)2", R=2.4748, Q=1.9643, index=101)
            self.define_sub_group("AMCH3CH2", R=2.2739, Q=1.5754, index=102)
            self.define_sub_group("AM(CH2)2", R=2.0767, Q=1.1866, index=103)
        with self.define_main_group("ACS", index=52):
            self.define_sub_group("AC2H2S", R=1.7943, Q=1.34, index=104)
            self.define_sub_group("AC2HS", R=1.6282, Q=1.06, index=105)
            self.define_sub_group("AC2S", R=1.4621, Q=0.78, index=106)
        with self.define_main_group("EPOXIDES", index=53):
            self.define_sub_group("H2COCH", R=1.3601, Q=1.8031, index=107)
            self.define_sub_group("COCH", R=0.683, Q=0.3418, index=108)
            self.define_sub_group("HCOCH", R=0.9104, Q=0.6538, index=109)
            self.define_sub_group("H2COCH2", R=1.063, Q=1.123, index=119)
            self.define_sub_group("H2COC", R=0.9104, Q=0.6538, index=153)
        with self.define_main_group("SULFONE", index=56):
            self.define_sub_group("(CH2)2SU", R=2.687, Q=2.12, index=110)
            self.define_sub_group("CH2SUCH", R=2.46, Q=1.808, index=111)
        with self.define_main_group("CARBONAT", index=55):
            self.define_sub_group("(CH3)2CB", R=2.42, Q=2.4976, index=112)
            self.define_sub_group("(CH2)2CB", R=2.42, Q=2.0018, index=113)
            self.define_sub_group("CH2CH3CB", R=2.42, Q=2.2497, index=114)
        with self.define_main_group("CH2S", index=61):
            self.define_sub_group("CH3S", R=1.613, Q=1.368, index=122)
            self.define_sub_group("CH2S", R=1.3863, Q=1.06, index=123)
            self.define_sub_group("CHS", R=1.1589, Q=0.748, index=124)
        with self.define_main_group("IMIDAZOL", index=84):
            self.define_sub_group("C3H2N2+", R=1.3662, Q=0.6797, index=178)
            self.define_sub_group("C3H3N2+", R=1.843, Q=1.6997, index=184)
        with self.define_main_group("BTI", index=85):
            self.define_sub_group("BTI-", R=5.621, Q=5.9463, index=179)
        with self.define_main_group("PYRROL", index=87):
            self.define_sub_group("C4H8N+", R=2.7867, Q=2.7723, index=189)
        with self.define_main_group("BF4", index=89):
            self.define_sub_group("BF4-", R=3.9628, Q=0.6214, index=195)
        with self.define_main_group("PYRIDIN", index=90):
            self.define_sub_group("C5H5N+", R=2.1094, Q=2.5106, index=196)
            self.define_sub_group("C5H4N+", R=2.4873, Q=2.4457, index=220)
        with self.define_main_group("OTF", index=91):
            self.define_sub_group("OTF-", R=3.371, Q=2.0001, index=197)
        with self.define_main_group("-S-S-", index=93):
            self.define_sub_group("-S-S-", R=1.0678, Q=2.244, index=201)
        with self.define_main_group("SO4", index=98):
            self.define_sub_group("SO4", R=0.9903, Q=3.5249, index=209)
            self.define_sub_group("HSO4", R=1.5654, Q=3.8076, index=210)
        with self.define_main_group("PF6", index=99):
            self.define_sub_group("PF6", R=3.8183, Q=3.6018, index=211)


    def define_interaction_table(self):
        self.set_interaction(1, 2, Aij=189.66, Bij=-0.27232, Aji=-95.41801, Bji=0.061708)
        self.set_interaction(1, 3, Aij=114.2, Bij=0.0933, Aji=16.07, Bji=-0.2998)
        self.set_interaction(1, 4, Aij=7.339, Bij=-0.4538, Aji=47.2, Bji=0.3575)
        self.set_interaction(1, 5, Aij=2777, Bij=-4.674, Cij=0.001551, Aji=1606, Bji=-4.746, Cji=0.0009181)
        self.set_interaction(1, 6, Aij=2409.4, Bij=-3.0099, Aji=82.593, Bji=-0.48575)
        self.set_interaction(1, 7, Aij=1391.3, Bij=-3.6156, Cij=0.001144, Aji=-17.253, Bji=0.8389, Cji=0.0009021)
        self.set_interaction(1, 8, Aij=1381, Bij=-0.9977, Aji=1987, Bji=-4.615)
        self.set_interaction(1, 9, Aij=433.6, Bij=0.1473, Aji=199, Bji=-0.8709)
        self.set_interaction(1, 10, Aij=875.85, Aji=256.21)
        self.set_interaction(1, 11, Aij=98.65601, Bij=1.9294, Cij=-0.0031331, Aji=632.22, Bji=-3.3912, Cji=0.0039282)
        self.set_interaction(1, 12, Aij=508.4, Bij=-0.6215, Aji=238.5, Bji=-0.5358)
        self.set_interaction(1, 13, Aij=233.1, Bij=-0.3155, Aji=-9.654, Bji=-0.03242)
        self.set_interaction(1, 14, Aij=-164.04, Bij=4.9683, Cij=-0.010252, Aji=326.04, Bji=-2.6348, Cji=0.0033576)
        self.set_interaction(1, 15, Aij=350.58, Bij=0.066729, Aji=207.26, Bji=-1.0916)
        self.set_interaction(1, 16, Aij=-175.7, Bij=1.857, Aji=205.65, Bji=-1.4436)
        self.set_interaction(1, 17, Aij=958.74, Bij=-0.14836, Aji=2257.3, Bji=-5.6676)
        self.set_interaction(1, 18, Aij=1802.3, Bij=-17.171, Cij=0.036, Aji=-436.15, Bji=3.4225, Cji=-0.0087)
        self.set_interaction(1, 19, Aij=593.07, Bij=0.7335, Aji=293.81, Bji=-1.3979)
        self.set_interaction(1, 20, Aij=1182.2, Bij=-3.2647, Cij=0.009198, Aji=2017.7, Bji=-9.0933, Cji=0.010238)
        self.set_interaction(1, 21, Aij=401, Bij=-0.72772, Aji=-65.68501, Bji=0.074091)
        self.set_interaction(1, 22, Aij=-233.66, Bij=1.2561, Aji=311.55, Bji=-1.1856)
        self.set_interaction(1, 23, Aij=-653.74, Bij=4.5311, Cij=-0.008735, Aji=1302.6, Bji=-8.427, Cji=0.014417)
        self.set_interaction(1, 24, Aij=267.51, Bij=-1.7109, Cij=0.003388, Aji=-148.07, Bji=1.0927, Cji=-0.002416)
        self.set_interaction(1, 25, Aij=-1385, Bij=15.89, Cij=-0.04831, Aji=3264, Bji=-20.84, Cji=0.03317)
        self.set_interaction(1, 26, Aij=2345, Bij=-13.2, Cij=0.02156, Aji=-396.5, Bji=3.092, Cji=-0.006266)
        self.set_interaction(1, 27, Aij=2383, Bij=-2.693, Aji=1744, Bji=-4.082)
        self.set_interaction(1, 28, Aij=24.33, Bij=1.521, Aji=72.12, Bji=-1.126)
        self.set_interaction(1, 29, Aij=465.9, Bij=-0.8557, Aji=-59.9, Bji=-0.008313)
        self.set_interaction(1, 30, Aij=577.7, Bij=0.9384, Aji=210.5, Bji=-1.081)
        self.set_interaction(1, 31, Aij=897.7, Aji=28.17)
        self.set_interaction(1, 32, Aij=559.9, Bij=-0.3564, Aji=166, Bji=-0.7116)
        self.set_interaction(1, 33, Aij=527.7, Bij=-0.499, Aji=-62.08, Bji=-0.3658)
        self.set_interaction(1, 34, Aij=477.5, Bij=-0.6581, Aji=-22.04, Bji=-0.1018)
        self.set_interaction(1, 35, Aij=-547.5, Bij=3.457, Aji=477.1, Bji=-2.175)
        self.set_interaction(1, 36, Aij=1662, Aji=-291.9)
        self.set_interaction(1, 37, Aij=334.5, Bij=0.009102, Aji=3.202, Bji=-0.6894)
        self.set_interaction(1, 38, Aij=468.5, Bij=-1.042, Aji=-160.3, Bji=-0.0108)
        self.set_interaction(1, 39, Aij=406.2, Bij=0.6525, Aji=151, Bji=-0.9023)
        self.set_interaction(1, 40, Aij=342, Bij=-1.679, Aji=-484.3, Bji=2.467)
        self.set_interaction(1, 41, Aij=1312, Bij=-3.643, Aji=-314.6, Bji=1.287)
        self.set_interaction(1, 42, Aij=-117.1, Bij=0.5481, Cij=-0.00098, Aji=170.9, Bji=-0.8062, Cji=0.001291)
        self.set_interaction(1, 43, Aij=79.507, Bij=0.70892, Cij=-0.0020983, Aji=186.71, Bji=-1.3546, Cji=0.0024016)
        self.set_interaction(1, 44, Aij=1935.7, Aji=-21.23)
        self.set_interaction(1, 45, Aij=164.25, Bij=-0.26834, Aji=-44.069, Bji=0.027778)
        self.set_interaction(1, 46, Aij=677.32, Bij=-2.0066, Aji=-249.85, Bji=1.7054)
        self.set_interaction(1, 47, Aij=3150.86, Bij=-3.1003, Aji=189.911, Bji=-0.9413)
        self.set_interaction(1, 48, Aij=1529.52, Bij=-6.2025, Cij=0.00975, Aji=82.6, Bji=-0.615, Cji=-0.000623)
        self.set_interaction(1, 49, Aij=-923.282, Bij=7.80959, Cij=-0.008808, Aji=183.02, Bji=-1.22588, Cji=0.000583)
        self.set_interaction(1, 52, Aij=24.432, Bij=0.2915, Aji=-16.034, Bji=-0.5337)
        self.set_interaction(1, 53, Aij=269.67, Aji=-41)
        self.set_interaction(1, 55, Aij=407.47, Bij=-0.2037, Cij=-0.00016, Aji=-0.067, Bji=-0.3935, Cji=4e-05)
        self.set_interaction(1, 56, Aij=1058.31, Bij=-1.1856, Aji=438.76, Bji=-1.2256)
        self.set_interaction(1, 61, Aij=860.51, Bij=-1.96009, Aji=-309.943, Bji=0.64856)
        self.set_interaction(1, 84, Aij=70.38, Bij=0.24511, Aji=77.7793, Bji=0.83298)
        self.set_interaction(1, 85, Aij=1260.02, Bij=-4.55446, Cij=0.0067607, Aji=996.5333, Bji=-5.48462, Cji=0.0095873)
        self.set_interaction(1, 87, Aij=283.8202, Bij=-3.9547, Cij=0.0103753, Aji=907.3842, Bji=0.46568, Cji=-0.0055872)
        self.set_interaction(1, 89, Aij=360.8016, Bij=2.05302, Cij=-0.0166992, Aji=454.5569, Bji=-1.47382,
                             Cji=0.0010454)
        self.set_interaction(1, 90, Aij=656.528, Bij=-4.2237, Cij=0.0074475, Aji=-307.8627, Bji=3.60996, Cji=-0.0081757)
        self.set_interaction(1, 91, Aij=1094.303, Bij=-1.1188, Aji=243.462, Bji=-1.03559)
        self.set_interaction(1, 93, Aij=468.398, Bij=-2.55668, Aji=2660.16, Bji=-6.86512)
        self.set_interaction(1, 98, Aij=1859.374, Bij=4.06308, Aji=1131.305, Bji=-0.93175)
        self.set_interaction(1, 99, Aij=4041.479, Bij=-3.90698, Cij=-0.0254482, Aji=3504.325, Bji=2.66288,
                             Cji=-0.0330202)
        self.set_interaction(2, 3, Aij=174.1, Bij=-0.5886, Aji=-157.2, Bji=0.6166)
        self.set_interaction(2, 4, Aij=117.3, Bij=-0.8552, Aji=-113.1, Bji=1.172)
        self.set_interaction(2, 5, Aij=2649, Bij=-6.508, Cij=0.004822, Aji=1566, Bji=-5.809, Cji=0.005197)
        self.set_interaction(2, 6, Aij=-628.07, Bij=10, Cij=-0.014972, Aji=-96.297, Bji=0.6304, Cji=-0.0018)
        self.set_interaction(2, 7, Aij=778.3, Bij=0.1482, Aji=-1301, Bji=4.072)
        self.set_interaction(2, 8, Aij=1207, Bij=-1.955, Aji=191.6, Bji=0.4936)
        self.set_interaction(2, 9, Aij=179.8, Bij=0.69911, Aji=91.811, Bji=-0.71715)
        self.set_interaction(2, 10, Aij=476.25, Aji=202.49)
        self.set_interaction(2, 11, Aij=980.74, Bij=-2.4224, Aji=-582.82, Bji=1.6732)
        self.set_interaction(2, 12, Aij=309.8, Aji=-28.63)
        self.set_interaction(2, 13, Aij=733.3, Bij=-2.509, Aji=-844.3, Bji=2.945)
        self.set_interaction(2, 14, Aij=1857, Bij=-8.653001, Cij=0.01088, Aji=498.8, Bji=-5.148, Cji=0.01039)
        self.set_interaction(2, 15, Aij=224.8, Aji=-124.32)
        self.set_interaction(2, 16, Aij=165.3, Aji=-131.5)
        self.set_interaction(2, 17, Aij=2800, Bij=-10.72, Cij=0.01339, Aji=3982, Bji=-19.72, Cji=0.02783)
        self.set_interaction(2, 18, Aij=13.502, Aji=-13.317)
        self.set_interaction(2, 19, Aij=634.85, Aji=-181.93)
        self.set_interaction(2, 20, Aij=-2026.1, Bij=8.1549, Aji=-347.5, Bji=1.216)
        self.set_interaction(2, 21, Aij=498.9, Bij=-1.487, Aji=-359.6, Bji=1.238)
        self.set_interaction(2, 22, Aij=-44.958, Aji=55.881)
        self.set_interaction(2, 23, Aij=-204.51, Aji=255.41)
        self.set_interaction(2, 24, Aij=616.62, Bij=-2.1164, Aji=-663.45, Bji=2.3281)
        self.set_interaction(2, 25, Aij=-56.69, Bij=9.805, Aji=215.5, Bji=-1.519)
        self.set_interaction(2, 26, Aij=417.6, Bij=0.08726, Aji=452.2, Bji=-1.997)
        self.set_interaction(2, 28, Aij=46.06, Bij=1.545, Aji=70.88, Bji=-1.09)
        self.set_interaction(2, 30, Aij=470.4, Aji=-146.1)
        self.set_interaction(2, 33, Aij=-19.82, Bij=0.5941, Aji=160.4, Bji=-0.5148)
        self.set_interaction(2, 34, Aij=642.27, Bij=-3.8641, Cij=0.0055, Aji=-197.06, Bji=1.8723, Cji=-0.0027)
        self.set_interaction(2, 35, Aij=-174.6, Bij=1.96, Aji=154, Bji=-1.303)
        self.set_interaction(2, 36, Aij=179.7, Aji=-113.8)
        self.set_interaction(2, 37, Aij=967.9, Bij=-0.03862, Aji=-300.6, Bji=-0.09576)
        self.set_interaction(2, 38, Aij=141.1, Bij=-0.3025, Aji=-139.5, Bji=0.9076)
        self.set_interaction(2, 39, Aij=388.4, Aji=-152.2)
        self.set_interaction(2, 41, Aij=-339.8, Bij=1.297, Aji=698.5, Bji=-2.159)
        self.set_interaction(2, 42, Aij=2.406, Bij=-0.1882, Aji=60.2, Bji=0.1565)
        self.set_interaction(2, 43, Aij=-322.1, Bij=-0.20367, Cij=0.004517, Aji=1182.6, Bji=-5, Cji=0.0037455)
        self.set_interaction(2, 45, Aij=389.28, Aji=-174.41)
        self.set_interaction(2, 46, Aij=491.23, Bij=-1.8285, Aji=-734.87, Bji=3.3463)
        self.set_interaction(2, 47, Aij=673.2271, Bij=1.9523, Aji=92.7, Bji=-0.8135)
        self.set_interaction(2, 48, Aij=249.18, Aji=-81.79)
        self.set_interaction(2, 49, Aij=1171.32, Bij=3.92127, Aji=90.0134, Bji=-0.920017)
        self.set_interaction(2, 52, Aij=-86.231, Aji=-39.373)
        self.set_interaction(2, 53, Aij=9.389999, Aji=57.86)
        self.set_interaction(2, 55, Aij=-456.08, Bij=2.4334, Aji=298.55, Bji=-1.4153)
        self.set_interaction(2, 56, Aij=775.56, Aji=-65.66)
        self.set_interaction(2, 61, Aij=324.316, Aji=-156.458)
        self.set_interaction(2, 84, Aij=-44.3408, Bij=-0.65212, Aji=1753.135, Bji=-3.112)
        self.set_interaction(2, 85, Aij=566.2915, Bij=-0.14057, Aji=-1213.93, Bji=5.14012)
        self.set_interaction(2, 87, Aij=-92.7302, Bij=-0.25088, Aji=-1151.296, Bji=4.7125)
        self.set_interaction(2, 89, Aij=1230.781, Bij=0.17033, Cij=0.0141503, Aji=4896.219, Bji=-1.86077,
                             Cji=-0.0254894)
        self.set_interaction(2, 90, Aij=1625.909, Bij=-2.40107, Aji=159.8103, Bji=1.85317)
        self.set_interaction(2, 91, Aij=2979.973, Bij=4.82369, Cij=-0.0272505, Aji=724.9113, Bji=-4.79411,
                             Cji=0.0056279)
        self.set_interaction(2, 93, Aij=-740.2421, Aji=232.905)
        self.set_interaction(2, 98, Aij=2487.632, Bij=-3.28182, Aji=1804.025, Bji=-5.18107)
        self.set_interaction(2, 99, Aij=-2399.738, Bij=-0.27451, Cij=0.0329518, Aji=658.1962, Bji=0.57714,
                             Cji=-0.0078001)
        self.set_interaction(3, 4, Aij=139.2, Bij=-0.65, Aji=-45.33, Bji=0.4223)
        self.set_interaction(3, 5, Aij=3972, Bij=-13.16, Cij=0.01208, Aji=3049, Bji=-12.77, Cji=0.01435)
        self.set_interaction(3, 6, Aij=1604.3, Bij=-2.0299, Aji=13.733, Bji=-0.11768)
        self.set_interaction(3, 7, Aij=792, Bij=-1.726, Aji=332.3, Bji=1.158)
        self.set_interaction(3, 8, Aij=1356, Bij=-2.118, Aji=2340, Bji=-5.043)
        self.set_interaction(3, 9, Aij=146.2, Bij=-1.237, Cij=0.004237, Aji=-57.53, Bji=1.212, Cji=-0.003715)
        self.set_interaction(3, 10, Aij=-365.5, Bij=1.874, Aji=1011, Bji=-2.167)
        self.set_interaction(3, 11, Aij=-274.54, Bij=0.91491, Aji=622.73, Bji=-1.7605)
        self.set_interaction(3, 12, Aij=170.5, Bij=-0.02393, Aji=108.3, Bji=-0.262)
        self.set_interaction(3, 13, Aij=-87.08, Bij=-0.1859, Aji=179, Bji=0.05615)
        self.set_interaction(3, 14, Aij=2036, Bij=-8.729, Cij=0.008138, Aji=-121, Bji=-1.901, Cji=0.006999)
        self.set_interaction(3, 15, Aij=139.67, Bij=0.037692, Aji=105.63, Bji=-0.60667)
        self.set_interaction(3, 16, Aij=-71.4, Bij=0.7078, Aji=16.29, Bji=-0.6022)
        self.set_interaction(3, 17, Aij=1044.7, Bij=-1.7112, Aji=154.39, Bji=1.2458)
        self.set_interaction(3, 18, Aij=-1553.9, Bij=0.1615, Cij=0.0082, Aji=1810.8, Bji=-5.7594, Cji=0.0023)
        self.set_interaction(3, 19, Aij=-17.44, Bij=0.9437, Aji=111.8, Bji=-0.5959)
        self.set_interaction(3, 20, Aij=69.561, Bij=1.8881, Aji=613.32, Bji=-1.595)
        self.set_interaction(3, 21, Aij=73.04601, Bij=-0.21322, Aji=-58.972, Bji=0.1046)
        self.set_interaction(3, 22, Aij=133.66, Bij=-0.46141, Aji=-142.2, Bji=0.39662)
        self.set_interaction(3, 23, Aij=66.214, Bij=-0.63629, Aji=-78.116, Bji=0.69975)
        self.set_interaction(3, 24, Aij=269, Bij=-1.776, Cij=0.002645, Aji=-305.5, Bji=2.12, Cji=-0.003239)
        self.set_interaction(3, 25, Aij=595.2, Bij=3.309, Cij=-0.02844, Aji=1885, Bji=-10.98, Cji=0.01661)
        self.set_interaction(3, 26, Aij=134.1, Bij=-0.8156, Cij=0.00145, Aji=-330, Bji=3.066, Cji=-0.005376)
        self.set_interaction(3, 27, Aij=936.3391, Bij=-4.4141, Cij=0.012232, Aji=1866.23, Bji=-9.28137, Cji=0.008299)
        self.set_interaction(3, 28, Aij=3736, Bij=-25, Cij=0.04593, Aji=574.6, Bji=-3.702, Cji=0.003682)
        self.set_interaction(3, 29, Aij=77.083, Bij=-0.3094, Aji=-2.1662, Bji=0.1196)
        self.set_interaction(3, 30, Aij=331.6, Bij=0.3778, Aji=0.4086, Bji=-0.4601)
        self.set_interaction(3, 31, Aij=-148.14, Bij=2.2658, Cij=-0.00353, Aji=-93.18999, Bji=1.5036, Cji=-0.00141)
        self.set_interaction(3, 32, Aij=-82.28, Bij=0.5677, Aji=298.9, Bji=-0.8374)
        self.set_interaction(3, 33, Aij=-248.2, Bij=1.214, Aji=187.5, Bji=-0.9020001)
        self.set_interaction(3, 35, Aij=347.6, Bij=-1.43, Aji=-345.6, Bji=1.545)
        self.set_interaction(3, 37, Aij=602.1, Bij=-7.798, Cij=0.01966, Aji=1887, Bji=-8.707001, Cji=0.007813)
        self.set_interaction(3, 38, Aij=808, Bij=-5.331, Cij=0.006077, Aji=-1367, Bji=7.883, Cji=-0.007754)
        self.set_interaction(3, 39, Aij=-81.68999, Bij=1.1229, Cij=-0.00105, Aji=58.76, Bji=-0.6263, Cji=0.000668)
        self.set_interaction(3, 40, Aij=123.26, Bij=-0.6133, Aji=419.04, Bji=-0.1703)
        self.set_interaction(3, 41, Aij=-126.2, Bij=0.386, Aji=670.8, Bji=-1.807)
        self.set_interaction(3, 42, Aij=134.6, Bij=-1.231, Cij=0.001488, Aji=-2.619, Bji=1.094, Cji=-0.001557)
        self.set_interaction(3, 43, Aij=-26.852, Bij=-0.44215, Aji=47.23, Bji=0.64039)
        self.set_interaction(3, 44, Aij=-1172, Bij=10.106, Cij=-0.01428, Aji=-1141.6, Bji=8.656199, Cji=-0.015439)
        self.set_interaction(3, 45, Aij=380.02, Bij=-0.23326, Aji=-201.52, Bji=-0.068774)
        self.set_interaction(3, 46, Aij=313.79, Bij=-1.1552, Aji=-258.12, Bji=1.4084)
        self.set_interaction(3, 47, Aij=1083.63, Bij=0.9144, Cij=-0.002141, Aji=160.13, Bji=-1.3578, Cji=0.001331)
        self.set_interaction(3, 48, Aij=92.5, Bij=1.4186, Cij=1.8e-05, Aji=-14.89, Bji=-0.8126, Cji=2e-06)
        self.set_interaction(3, 49, Aij=-103.146, Bij=2.6565, Cij=-0.002819, Aji=45.4296, Bji=-0.570848, Cji=0.000281)
        self.set_interaction(3, 52, Aij=1412, Bij=-3.9917, Aji=-660.25, Bji=1.4602)
        self.set_interaction(3, 53, Aij=392.56, Aji=-260.22)
        self.set_interaction(3, 55, Aij=849.08, Bij=-6.5044, Cij=0.01219, Aji=-358.06, Bji=3.6831, Cji=-0.00782)
        self.set_interaction(3, 56, Aij=-522.32, Bij=3.8856, Cij=-0.004212, Aji=562.72, Bji=-2.4316, Cji=0.00242)
        self.set_interaction(3, 61, Aij=454.316, Bij=-1.21943, Aji=-388.19, Bji=1.04305)
        self.set_interaction(3, 84, Aij=277.4415, Bij=-0.1843, Aji=-223.7499, Bji=0.2781)
        self.set_interaction(3, 85, Aij=825.9189, Bij=-4.35541, Cij=0.0085572, Aji=20.2719, Bji=-0.33187, Cji=0.0025793)
        self.set_interaction(3, 87, Aij=-1304.107, Bij=3.87435, Cij=0.0039496, Aji=26.3894, Bji=0.38631, Cji=-0.0027586)
        self.set_interaction(3, 89, Aij=626.9603, Bij=-4.30606, Aji=2116.398, Bji=-2.11062)
        self.set_interaction(3, 90, Aij=161.7745, Bij=-0.56713, Aji=785.4488, Bji=-3.09114)
        self.set_interaction(3, 91, Aij=1690.065, Bij=-3.38808, Cij=0.0202432, Aji=-15.3556, Bji=-0.15571,
                             Cji=-0.0001404)
        self.set_interaction(3, 93, Aij=-630.416, Aji=-72.2694)
        self.set_interaction(3, 98, Aij=1014.686, Bij=-1.73087, Aji=877.6001, Bji=-1.96581)
        self.set_interaction(3, 99, Aij=4899.846, Bij=-3.53433, Cij=-0.0358313, Aji=-969.6266, Bji=3.81305,
                             Cji=0.0028389)
        self.set_interaction(4, 5, Aij=3989, Bij=-14.09, Cij=0.0153, Aji=2673, Bji=-5.765, Cji=-0.000332)
        self.set_interaction(4, 6, Aij=436.21, Bij=1.9094, Aji=145.54, Bji=-0.48799)
        self.set_interaction(4, 7, Aij=1050.2, Bij=-1.9939, Aji=24.144, Bji=1.6504)
        self.set_interaction(4, 8, Aij=1375, Bij=-1.702, Aji=1825, Bji=-3.743)
        self.set_interaction(4, 9, Aij=1001, Bij=-1.871, Cij=0.000239, Aji=-146.6, Bji=0.2419, Cji=0.0001133)
        self.set_interaction(4, 10, Aij=683.6, Bij=-1.02, Cij=0.000869, Aji=1963, Bji=2.656, Cji=-0.01355)
        self.set_interaction(4, 11, Aij=-242.5, Bij=2.72, Cij=-0.003449, Aji=1624, Bji=-9.409, Cji=0.01338)
        self.set_interaction(4, 12, Aij=136.98, Aji=377.26)
        self.set_interaction(4, 13, Aij=-595.1, Bij=2.978, Aji=375, Bji=-1.57)
        self.set_interaction(4, 14, Aij=2977, Bij=-19.16, Cij=0.03333, Aji=-45.44, Bji=5.141, Cji=-0.0142)
        self.set_interaction(4, 15, Aij=1250, Aji=-316.22)
        self.set_interaction(4, 16, Aij=-2631, Bij=13.56, Cij=-0.007036, Aji=978.3, Bji=-6.481, Cji=0.007088)
        self.set_interaction(4, 17, Aij=4000, Bij=-16.68, Cij=0.02112, Aji=3969, Bji=8.497001, Cji=-0.005945)
        self.set_interaction(4, 18, Aij=135.3, Bij=-2.369, Cij=0.0039, Aji=1698.1, Bji=-9.8887, Cji=0.0153)
        self.set_interaction(4, 19, Aij=208.1, Bij=1.333, Aji=170.1, Bji=-0.8218)
        self.set_interaction(4, 20, Aij=1352.5, Aji=29.747)
        self.set_interaction(4, 21, Aij=-46.994, Bij=0.30443, Aji=113.07, Bji=-0.59569)
        self.set_interaction(4, 22, Aij=213.85, Bij=0.19998, Aji=-75.01, Bji=-0.42684)
        self.set_interaction(4, 23, Aij=192.52, Bij=-0.25363, Aji=-38.939, Bji=-0.19378)
        self.set_interaction(4, 24, Aij=-106.2, Bij=0.6081, Aji=107.8, Bji=-0.6785)
        self.set_interaction(4, 25, Aij=-113.6, Bij=19.72, Aji=-69.23, Bji=-0.7359)
        self.set_interaction(4, 26, Aij=1358, Bij=-9.968, Cij=0.02093, Aji=1014, Bji=-4.702, Cji=0.004381)
        self.set_interaction(4, 27, Aij=391.044, Bij=0.196357, Cij=2.24992e-05, Aji=1931.39, Bji=-5.15065,
                             Cji=4.51638e-07)
        self.set_interaction(4, 28, Aij=2586, Bij=-9.5, Cij=0.008819, Aji=-1494, Bji=7.679, Cji=-0.01225)
        self.set_interaction(4, 30, Aij=157.9, Bij=1.408, Aji=181.2, Bji=-1.081)
        self.set_interaction(4, 31, Aij=1856.45, Bij=5.9959, Cij=-0.01862, Aji=1041.33, Bji=-5.2068, Cji=0.00717)
        self.set_interaction(4, 32, Aij=69, Bij=0.4317, Aji=448.8, Bji=-0.9919)
        self.set_interaction(4, 33, Aij=277, Bij=-0.5955, Aji=-156.7, Bji=0.1221)
        self.set_interaction(4, 35, Aij=88.93, Bij=0.2236, Aji=178.6, Bji=-0.7113)
        self.set_interaction(4, 37, Aij=234.2, Bij=-0.2311, Aji=-115.9, Bji=-0.02925)
        self.set_interaction(4, 38, Aij=-172.2, Bij=0.2257, Aji=-13.15, Bji=0.001763)
        self.set_interaction(4, 39, Aij=444.29, Bij=0.1947, Aji=64.28, Bji=-0.5839)
        self.set_interaction(4, 40, Aij=29.57, Aji=153.32)
        self.set_interaction(4, 41, Aij=2303, Bij=-6.346, Aji=-736.8, Bji=3.332)
        self.set_interaction(4, 42, Aij=-107.1, Bij=0.2564, Aji=191.5, Bji=-0.5561)
        self.set_interaction(4, 43, Aij=-26.486, Bij=-0.29519, Aji=199.48, Bji=0.033695)
        self.set_interaction(4, 44, Aij=-514.79, Bij=4.9372, Aji=291.65, Bji=-1.2039)
        self.set_interaction(4, 45, Aij=297.73, Bij=-0.24204, Aji=-248.3, Bji=0.25471)
        self.set_interaction(4, 46, Aij=72.26, Bij=-0.1919, Aji=763.57, Bji=-1.3961)
        self.set_interaction(4, 47, Aij=1696.96, Bij=-0.7064, Aji=340.91, Bji=-1.008)
        self.set_interaction(4, 48, Aij=76.73, Bij=0.9394, Cij=-0.00045, Aji=246.08, Bji=-0.8198, Cji=-0.000137)
        self.set_interaction(4, 49, Aij=-182.325, Bij=3.36729, Cij=-0.004625, Aji=119.767, Bji=-0.608934, Cji=0.000607)
        self.set_interaction(4, 52, Aij=1000.8, Bij=-2.5194, Aji=-139.78, Bji=0.1146)
        self.set_interaction(4, 53, Aij=164.05, Aji=102.04)
        self.set_interaction(4, 55, Aij=558.08, Bij=0.565, Cij=-0.00331, Aji=562.34, Bji=-3.4373, Cji=0.00633)
        self.set_interaction(4, 56, Aij=-307.42, Bij=4.1819, Cij=-0.007723, Aji=1661.26, Bji=-8.697001, Cji=0.01424)
        self.set_interaction(4, 61, Aij=629.463, Bij=-1.39039, Aji=-305.273, Bji=0.669085)
        self.set_interaction(4, 84, Aij=930.7075, Bij=-4.30401, Cij=0.007841, Aji=365.3965, Bji=-1.20387,
                             Cji=-0.0007792)
        self.set_interaction(4, 85, Aij=-313.9384, Bij=2.06128, Aji=-134.6708, Bji=1.38462)
        self.set_interaction(4, 87, Aij=-935.3928, Bij=1.26386, Cij=0.0041351, Aji=4101.374, Bji=-1.19824,
                             Cji=-0.0186009)
        self.set_interaction(4, 89, Aij=524.3056, Bij=-4.26859, Aji=1259.47, Bji=-1.91791)
        self.set_interaction(4, 90, Aij=453.4621, Bij=-2.03845, Aji=3997.18, Bji=-2.994)
        self.set_interaction(4, 91, Aij=1077.47, Bij=-2.43883, Aji=-998.0735, Bji=4.48792)
        self.set_interaction(4, 93, Aij=-478.068, Aji=15.0501)
        self.set_interaction(4, 98, Aij=1113.436, Bij=-0.428312, Aji=-281.4692, Bji=1.00738)
        self.set_interaction(4, 99, Aij=335.1104, Bij=2.26653, Cij=0.0239277, Aji=3792.062, Bji=-2.55407,
                             Cji=-0.0217293)
        self.set_interaction(5, 6, Aij=346.31, Bij=-2.4583, Cij=0.0029287, Aji=-1218.2, Bji=9.7928, Cji=-0.016158)
        self.set_interaction(5, 7, Aij=-801.9, Bij=3.824, Cij=-0.007514, Aji=1460, Bji=-8.673, Cji=0.01641)
        self.set_interaction(5, 8, Aij=83.91, Bij=-1.262, Aji=465.4, Bji=-1.841)
        self.set_interaction(5, 9, Aij=-250, Bij=2.857, Cij=-0.006022, Aji=653.3, Bji=-1.412, Cji=0.000954)
        self.set_interaction(5, 10, Aij=-281.4, Bij=2.379, Cij=-0.006668, Aji=1590, Bji=-24.57, Cji=0.06212)
        self.set_interaction(5, 11, Aij=973.8, Bij=-5.633, Cij=0.00769, Aji=310.4, Bji=1.538, Cji=-0.004885)
        self.set_interaction(5, 12, Aij=235.9, Bij=-0.5874, Aji=839.6, Bji=-1.215)
        self.set_interaction(5, 13, Aij=816.7, Bij=-5.092, Cij=0.006065, Aji=650.9, Bji=-0.7132, Cji=0.000815)
        self.set_interaction(5, 14, Aij=-923.7, Bij=2.468, Aji=-75.63, Bji=-0.1511)
        self.set_interaction(5, 15, Aij=-355.1, Bij=0.58, Aji=-660.2, Bji=1.743)
        self.set_interaction(5, 16, Aij=104.6, Bij=-5.014, Cij=0.008854, Aji=1876, Bji=11.5, Cji=0.09)
        self.set_interaction(5, 17, Aij=-1114, Bij=5.916, Cij=-0.007126, Aji=1325, Bji=-6.263, Cji=0.007584)
        self.set_interaction(5, 18, Aij=-3061.2, Bij=9.5413, Cij=0.1842, Aji=-643.09, Bji=2.9346, Cji=-0.0055)
        self.set_interaction(5, 19, Aij=123.5, Bij=0.8503, Cij=-0.002478, Aji=-46, Bji=4.415, Cji=-0.00878)
        self.set_interaction(5, 20, Aij=-1295, Bij=4.3634, Aji=1525.8, Bji=-4.9155)
        self.set_interaction(5, 21, Aij=238.1, Bij=-0.7077, Aji=2177, Bji=-4.363)
        self.set_interaction(5, 22, Aij=-126, Bij=0.1322, Aji=2389, Bji=-4.509)
        self.set_interaction(5, 23, Aij=1314.8, Aji=963.37)
        self.set_interaction(5, 24, Aij=925.6, Bij=-2.027, Aji=3139, Bji=-5.964)
        self.set_interaction(5, 25, Aij=1862, Bij=32.07, Cij=-0.009397, Aji=3664, Bji=34.13, Cji=0.002987)
        self.set_interaction(5, 26, Aij=741.8, Bij=-2.098, Aji=1091, Bji=-1.274)
        self.set_interaction(5, 27, Aij=2100, Aji=316.6)
        self.set_interaction(5, 30, Aij=738.4, Bij=-1.771, Aji=616.5, Bji=-0.1457)
        self.set_interaction(5, 31, Aij=499.8, Bij=-2.41, Aji=-468.8, Bji=2.421)
        self.set_interaction(5, 32, Aij=838.8, Aji=774.7)
        self.set_interaction(5, 33, Aij=699.7, Bij=-1.767, Aji=1439, Bji=-1.673)
        self.set_interaction(5, 34, Aij=-148.9, Bij=1.034, Aji=1255, Bji=-2.538)
        self.set_interaction(5, 35, Aij=190.4, Bij=-1.2, Aji=-452.3, Bji=1.956)
        self.set_interaction(5, 36, Aij=1117, Bij=-2.196, Aji=1072, Bji=-1.248)
        self.set_interaction(5, 37, Aij=439.4, Aji=959.3)
        self.set_interaction(5, 38, Aij=848.6, Aji=1253)
        self.set_interaction(5, 39, Aij=1036, Bij=-2.995, Aji=-366.4, Bji=1.129)
        self.set_interaction(5, 41, Aij=403.8, Bij=-0.9346, Aji=703.4, Bji=-1.383)
        self.set_interaction(5, 42, Aij=3121, Bij=-13.69, Cij=0.01446, Aji=2601, Bji=-1.25, Cji=-0.006309)
        self.set_interaction(5, 43, Aij=401.89, Bij=-0.43627, Cij=-0.002004, Aji=-238.36, Bji=5, Cji=-0.008186501)
        self.set_interaction(5, 45, Aij=-32.643, Bij=-0.010434, Aji=2985.8, Bji=-6.227)
        self.set_interaction(5, 46, Aij=1201.3, Bij=-4.6803, Aji=-516.99, Bji=2.1888)
        self.set_interaction(5, 47, Aij=186.24, Bij=-3.1349, Cij=0.007737, Aji=1460.82, Bji=-5.962, Cji=0.004691)
        self.set_interaction(5, 48, Aij=1627.81, Bij=-5.8598, Cij=0.005497, Aji=-808.4, Bji=3.8321, Cji=-0.005877)
        self.set_interaction(5, 49, Aij=-150.961, Bij=0.527951, Aji=868.681, Bji=-2.20103)
        self.set_interaction(5, 52, Aij=-3745, Bij=4.5137, Cij=0.0636, Aji=7712.2, Bji=-7.6896, Cji=-0.035)
        self.set_interaction(5, 53, Aij=495.28, Bij=-2.6496, Cij=0.00159, Aji=700.05, Bji=0.2237, Cji=-0.00244)
        self.set_interaction(5, 56, Aij=34.94, Bij=1.0075, Cij=-0.002547, Aji=942.22, Bji=-2.3183, Cji=0.001188)
        self.set_interaction(5, 61, Aij=1105.94, Bij=-1.68763, Aji=1343.95, Bji=-2.16488)
        self.set_interaction(5, 84, Aij=292.6468, Bij=-1.59145, Aji=147.0663, Bji=2.08434)
        self.set_interaction(5, 85, Aij=1026.29, Bij=-3.17777, Aji=-792.988, Bji=3.87681)
        self.set_interaction(5, 87, Aij=-348.7564, Bij=1.92474, Aji=1003.498, Bji=-2.48421)
        self.set_interaction(5, 89, Aij=-2448.212, Bij=-2.13997, Cij=0.0332832, Aji=-234.3916, Bji=4.01945,
                             Cji=-0.0134709)
        self.set_interaction(5, 90, Aij=1178.048, Bij=5.33766, Aji=1044.876, Bji=-3.71273)
        self.set_interaction(5, 91, Aij=-159.8631, Bij=2.41804, Aji=662.1286, Bji=-3.28127)
        self.set_interaction(5, 93, Aij=-367.321, Aji=34.7132)
        self.set_interaction(5, 98, Aij=225.0235, Bij=-4.3271, Cij=0.0030855, Aji=448.381, Bji=-1.61662, Cji=0.000518)
        self.set_interaction(5, 99, Aij=-352.184, Bij=-1.58363, Cij=0.0090199, Aji=1751.43, Bji=-1.09998,
                             Cji=-0.0148923)
        self.set_interaction(6, 7, Aij=-328.5, Bij=1.0823, Cij=-0.0022, Aji=-524.3, Bji=4.6065, Cji=-0.004)
        self.set_interaction(6, 8, Aij=-867, Bij=-1.258, Cij=0.02998, Aji=265.5, Bji=-2.905, Cji=0.002283)
        self.set_interaction(6, 9, Aij=86.439, Bij=-0.46505, Aji=394.78, Bji=-0.36048)
        self.set_interaction(6, 10, Aij=-392.5, Bij=2.256, Aji=-158.4, Bji=-0.6469)
        self.set_interaction(6, 11, Aij=299.23, Bij=-1.2702, Aji=294.76, Bji=0.3745)
        self.set_interaction(6, 12, Aij=220.7, Bij=-0.6402, Aji=444.7, Bji=-0.06819)
        self.set_interaction(6, 13, Aij=-87.48, Bij=-0.5522, Aji=475.2, Bji=0.1198)
        self.set_interaction(6, 14, Aij=-495.25, Bij=1.0807, Aji=-467.95, Bji=0.65743)
        self.set_interaction(6, 15, Aij=-1508.5, Bij=4.4917, Aji=-278.09, Bji=-0.39888)
        self.set_interaction(6, 16, Aij=-1039, Bij=5.603, Cij=-0.006551, Aji=39.33, Bji=-3.354, Cji=0.006714)
        self.set_interaction(6, 17, Aij=-2012, Bij=13.46, Cij=-0.02004, Aji=251.2, Bji=-1.274, Cji=0.002214)
        self.set_interaction(6, 18, Aij=-341.34, Bij=-5.8042, Cij=0.0141, Aji=-230.38, Bji=-0.3825, Cji=0.0023)
        self.set_interaction(6, 19, Aij=97.97301, Bij=-0.28666, Aji=615.01, Bji=-0.9444399)
        self.set_interaction(6, 20, Aij=-733.07, Bij=2.3351, Aji=1075.5, Bji=-3.4339)
        self.set_interaction(6, 21, Aij=-16.521, Bij=-0.28137, Aji=1831.2, Bji=-2.9694)
        self.set_interaction(6, 22, Aij=-85.926, Bij=-0.26371, Aji=1904.4, Bji=-2.7981)
        self.set_interaction(6, 23, Aij=-139.58, Aji=893.38)
        self.set_interaction(6, 24, Aij=-40.13, Bij=-0.06709, Aji=2150, Bji=-2.813)
        self.set_interaction(6, 25, Aij=3000, Bij=-11.81, Aji=2955, Bji=-6.383)
        self.set_interaction(6, 26, Aij=374.2, Bij=-2.212, Cij=0.002688, Aji=1079, Bji=-1.766, Cji=0.001238)
        self.set_interaction(6, 28, Aij=-332.4, Bij=4.419, Cij=-0.007797, Aji=2645, Bji=-9.172001, Cji=0.01177)
        self.set_interaction(6, 29, Aij=75.71, Bij=-0.3753, Aji=1334, Bji=-1.863)
        self.set_interaction(6, 30, Aij=-369.8, Bij=2.203, Aji=662, Bji=-1.903)
        self.set_interaction(6, 31, Aij=33.19, Bij=-0.2074, Aji=-3.428, Bji=0.483)
        self.set_interaction(6, 32, Aij=180.5, Bij=0.3161, Aji=1965, Bji=-3.858)
        self.set_interaction(6, 33, Aij=28.95, Bij=-0.3303, Aji=648.8, Bji=0.8050001)
        self.set_interaction(6, 35, Aij=-185.9, Bij=0.03906, Aji=145, Bji=-0.6343)
        self.set_interaction(6, 36, Aij=164, Aji=135.9)
        self.set_interaction(6, 37, Aij=-43.88, Aji=2421)
        self.set_interaction(6, 38, Aij=-99.57999, Aji=1235)
        self.set_interaction(6, 39, Aij=101.2, Bij=-1.078, Aji=-269.7, Bji=1.732)
        self.set_interaction(6, 41, Aij=308.7, Bij=-1.175, Aji=678.1, Bji=-1.335)
        self.set_interaction(6, 42, Aij=68.972, Bij=-0.41997, Aji=2540.7, Bji=-3.5236)
        self.set_interaction(6, 43, Aij=-308.7, Bij=1.7454, Cij=-0.00335, Aji=952.24, Bji=-3.3287, Cji=0.0053598)
        self.set_interaction(6, 45, Aij=-242.6, Bij=0.29562, Cij=-9.0055e-05, Aji=4519.3, Bji=-19.456, Cji=0.025981)
        self.set_interaction(6, 47, Aij=-119.08, Bij=0.9938, Aji=-52.69, Bji=-0.7465)
        self.set_interaction(6, 48, Aij=210.91, Bij=-2.4953, Cij=0.0048, Aji=-132.93, Bji=0.6584, Cji=-0.00208)
        self.set_interaction(6, 49, Aij=197.395, Bij=-0.985958, Aji=-195.127, Bji=1.17652)
        self.set_interaction(6, 52, Aij=673.72, Bij=-2.3567, Aji=2676.7, Bji=-5.4899)
        self.set_interaction(6, 53, Aij=137.95, Bij=-1.4112, Cij=0.00172, Aji=524.52, Bji=0.4314, Cji=-0.00166)
        self.set_interaction(6, 56, Aij=166.42, Bij=0.994, Cij=-0.003346, Aji=1683.48, Bji=-8.003099, Cji=0.01159)
        self.set_interaction(6, 61, Aij=34.7884, Aji=641.041)
        self.set_interaction(6, 84, Aij=-704.2285, Bij=-1.26236, Cij=0.0149627, Aji=152.7192, Bji=2.17856,
                             Cji=-0.0098745)
        self.set_interaction(6, 85, Aij=-348.9494, Bij=1.41462, Aji=2127.895, Bji=-4.14737)
        self.set_interaction(6, 87, Aij=-1230.497, Bij=3.91001, Aji=1142.897, Bji=-2.92427)
        self.set_interaction(6, 89, Aij=-671.2604, Bij=1.58636, Cij=-0.0048637, Aji=985.4429, Bji=-2.42145,
                             Cji=-0.0062036)
        self.set_interaction(6, 90, Aij=53.7776, Bij=-0.9791, Aji=94.629, Bji=0.30192)
        self.set_interaction(6, 91, Aij=1741.061, Bij=-4.73718, Aji=-558.6751, Bji=1.11398)
        self.set_interaction(6, 93, Aij=1216.71, Bij=-5.44835, Aji=-386.355, Bji=3.77874)
        self.set_interaction(6, 98, Aij=1037.56, Bij=-5.46099, Aji=1421.557, Bji=0.34143)
        self.set_interaction(7, 8, Aij=-2686, Bij=19.44, Cij=-0.02702, Aji=148.4, Bji=-2.757, Cji=0.002329)
        self.set_interaction(7, 9, Aij=190.5, Bij=-3.669, Cij=0.008838, Aji=770.6, Bji=-0.5873, Cji=-0.003252)
        self.set_interaction(7, 10, Aij=-1545, Bij=6.512, Aji=512.6, Bji=-2.145)
        self.set_interaction(7, 11, Aij=-433.288, Bij=3.0862, Cij=-0.002012, Aji=311.974, Bji=-1.3412, Cji=0.001074)
        self.set_interaction(7, 12, Aij=140.71, Aji=53.28)
        self.set_interaction(7, 13, Aij=177.665, Bij=-3.72906, Cij=0.010763, Aji=433.207, Bji=-0.605276, Cji=-0.000914)
        self.set_interaction(7, 14, Aij=798.5, Bij=-5.869, Cij=0.01032, Aji=-980.6, Bji=3.671, Cji=-0.005908)
        self.set_interaction(7, 15, Aij=1524, Bij=-2.531, Aji=-851, Bji=1.034)
        self.set_interaction(7, 16, Aij=274.5, Bij=-0.5905, Cij=0.002205, Aji=-446, Bji=-0.7738, Cji=0.002634)
        self.set_interaction(7, 17, Aij=158.4, Bij=0.5246, Aji=-131, Bji=-0.7957)
        self.set_interaction(7, 18, Aij=-3178.5, Bij=9.8213, Cij=-0.0034, Aji=-393.18, Bji=0.5546, Cji=-0.0029)
        self.set_interaction(7, 19, Aij=-634.1, Bij=3.59, Aji=509.6, Bji=-1.916)
        self.set_interaction(7, 20, Aij=-1795.2, Bij=12.708, Cij=-0.015455, Aji=624.97, Bji=-4.6878, Cji=0.0052371)
        self.set_interaction(7, 21, Aij=86.68999, Bij=1.592, Aji=313.3, Bji=-0.5041)
        self.set_interaction(7, 22, Aij=134.1, Bij=0.9495, Aji=748.2, Bji=-1.319)
        self.set_interaction(7, 24, Aij=1008, Bij=-1.795, Aji=1282, Bji=-2.815)
        self.set_interaction(7, 25, Aij=-1895, Bij=9.303, Aji=591.6, Bji=-3.08)
        self.set_interaction(7, 26, Aij=-595.7, Bij=2.634, Aji=882.6, Bji=-2.606)
        self.set_interaction(7, 30, Aij=-123.8, Bij=3.847, Aji=501.4, Bji=-1.939)
        self.set_interaction(7, 31, Aij=372.5, Bij=-0.9091, Aji=-368.8, Bji=0.7775)
        self.set_interaction(7, 33, Aij=822.2, Aji=-17.99)
        self.set_interaction(7, 35, Aij=117, Bij=-0.611, Aji=-370.8, Bji=0.1043)
        self.set_interaction(7, 36, Aij=419.8, Bij=2.436, Aji=276.9, Bji=-0.9948)
        self.set_interaction(7, 39, Aij=-494.2, Bij=2.826, Aji=-121.8, Bji=-0.6029)
        self.set_interaction(7, 41, Aij=676, Bij=-0.9909, Aji=808.4, Bji=-2.929)
        self.set_interaction(7, 42, Aij=274.37, Bij=-0.5861, Cij=-0.00030011, Aji=1632.9, Bji=-2.8719, Cji=0.003455)
        self.set_interaction(7, 43, Aij=-75.7467, Bij=-0.98514, Cij=0.0033318, Aji=717.485, Bji=-1.48515,
                             Cji=-7.437899e-14)
        self.set_interaction(7, 44, Aij=-804.28, Bij=2.828, Aji=594.45, Bji=-2.2535)
        self.set_interaction(7, 45, Aij=509.3, Bij=-0.29, Aji=-523.8, Bji=3.158)
        self.set_interaction(7, 46, Aij=659.22, Bij=-1.8841, Aji=-588.21, Bji=0.9707)
        self.set_interaction(7, 47, Aij=934.67, Bij=-4.4232, Cij=0.006054, Aji=-649.31, Bji=1.3487, Cji=-0.001586)
        self.set_interaction(7, 48, Aij=64.43999, Bij=-0.0094, Aji=-439.58, Bji=0.3142)
        self.set_interaction(7, 49, Aij=-1119.8, Bij=7.95909, Cij=-0.008555, Aji=-128.903, Bji=-0.468519, Cji=-0.000916)
        self.set_interaction(7, 53, Aij=499.44, Bij=-6.6553, Cij=0.01656, Aji=952.57, Bji=-0.8437, Cji=-0.0048)
        self.set_interaction(7, 56, Aij=-302.93, Bij=0.5975, Aji=652.86, Bji=-1.8013)
        self.set_interaction(7, 61, Aij=1038.3, Aji=-270.53)
        self.set_interaction(7, 84, Aij=659.2109, Bij=-0.59252, Cij=0.0451602, Aji=151.31, Bji=-3.00895, Cji=0.0019814)
        self.set_interaction(7, 85, Aij=2770.851, Bij=-2.00944, Aji=1996.724, Bji=-5.47448)
        self.set_interaction(7, 89, Aij=2707.161, Bij=-2.51124, Cij=-0.0022558, Aji=-399.1902, Bji=0.16195,
                             Cji=-0.0067585)
        self.set_interaction(7, 90, Aij=-1616.174, Bij=5.32305, Aji=275.7479, Bji=-1.97057)
        self.set_interaction(7, 91, Aij=318.2128, Bij=-1.42376, Aji=385.0068, Bji=-2.49001)
        self.set_interaction(7, 93, Aij=4878.28, Bij=-6.38758, Aji=-444.615, Bji=-0.776483)
        self.set_interaction(7, 98, Aij=206.2641, Bij=-2.57259, Cij=0.0032526, Aji=-371.2031, Bji=0.99046,
                             Cji=-0.000807)
        self.set_interaction(8, 9, Aij=-145.2, Bij=-0.738, Aji=-666.8, Bji=1.918)
        self.set_interaction(8, 10, Aij=5.604, Aji=-410.21)
        self.set_interaction(8, 11, Aij=-212.9, Aji=-224.4)
        self.set_interaction(8, 13, Aij=-329.3, Aji=-80.58)
        self.set_interaction(8, 17, Aij=542, Bij=-6.792, Cij=0.01655, Aji=-131.1, Bji=-0.9399, Cji=0.000469)
        self.set_interaction(8, 18, Aij=-4080.9, Bij=12.859, Cij=-0.0101, Aji=-41.594, Bji=-3.5887, Cji=0.004)
        self.set_interaction(8, 20, Aij=401.88, Aji=281.08)
        self.set_interaction(8, 24, Aij=2356, Bij=-3.347, Aji=2157, Bji=-3.718)
        self.set_interaction(8, 25, Aij=555.5, Aji=1554)
        self.set_interaction(8, 31, Aij=-309, Aji=191.7)
        self.set_interaction(8, 42, Aij=391.2, Bij=-0.8799, Aji=3630, Bji=-6.23)
        self.set_interaction(8, 52, Aij=4911.4, Bij=9.8501, Aji=2332.5, Bji=-4.0832)
        self.set_interaction(9, 10, Aij=197.6, Aji=-93.07999)
        self.set_interaction(9, 11, Aij=-16.486, Bij=-0.27924, Aji=33.415, Bji=0.21913)
        self.set_interaction(9, 12, Aij=-83.57, Aji=101.3)
        self.set_interaction(9, 13, Aij=3645, Bij=-26.91, Cij=0.04757, Aji=695.8, Bji=-0.9619, Cji=-0.002462)
        self.set_interaction(9, 15, Aij=-47.97, Aji=119.5)
        self.set_interaction(9, 16, Aij=-389.6, Bij=0.1944, Cij=0.001863, Aji=2831, Bji=-13.01, Cji=0.01558)
        self.set_interaction(9, 17, Aij=1732, Bij=-9.896001, Cij=0.0141, Aji=1460, Bji=-13.73, Cji=0.02917)
        self.set_interaction(9, 18, Aij=-1908.7, Bij=4.1057, Aji=307.16, Bji=-0.8161)
        self.set_interaction(9, 19, Aij=-191, Bij=0.6835, Aji=79.08, Bji=-0.3808)
        self.set_interaction(9, 20, Aij=-109.51, Bij=0.96888, Aji=178.22, Bji=-0.91676)
        self.set_interaction(9, 21, Aij=-99.97601, Aji=55.27)
        self.set_interaction(9, 22, Aij=-18.695, Bij=-0.52606, Aji=-218.94, Bji=1.0749)
        self.set_interaction(9, 23, Aij=810.17, Bij=-3.2209, Cij=0.0021443, Aji=-48.641, Bji=-0.79498, Cji=0.0037129)
        self.set_interaction(9, 24, Aij=-208.71, Bij=0.8293, Cij=-0.0014, Aji=155.73, Bji=1.4412, Cji=-0.0025)
        self.set_interaction(9, 25, Aij=1297, Bij=-1.365, Cij=-0.02253, Aji=1375, Bji=2.499, Cji=0.006309)
        self.set_interaction(9, 26, Aij=-35.89, Bij=-0.1009, Aji=-32.6, Bji=0.0847)
        self.set_interaction(9, 27, Aij=-169.6, Bij=2.986, Aji=-328.1, Bji=0.3045)
        self.set_interaction(9, 28, Aij=419.9, Bij=0.09772, Aji=315.3, Bji=-0.5617)
        self.set_interaction(9, 29, Aij=2.714, Aji=64.41)
        self.set_interaction(9, 30, Aij=-986, Bij=3.847, Aji=277, Bji=-1.357)
        self.set_interaction(9, 31, Aij=478.5, Bij=-1.148, Aji=-72.58, Bji=0.4909)
        self.set_interaction(9, 32, Aij=346.6, Bij=-0.7017, Aji=-182, Bji=0.7905)
        self.set_interaction(9, 33, Aij=-717.76, Bij=2.9431, Aji=319.69, Bji=-1.4362)
        self.set_interaction(9, 34, Aij=-62.43, Bij=-0.8977, Aji=-481.2, Bji=9.604)
        self.set_interaction(9, 35, Aij=-76.87, Bij=0.5372, Aji=38.06, Bji=-0.5189)
        self.set_interaction(9, 37, Aij=64.01, Bij=-1.077, Aji=-153.4, Bji=3.143)
        self.set_interaction(9, 39, Aij=80.79201, Bij=-3.702, Cij=0.011586, Aji=1955.5, Bji=-10, Cji=0.011386)
        self.set_interaction(9, 41, Aij=64.21, Aji=-148.3)
        self.set_interaction(9, 42, Aij=437.739, Bij=-2.7983, Cij=0.00364, Aji=364.423, Bji=2.1022, Cji=-0.004653)
        self.set_interaction(9, 43, Aij=-62.857, Bij=0.2898, Aji=80.038, Bji=-0.10124)
        self.set_interaction(9, 45, Aij=-497.98, Bij=0.79715, Cij=0.00029817, Aji=945.14, Bji=-3.8168, Cji=0.0045351)
        self.set_interaction(9, 47, Aij=502.1, Bij=0.3401, Aji=141.01, Bji=-0.9829)
        self.set_interaction(9, 48, Aij=37.37, Aji=-116.4)
        self.set_interaction(9, 49, Aij=410.655, Bij=-1.1056, Aji=-201.105, Bji=0.7123)
        self.set_interaction(9, 52, Aij=-153.7, Bij=0.242, Aji=2311.5, Bji=-6.6054)
        self.set_interaction(9, 84, Aij=173.4451, Bij=-5.52695, Cij=0.0160979, Aji=958.3726, Bji=-4.73012,
                             Cji=0.0047161)
        self.set_interaction(9, 85, Aij=824.1235, Bij=-2.88368, Cij=0.0041569, Aji=670.0023, Bji=-5.05233,
                             Cji=0.0074865)
        self.set_interaction(9, 87, Aij=99.5504, Bij=0.45575, Aji=-1504.191, Bji=4.63058)
        self.set_interaction(9, 89, Aij=147.2528, Bij=-2.74448, Aji=177.8232, Bji=-2.24998)
        self.set_interaction(9, 90, Aij=331.1518, Bij=-1.11421, Aji=-30.955, Bji=-0.71244)
        self.set_interaction(9, 91, Aij=571.6566, Bij=-4.43975, Aji=-1449.573, Bji=4.44554)
        self.set_interaction(9, 98, Aij=519.4855, Bij=-5.4673, Cij=0.0068391, Aji=3306.365, Bji=0.15196, Cji=-0.0082641)
        self.set_interaction(10, 11, Aij=-208.4, Aji=389.7)
        self.set_interaction(10, 12, Aij=-160.7, Aji=226.6)
        self.set_interaction(10, 13, Aij=209, Bij=-0.6241, Aji=235.7, Bji=0.1314)
        self.set_interaction(10, 20, Aij=435.64, Aji=-188)
        self.set_interaction(10, 21, Aij=985.7, Bij=-2.986, Aji=-888.3, Bji=3.426)
        self.set_interaction(10, 22, Aij=-41.537, Aji=354.71)
        self.set_interaction(10, 29, Aij=-373.7, Aji=-397.5)
        self.set_interaction(10, 30, Aij=-742.7, Aji=-214.8)
        self.set_interaction(10, 32, Aij=114.3, Aji=293.5)
        self.set_interaction(10, 34, Aij=236.6, Aji=-93.066)
        self.set_interaction(10, 37, Aij=-43.56, Aji=945.6)
        self.set_interaction(10, 39, Aij=2371, Aji=-225.3)
        self.set_interaction(10, 42, Aij=716.7, Bij=-1.516, Aji=1161, Bji=-0.5724)
        self.set_interaction(11, 12, Aij=342.4, Aji=-251.7)
        self.set_interaction(11, 13, Aij=195.3, Bij=-9.75, Cij=0.04051, Aji=824.2, Bji=-6.009, Cji=0.008271)
        self.set_interaction(11, 15, Aij=3168, Bij=-24.07, Cij=0.04303, Aji=3329, Bji=-13.78, Cji=0.01193)
        self.set_interaction(11, 16, Aij=152.8, Bij=-1.099, Aji=160.8, Bji=0.8719)
        self.set_interaction(11, 17, Aij=-1355, Bij=7.683001, Cij=-0.01012, Aji=3499, Bji=-22.96, Cji=0.03543)
        self.set_interaction(11, 18, Aij=87.6, Bij=1.4109, Aji=45.309, Bji=-1.4281)
        self.set_interaction(11, 19, Aij=-193.23, Bij=0.43006, Aji=139.55, Bji=-0.43665)
        self.set_interaction(11, 20, Aij=62.031, Bij=1.0567, Aji=59.594, Bji=-0.71198)
        self.set_interaction(11, 21, Aij=-49.339, Aji=48.852)
        self.set_interaction(11, 22, Aij=168.17, Bij=-1.0536, Aji=-461.35, Bji=1.8569)
        self.set_interaction(11, 24, Aij=-5.71, Bij=-0.2724, Aji=223.4, Bji=0.1237)
        self.set_interaction(11, 25, Aij=3351, Bij=-14.54, Aji=-788.6, Bji=1.693)
        self.set_interaction(11, 26, Aij=9.222, Bij=-0.3292, Aji=-50.36, Bji=0.02448)
        self.set_interaction(11, 28, Aij=861.1, Bij=-1.693, Aji=280, Bji=-0.9491)
        self.set_interaction(11, 30, Aij=80.68999, Aji=-136.3)
        self.set_interaction(11, 31, Aij=-72.07, Aji=69.25)
        self.set_interaction(11, 32, Aij=82.96, Aji=11.62)
        self.set_interaction(11, 33, Aij=-386.3, Bij=1.892, Aji=248.3, Bji=-1.198)
        self.set_interaction(11, 35, Aij=296.8, Bij=-1.264, Aji=-337.1, Bji=0.8843)
        self.set_interaction(11, 36, Aij=-92.12, Bij=0.9031, Aji=503.5, Bji=-1.792)
        self.set_interaction(11, 37, Aij=-201.4, Bij=0.5487, Aji=-320, Bji=1.052)
        self.set_interaction(11, 39, Aij=96.77, Bij=0.4827, Aji=16.69, Bji=-0.5966)
        self.set_interaction(11, 41, Aij=-338.8, Bij=1.837, Aji=3.924, Bji=-1.174)
        self.set_interaction(11, 42, Aij=374.1, Bij=-1.976, Cij=0.001682, Aji=460.8, Bji=-0.06206, Cji=4.1e-05)
        self.set_interaction(11, 43, Aij=-28.231, Aji=36.948)
        self.set_interaction(11, 44, Aij=745.4, Aji=-447.04)
        self.set_interaction(11, 45, Aij=-579.11, Bij=0.9455, Aji=966.35, Bji=-2.1861)
        self.set_interaction(11, 47, Aij=758.13, Bij=-0.2593, Aji=32.71, Bji=-0.8641)
        self.set_interaction(11, 48, Aij=611.47, Bij=-2.4783, Aji=-1031.78, Bji=3.9906)
        self.set_interaction(11, 49, Aij=2973.58, Bij=-5.45555, Aji=81.6384, Bji=-1.3111)
        self.set_interaction(11, 56, Aij=-1336.2, Bij=6.7742, Cij=-0.0082, Aji=1915.2, Bji=-9.9994, Cji=0.0126)
        self.set_interaction(11, 61, Aij=-1366.59, Bij=5.90673, Cij=-0.007325, Aji=2622.04, Bji=-11.7013, Cji=0.013472)
        self.set_interaction(12, 18, Aij=-9.2978, Aji=-123.73)
        self.set_interaction(12, 19, Aij=92.21, Aji=-33.64)
        self.set_interaction(12, 24, Aij=-142.2, Bij=0.572, Aji=465.8, Bji=-0.7730001)
        self.set_interaction(12, 25, Aij=1894, Aji=18.79)
        self.set_interaction(12, 29, Aij=161.8, Aji=13.97)
        self.set_interaction(12, 37, Aij=745.4, Bij=-0.622, Aji=-479.1, Bji=0.8031)
        self.set_interaction(12, 39, Aij=580.3, Aji=-285.5)
        self.set_interaction(12, 42, Aij=227.6, Bij=-0.7762, Aji=509, Bji=-0.2248)
        self.set_interaction(12, 44, Aij=489.15, Aji=-441.01)
        self.set_interaction(12, 45, Aij=260.64, Bij=-1.2868, Aji=-597.09, Bji=2.5295)
        self.set_interaction(13, 18, Aij=-199.94, Bij=-6.9867, Cij=0.0193, Aji=-430.49, Bji=-0.728, Cji=0.004)
        self.set_interaction(13, 19, Aij=1987, Bij=-8.022, Cij=0.01065, Aji=-588.8, Bji=1.481, Cji=-0.002636)
        self.set_interaction(13, 20, Aij=521.48, Aji=-310.82)
        self.set_interaction(13, 21, Aij=-208.6, Bij=-0.2571, Cij=0.002418, Aji=872, Bji=-2.939, Cji=0.001269)
        self.set_interaction(13, 22, Aij=492.9, Bij=-6.475, Cij=0.01806, Aji=215.3, Bji=-2.482, Cji=0.002745)
        self.set_interaction(13, 23, Aij=-607.35, Bij=2.3467, Aji=97.12801, Bji=-0.64386)
        self.set_interaction(13, 24, Aij=-425.4, Bij=0.9514, Aji=641.2, Bji=-1.486)
        self.set_interaction(13, 25, Aij=974, Bij=-1.368, Cij=-0.01983, Aji=381.1, Bji=-5.682, Cji=0.01675)
        self.set_interaction(13, 26, Aij=-305.1, Bij=0.7063, Aji=319.6, Bji=-1.368)
        self.set_interaction(13, 28, Aij=35.02, Bij=1.702, Aji=198.5, Bji=-1.434)
        self.set_interaction(13, 29, Aij=102.6, Aji=-210.1)
        self.set_interaction(13, 30, Aij=513.7, Aji=-299.6)
        self.set_interaction(13, 32, Aij=-104.8, Aji=464)
        self.set_interaction(13, 34, Aij=-137.87, Aji=169.27)
        self.set_interaction(13, 37, Aij=-422.7, Bij=2.23, Aji=-326.4, Bji=0.24)
        self.set_interaction(13, 38, Aij=155.7, Bij=-1.578, Aji=-528.8, Bji=2.822)
        self.set_interaction(13, 42, Aij=397, Bij=-1.379, Aji=-214.1, Bji=1.134)
        self.set_interaction(13, 43, Aij=-124.33, Bij=-0.29398, Aji=561.14, Bji=-0.7058)
        self.set_interaction(13, 44, Aij=-454.92, Aji=310.75)
        self.set_interaction(13, 45, Aij=-515.93, Bij=0.3835, Aji=1368, Bji=-2.6254)
        self.set_interaction(13, 48, Aij=1321.52, Aji=-361.25)
        self.set_interaction(13, 49, Aij=3858.07, Bij=-9.278001, Cij=7e-06, Aji=-224.782, Bji=-0.410609, Cji=-4e-06)
        self.set_interaction(13, 56, Aij=-209.209, Bij=-0.470423, Cij=0.006918, Aji=1163.91, Bji=-7.1332, Cji=0.00839)
        self.set_interaction(13, 93, Aij=-584.728, Aji=178.501)
        self.set_interaction(14, 15, Aij=1517, Bij=-12.72, Cij=0.02557, Aji=-1074, Bji=9, Cji=-0.01795)
        self.set_interaction(14, 16, Aij=-472.4, Bij=-0.2051, Cij=0.01058, Aji=836.6, Bji=-5.208, Cji=0.004801)
        self.set_interaction(14, 19, Aij=-412.38, Bij=-0.49092, Cij=0.0062553, Aji=2412.2, Bji=-10.495, Cji=0.0097411)
        self.set_interaction(14, 24, Aij=-65.76, Bij=-0.3148, Aji=333.9, Bji=-0.1415)
        self.set_interaction(14, 25, Aij=2553, Bij=-11.9, Aji=3873, Bji=-9.315001)
        self.set_interaction(14, 29, Aij=-205.1, Aji=244.4)
        self.set_interaction(14, 39, Aij=162.14, Aji=-112.76)
        self.set_interaction(14, 42, Aij=124.3, Bij=-1.147, Aji=621.9, Bji=0.258)
        self.set_interaction(14, 43, Aij=-143.07, Aji=182.58)
        self.set_interaction(14, 47, Aij=413.11, Aji=-293.93)
        self.set_interaction(14, 56, Aij=190.882, Aji=14.9127)
        self.set_interaction(15, 16, Aij=402.6, Bij=-1.614, Aji=-639.9, Bji=2.561)
        self.set_interaction(15, 19, Aij=242.2, Aji=-131.9)
        self.set_interaction(15, 24, Aij=-3.28, Aji=43.83)
        self.set_interaction(15, 25, Aij=3888, Bij=-16.26, Aji=-868.8, Bji=2.948)
        self.set_interaction(15, 38, Aij=-330.2, Aji=904.1)
        self.set_interaction(15, 39, Aij=-88.49, Aji=-230.55)
        self.set_interaction(15, 42, Aij=-131.9, Bij=-0.007369, Aji=1248, Bji=-2.79)
        self.set_interaction(15, 43, Aij=-186.98, Aji=295.07)
        self.set_interaction(15, 47, Aij=444.67, Bij=0.8351, Aji=-196.23, Bji=-0.0511)
        self.set_interaction(15, 56, Aij=1178.79, Bij=-9.99911, Cij=0.020758, Aji=1667.89, Bji=-7.58704, Cji=0.008006)
        self.set_interaction(16, 22, Aij=-473, Bij=0.8883, Aji=406.8, Bji=-1.524)
        self.set_interaction(16, 24, Aij=215.9, Bij=-1.399, Aji=-825.9, Bji=3.15)
        self.set_interaction(16, 25, Aij=1622, Bij=-4.812, Cij=-0.01856, Aji=-94.87, Bji=-9.612, Cji=0.03722)
        self.set_interaction(16, 38, Aij=-7.532001, Bij=-0.8077, Aji=35.16, Bji=0.09722999)
        self.set_interaction(16, 39, Aij=965, Aji=-311.9)
        self.set_interaction(16, 42, Aij=324.15, Bij=-1.4264, Aji=-198.32, Bji=1.452)
        self.set_interaction(16, 45, Aij=-420.24, Bij=0.26315, Aji=-1035.8, Bji=3.078)
        self.set_interaction(16, 47, Aij=-729.5, Bij=6.2811, Aji=-66.61, Bji=-0.696)
        self.set_interaction(16, 48, Aij=2111.24, Aji=-235.61)
        self.set_interaction(16, 56, Aij=2662.02, Bij=-13.9855, Cij=0.017738, Aji=-1780.41, Bji=5.71975, Cji=-0.00165)
        self.set_interaction(17, 18, Aij=74.285, Bij=14.613, Cij=-0.0283, Aji=-558.39, Bji=1.2436, Cji=-0.0007)
        self.set_interaction(17, 19, Aij=393.9, Bij=-4.703, Cij=0.009003, Aji=2987, Bji=-9.336, Cji=0.007147)
        self.set_interaction(17, 21, Aij=582.1, Bij=-2.33, Cij=0.000377, Aji=-338, Bji=3.372, Cji=-0.003676)
        self.set_interaction(17, 24, Aij=3986, Bij=-16.15, Cij=0.01635, Aji=2626, Bji=-10.59, Cji=0.01466)
        self.set_interaction(17, 25, Aij=3709.2, Aji=1583.8)
        self.set_interaction(17, 27, Aij=3770, Bij=1.686, Aji=1.655, Bji=-1.164)
        self.set_interaction(17, 31, Aij=1268, Bij=-3.042, Aji=-818.8, Bji=3.229)
        self.set_interaction(17, 39, Aij=-391.9, Aji=650.7)
        self.set_interaction(17, 42, Aij=1371, Bij=-3.323, Aji=1091, Bji=-0.2657)
        self.set_interaction(17, 47, Aij=1271.42, Bij=-6.4373, Cij=0.006451, Aji=-2476.18, Bji=10.9894, Cji=-0.009617)
        self.set_interaction(18, 19, Aij=746.12, Bij=-2.1937, Aji=-2307.8, Bji=5.8186)
        self.set_interaction(18, 20, Aij=-451.49, Bij=-0.4031, Aji=-2617.7, Bji=4.614)
        self.set_interaction(18, 21, Aij=-67.106, Bij=-0.4933, Aji=-1592.8, Bji=2.8397)
        self.set_interaction(18, 22, Aij=-978.25, Bij=2.1884, Aji=946.79, Bji=-4.9886)
        self.set_interaction(18, 24, Aij=229.66, Bij=-1.5005, Aji=-2111, Bji=6.3625)
        self.set_interaction(18, 33, Aij=-713.65, Bij=1.9191, Aji=-763.19, Bji=0.603)
        self.set_interaction(18, 38, Aij=197.95, Bij=-0.3316, Cij=-0.0004, Aji=-1052.5, Bji=1.6854, Cji=0.0027)
        self.set_interaction(18, 42, Aij=268.23, Bij=-1.4137, Aji=-5894.1, Bji=13.088)
        self.set_interaction(18, 43, Aij=-250.25, Bij=0.3811, Aji=-1269.7, Bji=2.6335)
        self.set_interaction(18, 45, Aij=-516.31, Bij=1.6943, Aji=-1646.8, Bji=2.7237)
        self.set_interaction(18, 47, Aij=489.93, Bij=-0.5591, Cij=-0.00102, Aji=-2250.64, Bji=5.4638, Cji=-0.00015)
        self.set_interaction(18, 52, Aij=-674, Bij=1.8854, Aji=4998.6, Bji=1.1962)
        self.set_interaction(19, 21, Aij=176.5, Bij=-1.237, Aji=-368.7, Bji=1.992)
        self.set_interaction(19, 22, Aij=-78.96, Aji=14.76)
        self.set_interaction(19, 24, Aij=65.82, Bij=-0.6265, Aji=357.6, Bji=0.7676)
        self.set_interaction(19, 25, Aij=1283, Bij=3.361, Cij=-0.02978, Aji=2331, Bji=-9.238001, Cji=0.01158)
        self.set_interaction(19, 26, Aij=117.53, Bij=-0.44692, Aji=-128.21, Bji=0.50349)
        self.set_interaction(19, 28, Aij=468.8, Aji=434.8)
        self.set_interaction(19, 29, Aij=-18.8, Bij=-0.3652, Aji=41.54, Bji=0.646)
        self.set_interaction(19, 31, Aij=506.6, Bij=-1.263, Aji=11.72, Bji=0.7004)
        self.set_interaction(19, 33, Aij=-211.2, Aji=362.6)
        self.set_interaction(19, 34, Aij=11.65, Bij=-1.229, Aji=-1428, Bji=7.698)
        self.set_interaction(19, 36, Aij=267.1, Aji=-144.7)
        self.set_interaction(19, 37, Aij=61.96, Bij=-0.4161, Aji=-19.1, Bji=1.195)
        self.set_interaction(19, 39, Aij=57.08, Bij=-0.7417, Aji=-160.48, Bji=1.0644)
        self.set_interaction(19, 41, Aij=-75.67, Aji=26.8)
        self.set_interaction(19, 42, Aij=256.2, Bij=-1.526, Cij=0.001118, Aji=1336, Bji=-3.395, Cji=0.004586)
        self.set_interaction(19, 43, Aij=-28.653, Bij=-0.48152, Aji=56.754, Bji=0.89781)
        self.set_interaction(19, 45, Aij=237.42, Bij=-1.2928, Aji=-642.44, Bji=2.8574)
        self.set_interaction(19, 49, Aij=297.646, Bij=-0.66206, Aji=-116.794, Bji=0.54754)
        self.set_interaction(19, 56, Aij=-50.8443, Bij=-0.025287, Aji=180.813, Bji=-0.0641231)
        self.set_interaction(20, 21, Aij=27.618, Aji=702.4)
        self.set_interaction(20, 22, Aij=94.606, Aji=425.97)
        self.set_interaction(20, 24, Aij=701.95, Bij=-1.7576, Aji=213.34, Bji=2.1861)
        self.set_interaction(20, 25, Aij=-1398.7, Aji=1000)
        self.set_interaction(20, 32, Aij=146.06, Aji=780.71)
        self.set_interaction(20, 33, Aij=-18.328, Aji=753.21)
        self.set_interaction(20, 37, Aij=-447.95, Bij=1.5141, Aji=283.64, Bji=1.5491)
        self.set_interaction(20, 39, Aij=-421.21, Aji=93.773)
        self.set_interaction(20, 41, Aij=-271.176, Bij=1.4107, Aji=218.974, Bji=-0.8061)
        self.set_interaction(20, 42, Aij=1060, Bij=-2.822, Aji=578.3, Bji=1.493)
        self.set_interaction(20, 43, Aij=720.45, Bij=-1.5187, Aji=-140.77, Bji=0.309)
        self.set_interaction(20, 44, Aij=-65.631, Aji=-14.016)
        self.set_interaction(20, 45, Aij=508.72, Bij=-1.4005, Aji=-386.93, Bji=2.3961)
        self.set_interaction(20, 47, Aij=-369.31, Aji=-0.29)
        self.set_interaction(20, 48, Aij=-400.86, Aji=-244.69)
        self.set_interaction(21, 22, Aij=70.79, Aji=-66.21)
        self.set_interaction(21, 23, Aij=592.4, Bij=-4.2459, Cij=0.0069046, Aji=603.29, Bji=-3.977, Cji=0.0062484)
        self.set_interaction(21, 24, Aij=16.34, Bij=0.07287, Aji=95.05, Bji=-0.2348)
        self.set_interaction(21, 25, Aij=3985, Bij=-15.7, Aji=15.62, Bji=-1.099)
        self.set_interaction(21, 26, Aij=24.44, Bij=-0.4713, Aji=142.1, Bji=-0.153)
        self.set_interaction(21, 27, Aij=1248, Bij=-2.04, Aji=1295, Bji=-4.224)
        self.set_interaction(21, 28, Aij=295.9, Aji=-137.7)
        self.set_interaction(21, 30, Aij=666, Aji=-390.6)
        self.set_interaction(21, 33, Aij=128.8, Bij=-0.2077, Aji=-92.68, Bji=-0.01307)
        self.set_interaction(21, 37, Aij=280, Aji=-207.3)
        self.set_interaction(21, 39, Aij=-70.45, Aji=-59.29)
        self.set_interaction(21, 42, Aij=-31.42, Bij=-0.248, Aji=370.6, Bji=-0.322)
        self.set_interaction(21, 43, Aij=-325.77, Bij=2.0412, Aji=70.075, Bji=-1.149)
        self.set_interaction(21, 44, Aij=530.3, Aji=17.052)
        self.set_interaction(21, 45, Aij=207.12, Bij=-0.43964, Aji=-175.29, Bji=0.32745)
        self.set_interaction(21, 49, Aij=955.8281, Bij=-1.00012, Aji=-151.036, Bji=-0.458824)
        self.set_interaction(21, 56, Aij=842.486, Bij=-1.8434, Aji=-313.202, Bji=0.4139)
        self.set_interaction(22, 23, Aij=187.43, Bij=-3.446, Cij=0.0067179, Aji=1468.9, Bji=-5, Cji=0.003701)
        self.set_interaction(22, 24, Aij=46.29, Bij=-0.2115, Aji=46.03, Bji=0.05388)
        self.set_interaction(22, 25, Aij=3353, Bij=-14.2, Aji=368.6, Bji=-1.748)
        self.set_interaction(22, 26, Aij=822.4, Bij=-2.05, Aji=-423.1, Bji=0.8154)
        self.set_interaction(22, 30, Aij=-174.6, Aji=106.3)
        self.set_interaction(22, 32, Aij=132.7, Bij=-0.1183, Aji=-23.81, Bji=-0.09203999)
        self.set_interaction(22, 33, Aij=-139.6, Bij=1.022, Aji=96.4, Bji=-0.776)
        self.set_interaction(22, 35, Aij=-178.3, Bij=0.7426, Aji=-39.45, Bji=-0.8556)
        self.set_interaction(22, 37, Aij=160.7, Aji=-135.9)
        self.set_interaction(22, 39, Aij=-147.64, Aji=75.45)
        self.set_interaction(22, 42, Aij=10.7, Bij=-0.2702, Aji=224.4, Bji=-0.1399)
        self.set_interaction(22, 43, Aij=108.83, Bij=-0.86059, Aji=-358.57, Bji=1.3307)
        self.set_interaction(22, 45, Aij=7.3664, Bij=0.40456, Aji=-1.6641, Bji=-0.37825)
        self.set_interaction(22, 48, Aij=102.8, Bij=-0.0879, Aji=-446.86, Bji=0.4276)
        self.set_interaction(22, 49, Aij=-1017.26, Bij=3.0909, Aji=894.509, Bji=-2.93164)
        self.set_interaction(22, 56, Aij=262.138, Aji=-229.081)
        self.set_interaction(23, 24, Aij=-323.17, Bij=1.1973, Aji=350.92, Bji=-1.3456)
        self.set_interaction(23, 33, Aij=599.82, Aji=-364.76)
        self.set_interaction(23, 37, Aij=325.81, Aji=-199.87)
        self.set_interaction(23, 42, Aij=289.08, Bij=-0.5852, Aji=-69.60101, Bji=-0.4274)
        self.set_interaction(23, 56, Aij=401.698, Aji=157.634)
        self.set_interaction(24, 25, Aij=-131.8, Bij=9.802, Cij=-0.03582, Aji=972.1, Bji=-6.82, Cji=0.009219)
        self.set_interaction(24, 26, Aij=441.5, Bij=-0.5353, Aji=-65.74, Bji=0.0967)
        self.set_interaction(24, 27, Aij=3286, Aji=167.5)
        self.set_interaction(24, 28, Aij=9.362, Bij=1.033, Aji=52.01, Bji=-0.9095)
        self.set_interaction(24, 30, Aij=750.2, Bij=0.8165, Aji=100.5, Bji=-0.8269)
        self.set_interaction(24, 32, Aij=49.51, Bij=0.6829, Aji=186.4, Bji=-0.7294)
        self.set_interaction(24, 33, Aij=203.2, Bij=-1.328, Aji=-1360, Bji=7.402)
        self.set_interaction(24, 35, Aij=325.2, Bij=0.4405, Aji=-60.89, Bji=-0.6321)
        self.set_interaction(24, 36, Aij=902, Aji=-194.9)
        self.set_interaction(24, 37, Aij=220.6, Bij=0.3756, Aji=-134.4, Bji=-0.3226)
        self.set_interaction(24, 38, Aij=197.4, Bij=-0.4858, Aji=-98.98, Bji=-0.02128)
        self.set_interaction(24, 39, Aij=512.7, Bij=0.02702, Aji=-168.4, Bji=-0.123)
        self.set_interaction(24, 42, Aij=-37.183, Bij=-0.047827, Aji=60.78, Bji=0.024262)
        self.set_interaction(24, 43, Aij=190.45, Bij=0.012716, Aji=-131.87, Bji=-0.014204)
        self.set_interaction(24, 45, Aij=22.779, Bij=0.042145, Aji=14.947, Bji=-0.11086)
        self.set_interaction(24, 49, Aij=516.966, Bij=0.19471, Cij=0.001855, Aji=-198.677, Bji=0.435084, Cji=-0.001271)
        self.set_interaction(24, 52, Aij=1366.3, Bij=-3.8385, Aji=-580.46, Bji=1.1653)
        self.set_interaction(24, 56, Aij=120.59, Bij=1.2816, Aji=261.15, Bji=-0.8815)
        self.set_interaction(25, 26, Aij=3986, Bij=25, Aji=3638, Bji=-14.25)
        self.set_interaction(25, 27, Aij=-184.5, Bij=2.2806, Aji=2926.02, Bji=-1.8141)
        self.set_interaction(25, 33, Aij=-69.88, Aji=981.5)
        self.set_interaction(25, 39, Aij=-12.71, Bij=-0.4758, Aji=472.49, Bji=-0.5503)
        self.set_interaction(25, 42, Aij=207.16, Bij=-1.2206, Aji=223.782, Bji=-0.5826)
        self.set_interaction(25, 43, Aij=96.855, Bij=-1.2993, Aji=2991.9, Bji=-9.3959)
        self.set_interaction(25, 45, Aij=-27.161, Bij=0.611, Aji=4235.3, Bji=-16.954)
        self.set_interaction(25, 49, Aij=593.203, Bij=-0.7904, Aji=88.32201, Bji=0.1605)
        self.set_interaction(26, 27, Aij=85.6, Aji=986)
        self.set_interaction(26, 28, Aij=68.87, Bij=3.217, Aji=655.7, Bji=-2.203)
        self.set_interaction(26, 32, Aij=643.8, Bij=-0.7376, Aji=17.81, Bji=-0.2245)
        self.set_interaction(26, 33, Aij=9.258, Bij=-0.1079, Aji=121.4, Bji=0.05397)
        self.set_interaction(26, 34, Aij=-70.24, Aji=132.2)
        self.set_interaction(26, 37, Aij=159, Aji=108.4)
        self.set_interaction(26, 39, Aij=606.9, Aji=-340.9)
        self.set_interaction(26, 42, Aij=119.3, Bij=-0.5691, Aji=522.9, Bji=-0.2485)
        self.set_interaction(26, 43, Aij=53.75, Aji=-47.089)
        self.set_interaction(26, 52, Aij=324.62, Bij=-1.1828, Aji=11.442, Bji=0.7866)
        self.set_interaction(27, 33, Aij=505.4, Aji=81.44501)
        self.set_interaction(27, 42, Aij=2004, Bij=-4.5, Aji=2600, Bji=-4.8)
        self.set_interaction(28, 32, Aij=212.4, Bij=-0.3692, Aji=200.6, Bji=-0.228)
        self.set_interaction(28, 37, Aij=-93.31, Bij=-0.1286, Aji=319.4, Bji=0.03419)
        self.set_interaction(28, 42, Aij=29.45, Bij=-0.9194, Aji=92.4, Bji=1.191)
        self.set_interaction(28, 43, Aij=166.56, Bij=-1.0407, Aji=1.0902, Bji=1.5927)
        self.set_interaction(28, 45, Aij=89.744, Bij=-1.0122, Aji=40.987, Bji=1.1526)
        self.set_interaction(29, 35, Aij=467.1, Aji=-360)
        self.set_interaction(29, 39, Aij=356.6, Aji=-247.6)
        self.set_interaction(29, 42, Aij=-7.465, Bij=-0.4151, Aji=439.73, Bji=-0.2619)
        self.set_interaction(30, 37, Aij=-277.6, Aji=1168)
        self.set_interaction(30, 42, Aij=96.59, Bij=-0.7691, Aji=846.7, Bji=0.2545)
        self.set_interaction(30, 44, Aij=778.78, Aji=-384.29)
        self.set_interaction(31, 35, Aij=-228.4, Aji=-47.81)
        self.set_interaction(31, 39, Aij=373.8, Aji=-231.6)
        self.set_interaction(31, 47, Aij=-210.34, Bij=0.9103, Aji=106.79, Bji=-0.9536999)
        self.set_interaction(32, 33, Aij=-536.2, Bij=1.995, Aji=558, Bji=-1.924)
        self.set_interaction(32, 45, Aij=-47.772, Bij=-0.009201, Aji=92.429, Bji=0.17437)
        self.set_interaction(33, 35, Aij=-83.7, Bij=0.1436, Aji=-116.7, Bji=-0.06775)
        self.set_interaction(33, 41, Aij=-378.1, Bij=2.56, Aji=13.78, Bji=-0.936)
        self.set_interaction(33, 42, Aij=-122.5, Bij=0.0913, Aji=476.9, Bji=-0.61)
        self.set_interaction(33, 43, Aij=-186.4, Aji=265.42)
        self.set_interaction(33, 56, Aij=2459.01, Aji=-271.226)
        self.set_interaction(34, 39, Aij=1025, Aji=-416.5)
        self.set_interaction(35, 39, Aij=-133.35, Bij=0.1022, Aji=221.74, Bji=-0.5021)
        self.set_interaction(35, 45, Aij=-322.46, Bij=-0.020828, Aji=67.069, Bji=0.23965)
        self.set_interaction(35, 49, Aij=-3.44708, Aji=-72.16901)
        self.set_interaction(36, 37, Aij=-74.88, Aji=1004)
        self.set_interaction(36, 39, Aij=-110.34, Aji=97.36)
        self.set_interaction(37, 39, Aij=967.74, Aji=-306.22)
        self.set_interaction(37, 41, Aij=-211.1, Aji=516.5)
        self.set_interaction(37, 42, Aij=-24.82, Bij=-0.5324, Aji=81.56, Bji=0.7293)
        self.set_interaction(37, 43, Aij=321.62, Bij=-4.9963, Cij=0.013871, Aji=713.9, Bji=-2.7759, Cji=0.00091722)
        self.set_interaction(37, 45, Aij=185.82, Aji=-139)
        self.set_interaction(38, 39, Aij=190.06, Aji=-229.97)
        self.set_interaction(38, 40, Aij=-57.38, Aji=110.4)
        self.set_interaction(38, 42, Aij=112.7, Bij=-0.4103, Aji=21.04, Bji=-0.1288)
        self.set_interaction(38, 43, Aij=-22.572, Aji=-7.56)
        self.set_interaction(39, 42, Aij=141.2, Bij=-0.8783, Aji=666.5, Bji=-0.1555)
        self.set_interaction(39, 43, Aij=53.871, Bij=-0.67747, Aji=-54.26, Bji=1.0612)
        self.set_interaction(39, 44, Aij=-310.13, Aji=-367.48)
        self.set_interaction(39, 49, Aij=-18.3768, Bij=-0.450261, Aji=-2.27584, Bji=0.601128)
        self.set_interaction(40, 42, Aij=165.6, Aji=-109)
        self.set_interaction(41, 42, Aij=683.3, Bij=6.321, Aji=865, Bji=-2.89)
        self.set_interaction(42, 43, Aij=242.49, Bij=-0.038323, Aji=20.834, Bji=-0.34718)
        self.set_interaction(42, 45, Aij=183.79, Bij=-0.15182, Aji=-61.922, Bji=-0.059442)
        self.set_interaction(42, 46, Aij=298.46, Bij=-0.6823, Aji=499.59, Bji=-0.8158)
        self.set_interaction(42, 47, Aij=2187.67, Bij=-1.057, Aji=165.66, Bji=-0.9151)
        self.set_interaction(42, 48, Aij=499.19, Bij=0.5131, Aji=148.66, Bji=-1.0328)
        self.set_interaction(42, 49, Aij=-49.6851, Bij=3.86633, Aji=159.179, Bji=-1.11479)
        self.set_interaction(42, 52, Aij=313.43, Bij=0.0911, Aji=-30.564, Bji=-0.2455)
        self.set_interaction(42, 55, Aij=373.49, Bij=0.0647, Aji=75.45, Bji=-0.6572)
        self.set_interaction(42, 56, Aij=-581.16, Bij=4.1194, Aji=660.47, Bji=-2.0132)
        self.set_interaction(42, 61, Aij=509.274, Bij=-0.533745, Aji=-108.172, Bji=-0.132539)
        self.set_interaction(42, 84, Aij=-883.9302, Bij=4.0287, Aji=360.0808, Bji=-1.37211)
        self.set_interaction(42, 85, Aij=493.7747, Bij=-4.10779, Cij=0.0119213, Aji=986.9591, Bji=-2.85003,
                             Cji=0.0018563)
        self.set_interaction(42, 87, Aij=-501.6695, Bij=1.6261, Aji=-935.6061, Bji=4.67403)
        self.set_interaction(42, 89, Aij=-381.9636, Bij=-3.00577, Cij=0.0076136, Aji=2358.513, Bji=-1.30922,
                             Cji=0.0137839)
        self.set_interaction(42, 90, Aij=-1078.495, Bij=3.65154, Aji=1039.753, Bji=-3.51543)
        self.set_interaction(42, 91, Aij=1654.642, Bij=-4.24837, Aji=-171.2349, Bji=0.62135)
        self.set_interaction(42, 93, Aij=246.348, Bij=-1.9006, Aji=1120.89, Bji=-3.1028)
        self.set_interaction(42, 98, Aij=449.5142, Bij=2.1727, Cij=0.0360498, Aji=4258.272, Bji=0.34026, Cji=-0.029654)
        self.set_interaction(43, 45, Aij=-523.96, Bij=0.4945, Aji=1414, Bji=-2.8776)
        self.set_interaction(43, 56, Aij=19.221, Bij=0.4413, Aji=117.208, Bji=-0.3946)
        self.set_interaction(43, 84, Aij=-617.0431, Bij=2.40181, Aji=-323.6833, Bji=-0.13058)
        self.set_interaction(43, 85, Aij=826.0889, Bij=-0.93908, Aji=-450.7675, Bji=0.94786)
        self.set_interaction(43, 89, Aij=580.1045, Bij=-2.24427, Cij=-0.0080098, Aji=4927.307, Bji=-0.63511,
                             Cji=-0.0351738)
        self.set_interaction(43, 90, Aij=277.9974, Bij=0.25973, Aji=122.9861, Bji=-1.96464)
        self.set_interaction(43, 91, Aij=-312.4461, Bij=1.126, Aji=186.7351, Bji=-0.9786)
        self.set_interaction(45, 56, Aij=184.792, Aji=-76.1382)
        self.set_interaction(46, 52, Aij=-1423.7, Bij=3.9162, Aji=1738.5, Bji=-3.9729)
        self.set_interaction(47, 48, Aij=-91.65, Bij=-0.3411, Aji=117.57, Bji=0.7114)
        self.set_interaction(47, 49, Aij=-188.913, Aji=308.235)
        self.set_interaction(53, 56, Aij=912.2161, Bij=-2.02314, Aji=-604.897, Bji=1.04748)
        self.set_interaction(55, 56, Aij=-9.63211, Bij=0.555403, Cij=-6.5e-05, Aji=125.526, Bji=0.289226, Cji=-0.000969)
        self.set_interaction(56, 61, Aij=200.04, Bij=0.1979, Cij=-0.0016, Aji=742.31, Bji=-4.1234, Cji=0.0078)
        self.set_interaction(84, 85, Aij=-103.0519, Bij=-1.42845, Cij=0.0094424, Aji=283.1673, Bji=-5.14586,
                             Cji=0.0208006)
        self.set_interaction(84, 89, Aij=-107.3735, Bij=-3.90856, Cij=0.0005537, Aji=1094.895, Bji=-4.10982,
                             Cji=2.63e-05)
        self.set_interaction(84, 91, Aij=133.3952, Bij=-2.64806, Cij=-0.0025991, Aji=-591.0368, Bji=0.45605,
                             Cji=-0.0035024)
        self.set_interaction(84, 98, Aij=-113.0145, Bij=-1.48378, Aji=-35.932, Bji=-1.52136)
        self.set_interaction(84, 99, Aij=4542.127, Bij=0.85207, Cij=-0.0518203, Aji=-4116.333, Bji=-1.71395,
                             Cji=-0.0073933)
        self.set_interaction(85, 87, Aij=-1403.31, Bij=-4.02583, Cij=0.0382461, Aji=-516.843, Bji=-3.5941,
                             Cji=0.0160959)
        self.set_interaction(85, 90, Aij=855.7144, Bij=-1.14021, Aji=-254.6989, Bji=-0.46519)
        self.set_interaction(87, 91, Aij=-1138.034, Bij=0.70013, Aji=1824.841, Bji=-4.26948)
        self.set_interaction(89, 90, Aij=-1679.977, Bij=4.40837, Cij=-0.0092342, Aji=4209.48, Bji=-0.75036,
                             Cji=-0.0356672)
