#  @file
#  @author Christian Diddens <c.diddens@utwente.nl>
#  @author Duarte Rocha <d.rocha@utwente.nl>
#  
#  @section LICENSE
# 
#  pyoomph - a multi-physics finite element framework based on oomph-lib and GiNaC 
#  Copyright (C) 2021-2025  Christian Diddens & Duarte Rocha
# 
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
# 
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>. 
#
#  The authors may be contacted at c.diddens@utwente.nl and d.rocha@utwente.nl
#
# ========================================================================


##########################################################################################
# These parameter stem from the published data of the original UNIFAC model
#		https://www.ddbst.com/published-parameters-unifac.html
#
# Please cite the relevant publications when publishing results based on the original UNIFAC model
#		see: https://www.ddbst.com/published-parameters-unifac.html
#
# Also note that these are the published parameters. The UNIFAC Consortium offers updated
# parameters, which will lead to more accurate results. You can obtain these parameters
# at:
#
#		https://www.ddbst.com/unifac-consortium.html
#
# and register the updated parameter by following the template below.
#
# These updated parameters, however, are subject to the *terms of use* 
# of the UNIFAC Consortium.
##########################################################################################

from ..activity import UNIFACLikeActivityModel,ActivityModel

@ActivityModel.register_activity_model()
class OriginalUNIFAC(UNIFACLikeActivityModel):
    """
    Original UNIFAC model for activity coefficient calculations.
    
    The published parameters were obtained with kind permisson from the DDBST website (https://www.ddbst.com/published-parameters-unifac.html).
    
    Please cite the relevant publications when publishing results based on this model (https://www.ddbst.com/published-parameters-unifac.html).
    
    Note that these are the published parameters. The UNIFAC Consortium offers updated and revised parameters, which will lead to more accurate results. 
    You can obtain these parameters at: https://www.ddbst.com/unifac-consortium.html 
    
    These updated parameters, however, are subject to the *terms of use* of the UNIFAC Consortium. 
    
    Furthermore, you must implement these revised parameters in the code yourself.
    """    
    name = "Original"
    def __init__(self):
        super(OriginalUNIFAC, self).__init__()
        self.define_groups()
        self.define_interaction_table()

    def define_groups(self):
        with self.define_main_group("CH2", index=1):
            self.define_sub_group("CH3", R=0.9011, Q=0.848, index=1)
            self.define_sub_group("CH2", R=0.6744, Q=0.54, index=2)
            self.define_sub_group("CH", R=0.4469, Q=0.228, index=3)
            self.define_sub_group("C", R=0.2195, Q=0.0, index=4)
        with self.define_main_group("C=C", index=2):
            self.define_sub_group("CH2=CH", R=1.3454, Q=1.176, index=5)
            self.define_sub_group("CH=CH", R=1.1167, Q=0.867, index=6)
            self.define_sub_group("CH2=C", R=1.1173, Q=0.988, index=7)
            self.define_sub_group("CH=C", R=0.8886, Q=0.676, index=8)
            self.define_sub_group("C=C", R=0.6605, Q=0.485, index=70)
        with self.define_main_group("ACH", index=3):
            self.define_sub_group("ACH", R=0.5313, Q=0.4, index=9)
            self.define_sub_group("AC", R=0.3652, Q=0.12, index=10)
        with self.define_main_group("ACCH2", index=4):
            self.define_sub_group("ACCH3", R=1.2663, Q=0.968, index=11)
            self.define_sub_group("ACCH2", R=1.0396, Q=0.66, index=12)
            self.define_sub_group("ACCH", R=0.8121, Q=0.348, index=13)
        with self.define_main_group("OH", index=5):
            self.define_sub_group("OH", R=1.0, Q=1.2, index=14)
        with self.define_main_group("CH3OH", index=6):
            self.define_sub_group("CH3OH", R=1.4311, Q=1.432, index=15)
        with self.define_main_group("H2O", index=7):
            self.define_sub_group("H2O", R=0.92, Q=1.4, index=16)
        with self.define_main_group("ACOH", index=8):
            self.define_sub_group("ACOH", R=0.8952, Q=0.68, index=17)
        with self.define_main_group("CH2CO", index=9):
            self.define_sub_group("CH3CO", R=1.6724, Q=1.488, index=18)
            self.define_sub_group("CH2CO", R=1.4457, Q=1.18, index=19)
        with self.define_main_group("HCO", index=10):
            self.define_sub_group("HCO", R=0.998, Q=0.948, index=20)
        with self.define_main_group("CCOO", index=11):
            self.define_sub_group("CH3COO", R=1.9031, Q=1.728, index=21)
            self.define_sub_group("CH2COO", R=1.6764, Q=1.42, index=22)
        with self.define_main_group("HCOO", index=12):
            self.define_sub_group("HCOO", R=1.242, Q=1.188, index=23)
        with self.define_main_group("CH2O", index=13):
            self.define_sub_group("CH3O", R=1.145, Q=1.088, index=24)
            self.define_sub_group("CH2O", R=0.9183, Q=0.78, index=25)
            self.define_sub_group("CHO", R=0.6908, Q=0.468, index=26)
            self.define_sub_group("THF", R=0.9183, Q=1.1, index=27)
        with self.define_main_group("CNH2", index=14):
            self.define_sub_group("CH3NH2", R=1.5959, Q=1.544, index=28)
            self.define_sub_group("CH2NH2", R=1.3692, Q=1.236, index=29)
            self.define_sub_group("CHNH2", R=1.1417, Q=0.924, index=30)
        with self.define_main_group("CNH", index=15):
            self.define_sub_group("CH3NH", R=1.4337, Q=1.244, index=31)
            self.define_sub_group("CH2NH", R=1.207, Q=0.936, index=32)
            self.define_sub_group("CHNH", R=0.9795, Q=0.624, index=33)
        with self.define_main_group("(C)3N", index=16):
            self.define_sub_group("CH3N", R=1.1865, Q=0.94, index=34)
            self.define_sub_group("CH2N", R=0.9597, Q=0.632, index=35)
        with self.define_main_group("ACNH2", index=17):
            self.define_sub_group("ACNH2", R=1.06, Q=0.816, index=36)
        with self.define_main_group("PYRIDINE", index=18):
            self.define_sub_group("C5H5N", R=2.9993, Q=2.113, index=37)
            self.define_sub_group("C5H4N", R=2.8332, Q=1.833, index=38)
            self.define_sub_group("C5H3N", R=2.667, Q=1.553, index=39)
        with self.define_main_group("CCN", index=19):
            self.define_sub_group("CH3CN", R=1.8701, Q=1.724, index=40)
            self.define_sub_group("CH2CN", R=1.6434, Q=1.416, index=41)
        with self.define_main_group("COOH", index=20):
            self.define_sub_group("COOH", R=1.3013, Q=1.224, index=42)
            self.define_sub_group("HCOOH", R=1.528, Q=1.532, index=43)
        with self.define_main_group("CCL", index=21):
            self.define_sub_group("CH2CL", R=1.4654, Q=1.264, index=44)
            self.define_sub_group("CHCL", R=1.238, Q=0.952, index=45)
            self.define_sub_group("CCL", R=1.0106, Q=0.724, index=46)
        with self.define_main_group("CCL2", index=22):
            self.define_sub_group("CH2CL2", R=2.2564, Q=1.988, index=47)
            self.define_sub_group("CHCL2", R=2.0606, Q=1.684, index=48)
            self.define_sub_group("CCL2", R=1.8016, Q=1.448, index=49)
        with self.define_main_group("CCL3", index=23):
            self.define_sub_group("CHCL3", R=2.87, Q=2.41, index=50)
            self.define_sub_group("CCL3", R=2.6401, Q=2.184, index=51)
        with self.define_main_group("CCL4", index=24):
            self.define_sub_group("CCL4", R=3.39, Q=2.91, index=52)
        with self.define_main_group("ACCL", index=25):
            self.define_sub_group("ACCL", R=1.1562, Q=0.844, index=53)
        with self.define_main_group("CNO2", index=26):
            self.define_sub_group("CH3NO2", R=2.0086, Q=1.868, index=54)
            self.define_sub_group("CH2NO2", R=1.7818, Q=1.56, index=55)
            self.define_sub_group("CHNO2", R=1.5544, Q=1.248, index=56)
        with self.define_main_group("ACNO2", index=27):
            self.define_sub_group("ACNO2", R=1.4199, Q=1.104, index=57)
        with self.define_main_group("CS2", index=28):
            self.define_sub_group("CS2", R=2.057, Q=1.65, index=58)
        with self.define_main_group("CH3SH", index=29):
            self.define_sub_group("CH3SH", R=1.877, Q=1.676, index=59)
            self.define_sub_group("CH2SH", R=1.651, Q=1.368, index=60)
        with self.define_main_group("FURFURAL", index=30):
            self.define_sub_group("FURFURAL", R=3.168, Q=2.484, index=61)
        with self.define_main_group("DOH", index=31):
            self.define_sub_group("DOH", R=2.4088, Q=2.248, index=62)
        with self.define_main_group("I", index=32):
            self.define_sub_group("I", R=1.264, Q=0.992, index=63)
        with self.define_main_group("BR", index=33):
            self.define_sub_group("BR", R=0.9492, Q=0.832, index=64)
        with self.define_main_group("C=-C", index=34):
            self.define_sub_group("CH=-C", R=1.292, Q=1.088, index=65)
            self.define_sub_group("C=-C", R=1.0613, Q=0.784, index=66)
        with self.define_main_group("DMSO", index=35):
            self.define_sub_group("DMSO", R=2.8266, Q=2.472, index=67)
        with self.define_main_group("ACRY", index=36):
            self.define_sub_group("ACRY", R=2.3144, Q=2.052, index=68)
        with self.define_main_group("CLCC", index=37):
            self.define_sub_group("CL-(C=C)", R=0.791, Q=0.724, index=69)
        with self.define_main_group("ACF", index=38):
            self.define_sub_group("ACF", R=0.6948, Q=0.524, index=71)
        with self.define_main_group("DMF", index=39):
            self.define_sub_group("DMF", R=3.0856, Q=2.736, index=72)
            self.define_sub_group("HCON(..", R=2.6322, Q=2.12, index=73)
        with self.define_main_group("CF2", index=40):
            self.define_sub_group("CF3", R=1.406, Q=1.38, index=74)
            self.define_sub_group("CF2", R=1.0105, Q=0.92, index=75)
            self.define_sub_group("CF", R=0.615, Q=0.46, index=76)
        with self.define_main_group("COO", index=41):
            self.define_sub_group("COO", R=1.38, Q=1.2, index=77)
        with self.define_main_group("SIH2", index=42):
            self.define_sub_group("SIH3", R=1.6035, Q=1.2632, index=78)
            self.define_sub_group("SIH2", R=1.4443, Q=1.0063, index=79)
            self.define_sub_group("SIH", R=1.2853, Q=0.7494, index=80)
            self.define_sub_group("SI", R=1.047, Q=0.4099, index=81)
        with self.define_main_group("SIO", index=43):
            self.define_sub_group("SIH2O", R=1.4838, Q=1.0621, index=82)
            self.define_sub_group("SIHO", R=1.303, Q=0.7639, index=83)
            self.define_sub_group("SIO", R=1.1044, Q=0.4657, index=84)
        with self.define_main_group("NMP", index=44):
            self.define_sub_group("NMP", R=3.981, Q=3.2, index=85)
        with self.define_main_group("CCLF", index=45):
            self.define_sub_group("CCL3F", R=3.0356, Q=2.644, index=86)
            self.define_sub_group("CCL2F", R=2.2287, Q=1.916, index=87)
            self.define_sub_group("HCCL2F", R=2.406, Q=2.116, index=88)
            self.define_sub_group("HCCLF", R=1.6493, Q=1.416, index=89)
            self.define_sub_group("CCLF2", R=1.8174, Q=1.648, index=90)
            self.define_sub_group("HCCLF2", R=1.967, Q=1.828, index=91)
            self.define_sub_group("CCLF3", R=2.1721, Q=2.1, index=92)
            self.define_sub_group("CCL2F2", R=2.6243, Q=2.376, index=93)
        with self.define_main_group("CON(AM)", index=46):
            self.define_sub_group("AMH2", R=1.4515, Q=1.248, index=94)
            self.define_sub_group("AMHCH3", R=2.1905, Q=1.796, index=95)
            self.define_sub_group("AMHCH2", R=1.9637, Q=1.488, index=96)
            self.define_sub_group("AM(CH3)2", R=2.8589, Q=2.428, index=97)
            self.define_sub_group("AMCH3CH2", R=2.6322, Q=2.12, index=98)
            self.define_sub_group("AM(CH2)2", R=2.4054, Q=1.812, index=99)
        with self.define_main_group("OCCOH", index=47):
            self.define_sub_group("C2H5O2", R=2.1226, Q=1.904, index=100)
            self.define_sub_group("C2H4O2", R=1.8952, Q=1.592, index=101)
        with self.define_main_group("CH2S", index=48):
            self.define_sub_group("CH3S", R=1.613, Q=1.368, index=102)
            self.define_sub_group("CH2S", R=1.3863, Q=1.06, index=103)
            self.define_sub_group("CHS", R=1.1589, Q=0.748, index=104)
        with self.define_main_group("MORPH", index=49):
            self.define_sub_group("MORPH", R=3.474, Q=2.796, index=105)
        with self.define_main_group("THIOPHEN", index=50):
            self.define_sub_group("C4H4S", R=2.8569, Q=2.14, index=106)
            self.define_sub_group("C4H3S", R=2.6908, Q=1.86, index=107)
            self.define_sub_group("C4H2S", R=2.5247, Q=1.58, index=108)
        with self.define_main_group("NCO", index=51):
            self.define_sub_group("NCO", R=1.0567, Q=0.732, index=109)
        with self.define_main_group("SULFONES", index=55):
            self.define_sub_group("(CH2)2SU", R=2.6869, Q=2.12, index=118)
            self.define_sub_group("CH2CHSU", R=2.4595, Q=1.808, index=119)
        with self.define_main_group("IMIDAZOL", index=84):
            self.define_sub_group("IMIDAZOL", R=2.026, Q=0.868, index=178)
        with self.define_main_group("BTI", index=85):
            self.define_sub_group("BTI", R=5.774, Q=4.932, index=179)

    def define_interaction_table(self):
        self.set_interaction(1, 2, Aij=86.0200, Aji=-35.3600)
        self.set_interaction(1, 3, Aij=61.1300, Aji=-11.1200)
        self.set_interaction(1, 4, Aij=76.5000, Aji=-69.7000)
        self.set_interaction(1, 5, Aij=986.5000, Aji=156.4000)
        self.set_interaction(1, 6, Aij=697.2000, Aji=16.5100)
        self.set_interaction(1, 7, Aij=1318.0000, Aji=300.0000)
        self.set_interaction(1, 8, Aij=1333.0000, Aji=275.8000)
        self.set_interaction(1, 9, Aij=476.4000, Aji=26.7600)
        self.set_interaction(1, 10, Aij=677.0000, Aji=505.7000)
        self.set_interaction(1, 11, Aij=232.1000, Aji=114.8000)
        self.set_interaction(1, 12, Aij=507.0000, Aji=329.3000)
        self.set_interaction(1, 13, Aij=251.5000, Aji=83.3600)
        self.set_interaction(1, 14, Aij=391.5000, Aji=-30.4800)
        self.set_interaction(1, 15, Aij=255.7000, Aji=65.3300)
        self.set_interaction(1, 16, Aij=206.6000, Aji=-83.9800)
        self.set_interaction(1, 17, Aij=920.7000, Aji=1139.0000)
        self.set_interaction(1, 18, Aij=287.7700, Aji=-101.5600)
        self.set_interaction(1, 19, Aij=597.0000, Aji=24.8200)
        self.set_interaction(1, 20, Aij=663.5000, Aji=315.3000)
        self.set_interaction(1, 21, Aij=35.9300, Aji=91.4600)
        self.set_interaction(1, 22, Aij=53.7600, Aji=34.0100)
        self.set_interaction(1, 23, Aij=24.9000, Aji=36.7000)
        self.set_interaction(1, 24, Aij=104.3000, Aji=-78.4500)
        self.set_interaction(1, 25, Aij=11.4400, Aji=106.8000)
        self.set_interaction(1, 26, Aij=661.5000, Aji=-32.6900)
        self.set_interaction(1, 27, Aij=543.0000, Aji=5541.0000)
        self.set_interaction(1, 28, Aij=153.6000, Aji=-52.6500)
        self.set_interaction(1, 29, Aij=184.4000, Aji=-7.4810)
        self.set_interaction(1, 30, Aij=354.5500, Aji=-25.3100)
        self.set_interaction(1, 31, Aij=3025.0000, Aji=139.9300)
        self.set_interaction(1, 32, Aij=335.8000, Aji=128.0000)
        self.set_interaction(1, 33, Aij=479.5000, Aji=-31.5200)
        self.set_interaction(1, 34, Aij=298.9000, Aji=-72.8800)
        self.set_interaction(1, 35, Aij=526.5000, Aji=50.4900)
        self.set_interaction(1, 36, Aij=689.0000, Aji=-165.9000)
        self.set_interaction(1, 37, Aij=-4.1890, Aji=47.4100)
        self.set_interaction(1, 38, Aij=125.8000, Aji=-5.1320)
        self.set_interaction(1, 39, Aij=485.3000, Aji=-31.9500)
        self.set_interaction(1, 40, Aij=-2.8590, Aji=147.3000)
        self.set_interaction(1, 41, Aij=387.1000, Aji=529.0000)
        self.set_interaction(1, 42, Aij=-450.4000, Aji=-34.3600)
        self.set_interaction(1, 43, Aij=252.7000, Aji=110.2000)
        self.set_interaction(1, 44, Aij=220.3000, Aji=13.8900)
        self.set_interaction(1, 45, Aij=-5.8690, Aji=30.7400)
        self.set_interaction(1, 46, Aij=390.9000, Aji=27.9700)
        self.set_interaction(1, 47, Aij=553.3000, Aji=-11.9200)
        self.set_interaction(1, 48, Aij=187.0000, Aji=39.9300)
        self.set_interaction(1, 49, Aij=216.1000, Aji=-23.6100)
        self.set_interaction(1, 50, Aij=92.9900, Aji=-8.4790)
        self.set_interaction(1, 51, Aij=699.1300, Aji=456.1900)
        self.set_interaction(1, 55, Aij=808.5900, Aji=245.2100)
        self.set_interaction(1, 84, Aij=-1243.0000, Aji=125.3600)
        self.set_interaction(1, 85, Aij=637.6500, Aji=221.5600)
        self.set_interaction(2, 3, Aij=38.8100, Aji=3.4460)
        self.set_interaction(2, 4, Aij=74.1500, Aji=-113.6000)
        self.set_interaction(2, 5, Aij=524.1000, Aji=457.0000)
        self.set_interaction(2, 6, Aij=787.6000, Aji=-12.5200)
        self.set_interaction(2, 7, Aij=270.6000, Aji=496.1000)
        self.set_interaction(2, 8, Aij=526.1000, Aji=217.5000)
        self.set_interaction(2, 9, Aij=182.6000, Aji=42.9200)
        self.set_interaction(2, 10, Aij=448.7500, Aji=56.3000)
        self.set_interaction(2, 11, Aij=37.8500, Aji=132.1000)
        self.set_interaction(2, 12, Aij=333.5000, Aji=110.4000)
        self.set_interaction(2, 13, Aij=214.5000, Aji=26.5100)
        self.set_interaction(2, 14, Aij=240.9000, Aji=1.1630)
        self.set_interaction(2, 15, Aij=163.9000, Aji=-28.7000)
        self.set_interaction(2, 16, Aij=61.1100, Aji=-25.3800)
        self.set_interaction(2, 17, Aij=749.3000, Aji=2000.0000)
        self.set_interaction(2, 18, Aij=280.5000, Aji=-47.6300)
        self.set_interaction(2, 19, Aij=336.9000, Aji=-40.6200)
        self.set_interaction(2, 20, Aij=318.9000, Aji=1264.0000)
        self.set_interaction(2, 21, Aij=-36.8700, Aji=40.2500)
        self.set_interaction(2, 22, Aij=58.5500, Aji=-23.5000)
        self.set_interaction(2, 23, Aij=-13.9900, Aji=51.0600)
        self.set_interaction(2, 24, Aij=-109.7000, Aji=160.9000)
        self.set_interaction(2, 25, Aij=100.1000, Aji=70.3200)
        self.set_interaction(2, 26, Aij=357.5000, Aji=-1.9960)
        self.set_interaction(2, 28, Aij=76.3000, Aji=16.6230)
        self.set_interaction(2, 30, Aij=262.9000, Aji=82.6400)
        self.set_interaction(2, 33, Aij=183.8000, Aji=174.6000)
        self.set_interaction(2, 34, Aij=31.1400, Aji=41.3800)
        self.set_interaction(2, 35, Aij=179.0000, Aji=64.0700)
        self.set_interaction(2, 36, Aij=-52.8700, Aji=573.0000)
        self.set_interaction(2, 37, Aij=-66.4600, Aji=124.2000)
        self.set_interaction(2, 38, Aij=359.3000, Aji=-131.7000)
        self.set_interaction(2, 39, Aij=-70.4500, Aji=249.0000)
        self.set_interaction(2, 40, Aij=449.4000, Aji=62.4000)
        self.set_interaction(2, 41, Aij=48.3300, Aji=1397.0000)
        self.set_interaction(2, 44, Aij=86.4600, Aji=-16.1100)
        self.set_interaction(2, 46, Aij=200.2000, Aji=9.7550)
        self.set_interaction(2, 47, Aij=268.1000, Aji=132.4000)
        self.set_interaction(2, 48, Aij=-617.0000, Aji=543.6000)
        self.set_interaction(2, 49, Aij=62.5600, Aji=161.1000)
        self.set_interaction(2, 55, Aij=200.9400, Aji=384.4500)
        self.set_interaction(2, 84, Aij=-861.4600, Aji=-391.8100)
        self.set_interaction(2, 85, Aij=424.9300, Aji=629.9600)
        self.set_interaction(3, 4, Aij=167.0000, Aji=-146.8000)
        self.set_interaction(3, 5, Aij=636.1000, Aji=89.6000)
        self.set_interaction(3, 6, Aij=637.3500, Aji=-50.0000)
        self.set_interaction(3, 7, Aij=903.8000, Aji=362.3000)
        self.set_interaction(3, 8, Aij=1329.0000, Aji=25.3400)
        self.set_interaction(3, 9, Aij=25.7700, Aji=140.1000)
        self.set_interaction(3, 10, Aij=347.3000, Aji=23.3900)
        self.set_interaction(3, 11, Aij=5.9940, Aji=85.8400)
        self.set_interaction(3, 12, Aij=287.1000, Aji=18.1200)
        self.set_interaction(3, 13, Aij=32.1400, Aji=52.1300)
        self.set_interaction(3, 14, Aij=161.7000, Aji=-44.8500)
        self.set_interaction(3, 15, Aij=122.8000, Aji=-22.3100)
        self.set_interaction(3, 16, Aij=90.4900, Aji=-223.9000)
        self.set_interaction(3, 17, Aij=648.2000, Aji=247.5000)
        self.set_interaction(3, 18, Aij=-4.4490, Aji=31.8700)
        self.set_interaction(3, 19, Aij=212.5000, Aji=-22.9700)
        self.set_interaction(3, 20, Aij=537.4000, Aji=62.3200)
        self.set_interaction(3, 21, Aij=-18.8100, Aji=4.6800)
        self.set_interaction(3, 22, Aij=-144.4000, Aji=121.3000)
        self.set_interaction(3, 23, Aij=-231.9000, Aji=288.5000)
        self.set_interaction(3, 24, Aij=3.0000, Aji=-4.7000)
        self.set_interaction(3, 25, Aij=187.0000, Aji=-97.2700)
        self.set_interaction(3, 26, Aij=168.0400, Aji=10.3800)
        self.set_interaction(3, 27, Aij=194.9000, Aji=1824.0000)
        self.set_interaction(3, 28, Aij=52.0680, Aji=21.4970)
        self.set_interaction(3, 29, Aij=-10.4300, Aji=28.4100)
        self.set_interaction(3, 30, Aij=-64.6900, Aji=157.2900)
        self.set_interaction(3, 31, Aij=210.3660, Aji=221.4000)
        self.set_interaction(3, 32, Aij=113.3000, Aji=58.6800)
        self.set_interaction(3, 33, Aij=261.3000, Aji=-154.2000)
        self.set_interaction(3, 34, Aij=154.2600, Aji=-101.1200)
        self.set_interaction(3, 35, Aij=169.9000, Aji=-2.5040)
        self.set_interaction(3, 36, Aij=383.9000, Aji=-123.6000)
        self.set_interaction(3, 37, Aij=-259.1000, Aji=395.8000)
        self.set_interaction(3, 38, Aij=389.3000, Aji=-237.2000)
        self.set_interaction(3, 39, Aij=245.6000, Aji=-133.9000)
        self.set_interaction(3, 40, Aij=22.6700, Aji=140.6000)
        self.set_interaction(3, 41, Aij=103.5000, Aji=317.6000)
        self.set_interaction(3, 42, Aij=-432.3000, Aji=787.9000)
        self.set_interaction(3, 43, Aij=238.9000, Aji=234.4000)
        self.set_interaction(3, 44, Aij=30.0400, Aji=-23.8800)
        self.set_interaction(3, 45, Aij=-88.1100, Aji=167.9000)
        self.set_interaction(3, 47, Aij=333.3000, Aji=-86.8800)
        self.set_interaction(3, 49, Aij=-59.5800, Aji=142.9000)
        self.set_interaction(3, 50, Aij=-39.1600, Aji=23.9300)
        self.set_interaction(3, 55, Aij=360.8200, Aji=47.0500)
        self.set_interaction(4, 5, Aij=803.2000, Aji=25.8200)
        self.set_interaction(4, 6, Aij=603.2500, Aji=-44.5000)
        self.set_interaction(4, 7, Aij=5695.0000, Aji=377.6000)
        self.set_interaction(4, 8, Aij=884.9000, Aji=244.2000)
        self.set_interaction(4, 9, Aij=-52.1000, Aji=365.8000)
        self.set_interaction(4, 10, Aij=586.8000, Aji=106.0000)
        self.set_interaction(4, 11, Aij=5688.0000, Aji=-170.0000)
        self.set_interaction(4, 12, Aij=197.8000, Aji=428.0000)
        self.set_interaction(4, 13, Aij=213.1000, Aji=65.6900)
        self.set_interaction(4, 14, Aij=19.0200, Aji=296.4000)
        self.set_interaction(4, 15, Aij=-49.2900, Aji=223.0000)
        self.set_interaction(4, 16, Aij=23.5000, Aji=109.9000)
        self.set_interaction(4, 17, Aij=664.2000, Aji=762.8000)
        self.set_interaction(4, 18, Aij=52.8000, Aji=49.8000)
        self.set_interaction(4, 19, Aij=6096.0000, Aji=-138.4000)
        self.set_interaction(4, 20, Aij=872.3000, Aji=89.8600)
        self.set_interaction(4, 21, Aij=-114.1400, Aji=122.9100)
        self.set_interaction(4, 22, Aij=-111.0000, Aji=140.7800)
        self.set_interaction(4, 23, Aij=-80.2500, Aji=69.9000)
        self.set_interaction(4, 24, Aij=-141.3000, Aji=134.7000)
        self.set_interaction(4, 25, Aij=-211.0000, Aji=402.5000)
        self.set_interaction(4, 26, Aij=3629.0000, Aji=-97.0500)
        self.set_interaction(4, 27, Aij=4448.0000, Aji=-127.8000)
        self.set_interaction(4, 28, Aij=-9.4510, Aji=40.6750)
        self.set_interaction(4, 29, Aij=393.6000, Aji=19.5600)
        self.set_interaction(4, 30, Aij=48.4900, Aji=128.8000)
        self.set_interaction(4, 31, Aij=4975.0000, Aji=150.6400)
        self.set_interaction(4, 32, Aij=259.0000, Aji=26.4100)
        self.set_interaction(4, 33, Aij=210.0000, Aji=1112.0000)
        self.set_interaction(4, 34, Aij=-152.5500, Aji=614.5200)
        self.set_interaction(4, 35, Aij=4284.0000, Aji=-143.2000)
        self.set_interaction(4, 36, Aij=-119.2000, Aji=397.4000)
        self.set_interaction(4, 37, Aij=-282.5000, Aji=419.1000)
        self.set_interaction(4, 38, Aij=101.4000, Aji=-157.3000)
        self.set_interaction(4, 39, Aij=5629.0000, Aji=-240.2000)
        self.set_interaction(4, 40, Aij=-245.3900, Aji=839.8300)
        self.set_interaction(4, 41, Aij=69.2600, Aji=615.8000)
        self.set_interaction(4, 44, Aij=46.3800, Aji=6.2140)
        self.set_interaction(4, 47, Aij=421.9000, Aji=-19.4500)
        self.set_interaction(4, 49, Aij=-203.6000, Aji=274.1000)
        self.set_interaction(4, 50, Aij=184.9000, Aji=2.8450)
        self.set_interaction(4, 55, Aij=233.5100, Aji=347.1300)
        self.set_interaction(5, 6, Aij=-137.1000, Aji=249.1000)
        self.set_interaction(5, 7, Aij=353.5000, Aji=-229.1000)
        self.set_interaction(5, 8, Aij=-259.7000, Aji=-451.6000)
        self.set_interaction(5, 9, Aij=84.0000, Aji=164.5000)
        self.set_interaction(5, 10, Aij=-203.6000, Aji=529.0000)
        self.set_interaction(5, 11, Aij=101.1000, Aji=245.4000)
        self.set_interaction(5, 12, Aij=267.8000, Aji=139.4000)
        self.set_interaction(5, 13, Aij=28.0600, Aji=237.7000)
        self.set_interaction(5, 14, Aij=8.6420, Aji=-242.8000)
        self.set_interaction(5, 15, Aij=42.7000, Aji=-150.0000)
        self.set_interaction(5, 16, Aij=-323.0000, Aji=28.6000)
        self.set_interaction(5, 17, Aij=-52.3900, Aji=-17.4000)
        self.set_interaction(5, 18, Aij=170.0290, Aji=-132.3000)
        self.set_interaction(5, 19, Aij=6.7120, Aji=185.4000)
        self.set_interaction(5, 20, Aij=199.0000, Aji=-151.0000)
        self.set_interaction(5, 21, Aij=75.6200, Aji=562.2000)
        self.set_interaction(5, 22, Aij=65.2800, Aji=527.6000)
        self.set_interaction(5, 23, Aij=-98.1200, Aji=742.1000)
        self.set_interaction(5, 24, Aij=143.1000, Aji=856.3000)
        self.set_interaction(5, 25, Aij=123.5000, Aji=325.7000)
        self.set_interaction(5, 26, Aij=256.5000, Aji=261.6000)
        self.set_interaction(5, 27, Aij=157.1000, Aji=561.6000)
        self.set_interaction(5, 28, Aij=488.9000, Aji=609.8000)
        self.set_interaction(5, 29, Aij=147.5000, Aji=461.6000)
        self.set_interaction(5, 30, Aij=-120.4600, Aji=521.6300)
        self.set_interaction(5, 31, Aij=-318.9300, Aji=267.6000)
        self.set_interaction(5, 32, Aij=313.5000, Aji=501.3000)
        self.set_interaction(5, 33, Aij=202.1000, Aji=524.9000)
        self.set_interaction(5, 34, Aij=727.8000, Aji=68.9500)
        self.set_interaction(5, 35, Aij=-202.1000, Aji=-25.8700)
        self.set_interaction(5, 36, Aij=74.2700, Aji=389.3000)
        self.set_interaction(5, 37, Aij=225.8000, Aji=738.9000)
        self.set_interaction(5, 38, Aij=44.7800, Aji=649.7000)
        self.set_interaction(5, 39, Aij=-143.9000, Aji=64.1600)
        self.set_interaction(5, 41, Aij=190.3000, Aji=88.6300)
        self.set_interaction(5, 42, Aij=-817.7000, Aji=1913.0000)
        self.set_interaction(5, 43, Aij=-1712.8000, Aji=430.0600)
        self.set_interaction(5, 44, Aij=-504.2000, Aji=796.9000)
        self.set_interaction(5, 45, Aij=72.9600, Aji=794.4000)
        self.set_interaction(5, 46, Aij=-382.7000, Aji=394.8000)
        self.set_interaction(5, 47, Aij=-248.3000, Aji=517.5000)
        self.set_interaction(5, 49, Aij=104.7000, Aji=-61.2000)
        self.set_interaction(5, 50, Aij=57.6500, Aji=682.5000)
        self.set_interaction(5, 55, Aij=215.8100, Aji=72.1900)
        self.set_interaction(5, 84, Aij=-1840.8000, Aji=111.6500)
        self.set_interaction(5, 85, Aij=56.2980, Aji=122.1900)
        self.set_interaction(6, 7, Aij=-180.9500, Aji=289.6000)
        self.set_interaction(6, 8, Aij=-101.7000, Aji=-265.2000)
        self.set_interaction(6, 9, Aij=23.3900, Aji=108.6500)
        self.set_interaction(6, 10, Aij=306.4200, Aji=-340.1800)
        self.set_interaction(6, 11, Aij=-10.7200, Aji=249.6300)
        self.set_interaction(6, 12, Aij=179.7000, Aji=227.8000)
        self.set_interaction(6, 13, Aij=-128.6000, Aji=238.4000)
        self.set_interaction(6, 14, Aij=359.3000, Aji=-481.6500)
        self.set_interaction(6, 15, Aij=-20.9800, Aji=-370.3000)
        self.set_interaction(6, 16, Aij=53.9000, Aji=-406.8000)
        self.set_interaction(6, 17, Aij=489.7000, Aji=-118.1000)
        self.set_interaction(6, 18, Aij=580.4800, Aji=-378.2400)
        self.set_interaction(6, 19, Aij=53.2800, Aji=162.6000)
        self.set_interaction(6, 20, Aij=-202.0000, Aji=339.8000)
        self.set_interaction(6, 21, Aij=-38.3200, Aji=529.0000)
        self.set_interaction(6, 22, Aij=-102.5400, Aji=669.9000)
        self.set_interaction(6, 23, Aij=-139.3500, Aji=649.1000)
        self.set_interaction(6, 24, Aij=-44.7600, Aji=709.6000)
        self.set_interaction(6, 25, Aij=-28.2500, Aji=612.8000)
        self.set_interaction(6, 26, Aij=75.1400, Aji=252.5600)
        self.set_interaction(6, 27, Aij=457.8800, Aji=511.2900)
        self.set_interaction(6, 28, Aij=-31.0900, Aji=914.2000)
        self.set_interaction(6, 29, Aij=17.5000, Aji=448.6000)
        self.set_interaction(6, 30, Aij=-61.7600, Aji=287.0000)
        self.set_interaction(6, 31, Aij=-119.2000, Aji=240.8000)
        self.set_interaction(6, 32, Aij=212.1000, Aji=431.3000)
        self.set_interaction(6, 33, Aij=106.3000, Aji=494.7000)
        self.set_interaction(6, 34, Aij=-119.1000, Aji=967.7100)
        self.set_interaction(6, 35, Aij=-399.3000, Aji=695.0000)
        self.set_interaction(6, 36, Aij=-5.2240, Aji=218.8000)
        self.set_interaction(6, 37, Aij=33.4700, Aji=528.0000)
        self.set_interaction(6, 38, Aij=-48.2500, Aji=645.9000)
        self.set_interaction(6, 39, Aij=-172.4000, Aji=172.2000)
        self.set_interaction(6, 41, Aij=165.7000, Aji=171.0000)
        self.set_interaction(6, 45, Aij=-52.1000, Aji=762.7000)
        self.set_interaction(6, 48, Aij=37.6300, Aji=420.0000)
        self.set_interaction(6, 49, Aij=-59.4000, Aji=-89.2400)
        self.set_interaction(6, 50, Aij=-46.0100, Aji=597.8000)
        self.set_interaction(6, 55, Aij=150.0200, Aji=265.7500)
        self.set_interaction(7, 8, Aij=324.5000, Aji=-601.8000)
        self.set_interaction(7, 9, Aij=-195.4000, Aji=472.5000)
        self.set_interaction(7, 10, Aij=-116.0000, Aji=480.8000)
        self.set_interaction(7, 11, Aij=72.8700, Aji=200.8000)
        self.set_interaction(7, 12, Aij=233.8700, Aji=124.6300)
        self.set_interaction(7, 13, Aij=540.5000, Aji=-314.7000)
        self.set_interaction(7, 14, Aij=48.8900, Aji=-330.4000)
        self.set_interaction(7, 15, Aij=168.0000, Aji=-448.2000)
        self.set_interaction(7, 16, Aij=304.0000, Aji=-598.8000)
        self.set_interaction(7, 17, Aij=243.2000, Aji=-341.6000)
        self.set_interaction(7, 18, Aij=459.0000, Aji=-332.9000)
        self.set_interaction(7, 19, Aij=112.6000, Aji=242.8000)
        self.set_interaction(7, 20, Aij=-14.0900, Aji=-66.1700)
        self.set_interaction(7, 21, Aij=325.4400, Aji=698.2400)
        self.set_interaction(7, 22, Aij=370.4000, Aji=708.6900)
        self.set_interaction(7, 23, Aij=353.6800, Aji=826.7600)
        self.set_interaction(7, 24, Aij=497.5400, Aji=1201.0000)
        self.set_interaction(7, 25, Aij=133.9000, Aji=-274.5000)
        self.set_interaction(7, 26, Aij=220.6000, Aji=417.9000)
        self.set_interaction(7, 27, Aij=399.5000, Aji=360.7000)
        self.set_interaction(7, 28, Aij=887.1000, Aji=1081.0000)
        self.set_interaction(7, 30, Aij=188.0260, Aji=23.4840)
        self.set_interaction(7, 31, Aij=12.7200, Aji=-137.4000)
        self.set_interaction(7, 33, Aij=777.1000, Aji=79.1800)
        self.set_interaction(7, 35, Aij=-139.0000, Aji=-240.0000)
        self.set_interaction(7, 36, Aij=160.8000, Aji=386.6000)
        self.set_interaction(7, 39, Aij=319.0000, Aji=-287.1000)
        self.set_interaction(7, 41, Aij=-197.5000, Aji=284.4000)
        self.set_interaction(7, 42, Aij=-363.8000, Aji=180.2000)
        self.set_interaction(7, 44, Aij=-452.2000, Aji=832.2000)
        self.set_interaction(7, 46, Aij=835.6000, Aji=-509.3000)
        self.set_interaction(7, 47, Aij=139.6000, Aji=-205.7000)
        self.set_interaction(7, 49, Aij=407.9000, Aji=-384.3000)
        self.set_interaction(7, 55, Aij=-255.6300, Aji=627.3900)
        self.set_interaction(8, 9, Aij=-356.1000, Aji=-133.1000)
        self.set_interaction(8, 10, Aij=-271.1000, Aji=-155.6000)
        self.set_interaction(8, 11, Aij=-449.4000, Aji=-36.7200)
        self.set_interaction(8, 12, Aij=-32.5200, Aji=-234.2500)
        self.set_interaction(8, 13, Aij=-162.8742, Aji=-178.5461)
        self.set_interaction(8, 14, Aij=-832.9700, Aji=-870.8000)
        self.set_interaction(8, 17, Aij=119.9000, Aji=-253.1000)
        self.set_interaction(8, 18, Aij=-305.5000, Aji=-341.6000)
        self.set_interaction(8, 20, Aij=408.9000, Aji=-11.0000)
        self.set_interaction(8, 22, Aij=517.2700, Aji=1633.5000)
        self.set_interaction(8, 24, Aij=1827.0000, Aji=10000.0000)
        self.set_interaction(8, 25, Aij=6915.0000, Aji=622.3000)
        self.set_interaction(8, 27, Aij=-413.4800, Aji=815.1200)
        self.set_interaction(8, 28, Aij=8483.5000, Aji=1421.3000)
        self.set_interaction(8, 31, Aij=-687.1000, Aji=838.4000)
        self.set_interaction(8, 41, Aij=-494.2000, Aji=-167.3000)
        self.set_interaction(8, 44, Aij=-659.0000, Aji=-234.7000)
        self.set_interaction(8, 50, Aij=1005.0000, Aji=810.5000)
        self.set_interaction(9, 10, Aij=-37.3600, Aji=128.0000)
        self.set_interaction(9, 11, Aij=-213.7000, Aji=372.2000)
        self.set_interaction(9, 12, Aij=-190.4000, Aji=385.4000)
        self.set_interaction(9, 13, Aij=-103.6000, Aji=191.1000)
        self.set_interaction(9, 15, Aij=-174.2000, Aji=394.6000)
        self.set_interaction(9, 16, Aij=-169.0000, Aji=225.3000)
        self.set_interaction(9, 17, Aij=6201.0000, Aji=-450.3000)
        self.set_interaction(9, 18, Aij=7.3410, Aji=29.1000)
        self.set_interaction(9, 19, Aij=481.7000, Aji=-287.5000)
        self.set_interaction(9, 20, Aij=669.4000, Aji=-297.8000)
        self.set_interaction(9, 21, Aij=-191.6900, Aji=286.2800)
        self.set_interaction(9, 22, Aij=-130.3000, Aji=82.8600)
        self.set_interaction(9, 23, Aij=-354.5500, Aji=552.1000)
        self.set_interaction(9, 24, Aij=-39.2000, Aji=372.0000)
        self.set_interaction(9, 25, Aij=-119.8000, Aji=518.4000)
        self.set_interaction(9, 26, Aij=137.5000, Aji=-142.6100)
        self.set_interaction(9, 27, Aij=548.5000, Aji=-101.5000)
        self.set_interaction(9, 28, Aij=216.1380, Aji=303.6570)
        self.set_interaction(9, 29, Aij=-46.2800, Aji=160.6000)
        self.set_interaction(9, 30, Aij=-163.7000, Aji=317.5000)
        self.set_interaction(9, 31, Aij=71.4600, Aji=135.4000)
        self.set_interaction(9, 32, Aij=53.5900, Aji=138.0000)
        self.set_interaction(9, 33, Aij=245.2000, Aji=-142.6000)
        self.set_interaction(9, 34, Aij=-246.6000, Aji=443.6150)
        self.set_interaction(9, 35, Aij=-44.5800, Aji=110.4000)
        self.set_interaction(9, 36, Aij=-63.5000, Aji=114.5500)
        self.set_interaction(9, 37, Aij=-34.5700, Aji=-40.9000)
        self.set_interaction(9, 39, Aij=-61.7000, Aji=97.0400)
        self.set_interaction(9, 41, Aij=-18.8000, Aji=123.4000)
        self.set_interaction(9, 42, Aij=-588.9000, Aji=992.4000)
        self.set_interaction(9, 47, Aij=37.5400, Aji=156.4000)
        self.set_interaction(9, 50, Aij=-162.6000, Aji=278.8000)
        self.set_interaction(10, 11, Aij=-110.3000, Aji=185.1000)
        self.set_interaction(10, 12, Aij=766.0000, Aji=-236.5000)
        self.set_interaction(10, 13, Aij=304.1000, Aji=-7.8380)
        self.set_interaction(10, 19, Aij=-106.4000, Aji=224.6600)
        self.set_interaction(10, 20, Aij=497.5000, Aji=-165.5000)
        self.set_interaction(10, 21, Aij=751.9000, Aji=-47.5100)
        self.set_interaction(10, 22, Aij=67.5200, Aji=190.6000)
        self.set_interaction(10, 23, Aij=-483.7000, Aji=242.8000)
        self.set_interaction(10, 32, Aij=117.0000, Aji=245.9000)
        self.set_interaction(10, 34, Aij=2.2100, Aji=-55.8700)
        self.set_interaction(10, 36, Aij=-339.2000, Aji=354.0000)
        self.set_interaction(10, 37, Aij=172.4000, Aji=183.8000)
        self.set_interaction(10, 39, Aij=-268.8000, Aji=13.8900)
        self.set_interaction(10, 41, Aij=-275.5000, Aji=577.5000)
        self.set_interaction(11, 12, Aij=-241.8000, Aji=1167.0000)
        self.set_interaction(11, 13, Aij=-235.7000, Aji=461.3000)
        self.set_interaction(11, 15, Aij=-73.5000, Aji=136.0000)
        self.set_interaction(11, 16, Aij=-196.7000, Aji=2888.6001)
        self.set_interaction(11, 17, Aij=475.5000, Aji=-294.8000)
        self.set_interaction(11, 18, Aij=-0.1300, Aji=8.8700)
        self.set_interaction(11, 19, Aij=494.6000, Aji=-266.6000)
        self.set_interaction(11, 20, Aij=660.2000, Aji=-256.3000)
        self.set_interaction(11, 21, Aij=-34.7400, Aji=35.3800)
        self.set_interaction(11, 22, Aij=108.8500, Aji=-132.9500)
        self.set_interaction(11, 23, Aij=-209.6600, Aji=176.4500)
        self.set_interaction(11, 24, Aij=54.5700, Aji=129.4900)
        self.set_interaction(11, 25, Aij=442.4000, Aji=-171.1000)
        self.set_interaction(11, 26, Aij=-81.1300, Aji=129.3000)
        self.set_interaction(11, 28, Aij=183.0460, Aji=243.7750)
        self.set_interaction(11, 30, Aij=202.2500, Aji=-146.3100)
        self.set_interaction(11, 31, Aij=-101.7000, Aji=152.0000)
        self.set_interaction(11, 32, Aij=148.3000, Aji=21.9200)
        self.set_interaction(11, 33, Aij=18.8800, Aji=24.3700)
        self.set_interaction(11, 34, Aij=71.4800, Aji=-111.4500)
        self.set_interaction(11, 35, Aij=52.0800, Aji=41.5700)
        self.set_interaction(11, 36, Aij=-28.6100, Aji=175.5300)
        self.set_interaction(11, 37, Aij=-275.2000, Aji=611.3000)
        self.set_interaction(11, 39, Aij=85.3300, Aji=-82.1200)
        self.set_interaction(11, 41, Aij=560.2000, Aji=-234.9000)
        self.set_interaction(11, 47, Aij=151.8000, Aji=-3.4440)
        self.set_interaction(12, 13, Aij=-234.0000, Aji=457.3000)
        self.set_interaction(12, 18, Aij=-233.4000, Aji=554.4000)
        self.set_interaction(12, 19, Aij=-47.2500, Aji=99.3700)
        self.set_interaction(12, 20, Aij=-268.1000, Aji=193.9000)
        self.set_interaction(12, 22, Aij=31.0000, Aji=80.9900)
        self.set_interaction(12, 23, Aij=-126.2000, Aji=235.6000)
        self.set_interaction(12, 24, Aij=179.7000, Aji=351.9000)
        self.set_interaction(12, 25, Aij=24.2800, Aji=383.3000)
        self.set_interaction(12, 29, Aij=103.9000, Aji=201.5000)
        self.set_interaction(12, 33, Aij=298.1300, Aji=-92.2600)
        self.set_interaction(12, 37, Aij=-11.4000, Aji=134.5000)
        self.set_interaction(12, 39, Aij=308.9000, Aji=-116.7000)
        self.set_interaction(12, 41, Aij=-122.3000, Aji=145.4000)
        self.set_interaction(13, 14, Aij=-78.3600, Aji=222.1000)
        self.set_interaction(13, 15, Aij=251.5000, Aji=-56.0800)
        self.set_interaction(13, 16, Aij=5422.2998, Aji=-194.1000)
        self.set_interaction(13, 17, Aij=-46.3900, Aji=285.3600)
        self.set_interaction(13, 18, Aij=213.2000, Aji=-156.1000)
        self.set_interaction(13, 19, Aij=-18.5100, Aji=38.8100)
        self.set_interaction(13, 20, Aij=664.6000, Aji=-338.5000)
        self.set_interaction(13, 21, Aij=301.1400, Aji=225.3900)
        self.set_interaction(13, 22, Aij=137.7700, Aji=-197.7100)
        self.set_interaction(13, 23, Aij=-154.3000, Aji=-20.9300)
        self.set_interaction(13, 24, Aij=47.6700, Aji=113.9000)
        self.set_interaction(13, 25, Aij=134.8000, Aji=-25.1500)
        self.set_interaction(13, 26, Aij=95.1800, Aji=-94.4900)
        self.set_interaction(13, 27, Aij=155.1100, Aji=220.6600)
        self.set_interaction(13, 28, Aij=140.8960, Aji=112.3820)
        self.set_interaction(13, 29, Aij=-8.5380, Aji=63.7100)
        self.set_interaction(13, 30, Aij=170.1000, Aji=-87.3100)
        self.set_interaction(13, 31, Aij=-20.1100, Aji=9.2070)
        self.set_interaction(13, 32, Aij=-149.5000, Aji=476.6000)
        self.set_interaction(13, 33, Aij=-202.3000, Aji=736.4000)
        self.set_interaction(13, 34, Aij=-156.5700, Aji=173.7700)
        self.set_interaction(13, 35, Aij=128.8000, Aji=-93.5100)
        self.set_interaction(13, 37, Aij=240.2000, Aji=-217.9000)
        self.set_interaction(13, 38, Aij=-273.9500, Aji=167.3000)
        self.set_interaction(13, 39, Aij=254.8000, Aji=-158.2000)
        self.set_interaction(13, 40, Aij=-172.5100, Aji=278.1500)
        self.set_interaction(13, 41, Aij=417.0000, Aji=-247.8000)
        self.set_interaction(13, 42, Aij=1338.0000, Aji=448.5000)
        self.set_interaction(14, 15, Aij=-107.2000, Aji=127.4000)
        self.set_interaction(14, 16, Aij=-41.1100, Aji=38.8900)
        self.set_interaction(14, 17, Aij=-200.7000, Aji=-15.0700)
        self.set_interaction(14, 19, Aij=358.9000, Aji=-157.3000)
        self.set_interaction(14, 21, Aij=-82.9200, Aji=131.2000)
        self.set_interaction(14, 24, Aij=-99.8100, Aji=261.1000)
        self.set_interaction(14, 25, Aij=30.0500, Aji=108.5000)
        self.set_interaction(14, 29, Aij=-70.1400, Aji=106.7000)
        self.set_interaction(14, 35, Aij=874.1900, Aji=-366.5100)
        self.set_interaction(14, 39, Aij=-164.0000, Aji=49.7000)
        self.set_interaction(14, 42, Aij=-664.4000, Aji=961.8000)
        self.set_interaction(14, 43, Aij=275.9000, Aji=-125.2000)
        self.set_interaction(15, 16, Aij=-189.2000, Aji=865.9000)
        self.set_interaction(15, 17, Aij=138.5400, Aji=64.3000)
        self.set_interaction(15, 18, Aij=431.4900, Aji=-207.6600)
        self.set_interaction(15, 19, Aij=147.1000, Aji=-108.5000)
        self.set_interaction(15, 24, Aij=71.2300, Aji=91.1300)
        self.set_interaction(15, 25, Aij=-18.9300, Aji=102.2000)
        self.set_interaction(15, 31, Aij=939.0700, Aji=-213.7400)
        self.set_interaction(15, 38, Aij=570.9000, Aji=-198.8000)
        self.set_interaction(15, 39, Aij=-255.2200, Aji=10.0300)
        self.set_interaction(15, 41, Aij=-38.7700, Aji=284.5000)
        self.set_interaction(15, 42, Aij=448.1000, Aji=1464.2000)
        self.set_interaction(15, 43, Aij=-1327.0000, Aji=1603.8000)
        self.set_interaction(16, 17, Aij=287.4300, Aji=-24.4600)
        self.set_interaction(16, 19, Aij=1255.1000, Aji=-446.8600)
        self.set_interaction(16, 21, Aij=-182.9100, Aji=151.3800)
        self.set_interaction(16, 22, Aij=-73.8500, Aji=-141.4000)
        self.set_interaction(16, 23, Aij=-352.9000, Aji=-293.7000)
        self.set_interaction(16, 24, Aij=-262.0000, Aji=316.9000)
        self.set_interaction(16, 25, Aij=-181.9000, Aji=2951.0000)
        self.set_interaction(16, 35, Aij=243.1000, Aji=-257.2000)
        self.set_interaction(16, 38, Aij=-196.3120, Aji=116.4780)
        self.set_interaction(16, 39, Aij=22.0500, Aji=-185.2000)
        self.set_interaction(17, 18, Aij=89.7000, Aji=117.4000)
        self.set_interaction(17, 19, Aij=-281.6000, Aji=777.4000)
        self.set_interaction(17, 20, Aij=-396.0000, Aji=493.8000)
        self.set_interaction(17, 21, Aij=287.0000, Aji=429.7000)
        self.set_interaction(17, 22, Aij=-111.0000, Aji=140.8000)
        self.set_interaction(17, 24, Aij=882.0000, Aji=898.2000)
        self.set_interaction(17, 25, Aij=617.5000, Aji=334.9000)
        self.set_interaction(17, 27, Aij=-139.3000, Aji=134.9000)
        self.set_interaction(17, 31, Aij=0.1004, Aji=192.3000)
        self.set_interaction(17, 39, Aij=-334.4000, Aji=343.7000)
        self.set_interaction(17, 41, Aij=-89.4200, Aji=-22.1000)
        self.set_interaction(18, 19, Aij=-169.6700, Aji=134.2800)
        self.set_interaction(18, 20, Aij=-153.7000, Aji=-313.5000)
        self.set_interaction(18, 22, Aij=-351.6000, Aji=587.3000)
        self.set_interaction(18, 23, Aij=-114.7300, Aji=18.9800)
        self.set_interaction(18, 24, Aij=-205.3000, Aji=368.5000)
        self.set_interaction(18, 25, Aij=-2.1700, Aji=20.1800)
        self.set_interaction(18, 27, Aij=2845.0000, Aji=2475.0000)
        self.set_interaction(18, 33, Aij=-60.7800, Aji=-42.7100)
        self.set_interaction(18, 37, Aij=160.7000, Aji=281.6000)
        self.set_interaction(18, 38, Aij=-158.8000, Aji=159.8000)
        self.set_interaction(18, 50, Aij=-136.6000, Aji=221.4000)
        self.set_interaction(19, 20, Aij=205.2700, Aji=92.0700)
        self.set_interaction(19, 21, Aij=4.9330, Aji=54.3200)
        self.set_interaction(19, 22, Aij=-152.7000, Aji=258.6000)
        self.set_interaction(19, 23, Aij=-15.6200, Aji=74.0400)
        self.set_interaction(19, 24, Aij=-54.8600, Aji=491.9500)
        self.set_interaction(19, 25, Aij=-4.6240, Aji=363.5000)
        self.set_interaction(19, 26, Aij=-0.5150, Aji=0.2830)
        self.set_interaction(19, 28, Aij=230.8520, Aji=335.7430)
        self.set_interaction(19, 29, Aij=0.4604, Aji=161.0000)
        self.set_interaction(19, 31, Aij=177.5000, Aji=169.6000)
        self.set_interaction(19, 33, Aij=-62.1700, Aji=136.9000)
        self.set_interaction(19, 34, Aij=-203.0200, Aji=329.1200)
        self.set_interaction(19, 36, Aij=81.5700, Aji=-42.3100)
        self.set_interaction(19, 37, Aij=-55.7700, Aji=335.2000)
        self.set_interaction(19, 39, Aij=-151.5000, Aji=150.6000)
        self.set_interaction(19, 41, Aij=120.3000, Aji=-61.6000)
        self.set_interaction(19, 47, Aij=16.2300, Aji=119.2000)
        self.set_interaction(20, 21, Aij=13.4100, Aji=519.1000)
        self.set_interaction(20, 22, Aij=-44.7000, Aji=543.3000)
        self.set_interaction(20, 23, Aij=39.6300, Aji=504.2000)
        self.set_interaction(20, 24, Aij=183.4000, Aji=631.0000)
        self.set_interaction(20, 25, Aij=-79.0800, Aji=993.4000)
        self.set_interaction(20, 30, Aij=-208.9000, Aji=570.6000)
        self.set_interaction(20, 32, Aij=228.4000, Aji=616.6000)
        self.set_interaction(20, 33, Aij=-95.0000, Aji=5256.0000)
        self.set_interaction(20, 35, Aij=-463.6000, Aji=-180.2000)
        self.set_interaction(20, 37, Aij=-11.1600, Aji=898.2000)
        self.set_interaction(20, 39, Aij=-228.0000, Aji=-97.7700)
        self.set_interaction(20, 41, Aij=-337.0000, Aji=1179.0000)
        self.set_interaction(20, 46, Aij=-322.3000, Aji=-70.2500)
        self.set_interaction(21, 22, Aij=108.3100, Aji=-84.5300)
        self.set_interaction(21, 23, Aij=249.1500, Aji=-157.1000)
        self.set_interaction(21, 24, Aij=62.4200, Aji=11.8000)
        self.set_interaction(21, 25, Aij=153.0000, Aji=-129.7000)
        self.set_interaction(21, 26, Aij=32.7300, Aji=113.0000)
        self.set_interaction(21, 27, Aij=86.2000, Aji=1971.0000)
        self.set_interaction(21, 28, Aij=450.0880, Aji=-73.0920)
        self.set_interaction(21, 29, Aij=59.0200, Aji=-27.9400)
        self.set_interaction(21, 30, Aij=65.5600, Aji=-39.4600)
        self.set_interaction(21, 32, Aij=2.2200, Aji=179.2500)
        self.set_interaction(21, 33, Aij=344.4000, Aji=-262.3000)
        self.set_interaction(21, 37, Aij=-168.2000, Aji=383.2000)
        self.set_interaction(21, 39, Aij=6.5700, Aji=-55.2100)
        self.set_interaction(21, 41, Aij=63.6700, Aji=182.2000)
        self.set_interaction(22, 23, Aij=0.0000, Aji=0.0000)
        self.set_interaction(22, 24, Aij=56.3300, Aji=17.9700)
        self.set_interaction(22, 25, Aij=223.1000, Aji=-8.3090)
        self.set_interaction(22, 26, Aij=108.9000, Aji=-9.6390)
        self.set_interaction(22, 30, Aij=149.5600, Aji=-116.2100)
        self.set_interaction(22, 32, Aij=177.6000, Aji=-40.8200)
        self.set_interaction(22, 33, Aij=315.9000, Aji=-174.5000)
        self.set_interaction(22, 35, Aij=215.0000, Aji=-215.0000)
        self.set_interaction(22, 37, Aij=-91.8000, Aji=301.9000)
        self.set_interaction(22, 39, Aij=-160.2800, Aji=397.2400)
        self.set_interaction(22, 41, Aij=-96.8700, Aji=305.4000)
        self.set_interaction(22, 47, Aij=361.1000, Aji=-194.7000)
        self.set_interaction(23, 24, Aij=-30.1000, Aji=51.9000)
        self.set_interaction(23, 25, Aij=192.1000, Aji=-0.2266)
        self.set_interaction(23, 28, Aij=116.6120, Aji=-26.0580)
        self.set_interaction(23, 30, Aij=-64.3800, Aji=48.4840)
        self.set_interaction(23, 32, Aij=86.4000, Aji=21.7600)
        self.set_interaction(23, 33, Aij=168.8000, Aji=-46.8000)
        self.set_interaction(23, 35, Aij=363.7000, Aji=-343.6000)
        self.set_interaction(23, 37, Aij=111.2000, Aji=-149.8000)
        self.set_interaction(23, 41, Aij=255.8000, Aji=-193.0000)
        self.set_interaction(23, 44, Aij=-35.6800, Aji=-196.2000)
        self.set_interaction(23, 48, Aij=565.9000, Aji=-363.1000)
        self.set_interaction(24, 25, Aij=-75.9700, Aji=248.4000)
        self.set_interaction(24, 26, Aij=490.8800, Aji=-34.6800)
        self.set_interaction(24, 27, Aij=534.7000, Aji=514.6000)
        self.set_interaction(24, 28, Aij=132.2000, Aji=-60.7100)
        self.set_interaction(24, 30, Aij=546.6800, Aji=-133.1600)
        self.set_interaction(24, 32, Aij=247.8000, Aji=48.4900)
        self.set_interaction(24, 33, Aij=146.6000, Aji=77.5500)
        self.set_interaction(24, 35, Aij=337.7000, Aji=-58.4300)
        self.set_interaction(24, 36, Aij=369.4900, Aji=-85.1480)
        self.set_interaction(24, 37, Aij=187.1000, Aji=-134.2000)
        self.set_interaction(24, 38, Aij=215.2000, Aji=-124.6000)
        self.set_interaction(24, 39, Aij=498.6000, Aji=-186.7000)
        self.set_interaction(24, 41, Aij=256.5000, Aji=335.7000)
        self.set_interaction(24, 43, Aij=233.1000, Aji=70.8100)
        self.set_interaction(24, 47, Aij=423.1000, Aji=3.1630)
        self.set_interaction(24, 48, Aij=63.9500, Aji=-11.3000)
        self.set_interaction(24, 50, Aij=108.5000, Aji=-79.3400)
        self.set_interaction(24, 55, Aij=585.1900, Aji=75.0400)
        self.set_interaction(25, 26, Aij=132.7000, Aji=132.9000)
        self.set_interaction(25, 27, Aij=2213.0000, Aji=-123.1000)
        self.set_interaction(25, 33, Aij=593.4000, Aji=-185.3000)
        self.set_interaction(25, 35, Aij=1337.3700, Aji=-334.1200)
        self.set_interaction(25, 39, Aij=5143.1401, Aji=-374.1600)
        self.set_interaction(25, 40, Aij=309.5800, Aji=33.9500)
        self.set_interaction(25, 41, Aij=-71.1800, Aji=956.1000)
        self.set_interaction(25, 44, Aij=-209.7000, Aji=161.5000)
        self.set_interaction(25, 47, Aij=434.1000, Aji=7.0820)
        self.set_interaction(26, 27, Aij=533.2000, Aji=-85.1200)
        self.set_interaction(26, 28, Aij=320.2000, Aji=277.8000)
        self.set_interaction(26, 31, Aij=139.8220, Aji=481.3480)
        self.set_interaction(26, 32, Aij=304.3000, Aji=64.2800)
        self.set_interaction(26, 33, Aij=10.1700, Aji=125.3000)
        self.set_interaction(26, 34, Aij=-27.7010, Aji=174.4330)
        self.set_interaction(26, 37, Aij=10.7600, Aji=379.4000)
        self.set_interaction(26, 39, Aij=-223.1000, Aji=223.6000)
        self.set_interaction(26, 41, Aij=248.4000, Aji=-124.7000)
        self.set_interaction(26, 45, Aij=-218.9000, Aji=844.0000)
        self.set_interaction(26, 50, Aij=-4.5650, Aji=176.3000)
        self.set_interaction(27, 32, Aij=2990.0000, Aji=2448.0000)
        self.set_interaction(27, 33, Aij=-124.0000, Aji=4288.0000)
        self.set_interaction(28, 32, Aij=292.7000, Aji=-27.4500)
        self.set_interaction(28, 37, Aij=-47.3700, Aji=167.9000)
        self.set_interaction(28, 41, Aij=469.8000, Aji=885.5000)
        self.set_interaction(29, 35, Aij=31.6600, Aji=85.7000)
        self.set_interaction(29, 39, Aij=78.9200, Aji=-71.0000)
        self.set_interaction(29, 44, Aij=1004.2000, Aji=-274.1000)
        self.set_interaction(29, 48, Aij=-18.2700, Aji=6.9710)
        self.set_interaction(30, 41, Aij=43.3700, Aji=-64.2800)
        self.set_interaction(31, 35, Aij=-417.2000, Aji=535.8000)
        self.set_interaction(31, 39, Aij=302.2000, Aji=-191.7000)
        self.set_interaction(31, 41, Aij=347.8000, Aji=-264.3000)
        self.set_interaction(31, 44, Aij=-262.0000, Aji=262.0000)
        self.set_interaction(31, 47, Aij=-353.5000, Aji=515.8000)
        self.set_interaction(32, 33, Aij=6.3700, Aji=37.1000)
        self.set_interaction(32, 41, Aij=68.5500, Aji=288.1000)
        self.set_interaction(33, 35, Aij=32.9000, Aji=-111.2000)
        self.set_interaction(33, 37, Aij=-48.3300, Aji=322.4200)
        self.set_interaction(33, 39, Aij=336.2500, Aji=-176.2600)
        self.set_interaction(33, 41, Aij=-195.1000, Aji=627.7000)
        self.set_interaction(34, 37, Aij=2073.2000, Aji=631.5000)
        self.set_interaction(34, 39, Aij=-119.8000, Aji=6.6990)
        self.set_interaction(35, 39, Aij=-97.7100, Aji=136.6000)
        self.set_interaction(35, 41, Aij=153.7000, Aji=-29.3400)
        self.set_interaction(36, 37, Aij=-208.8000, Aji=837.2000)
        self.set_interaction(36, 39, Aij=-8.8040, Aji=5.1500)
        self.set_interaction(36, 41, Aij=423.4000, Aji=-53.9100)
        self.set_interaction(37, 39, Aij=255.0000, Aji=-137.7000)
        self.set_interaction(37, 41, Aij=730.8000, Aji=-198.0000)
        self.set_interaction(37, 44, Aij=26.3500, Aji=-66.3100)
        self.set_interaction(37, 48, Aij=2429.0000, Aji=148.9000)
        self.set_interaction(38, 39, Aij=-110.6500, Aji=50.0600)
        self.set_interaction(38, 40, Aij=-117.1700, Aji=185.6000)
        self.set_interaction(39, 40, Aij=-5.5790, Aji=55.8000)
        self.set_interaction(39, 41, Aij=72.3100, Aji=-28.6500)
        self.set_interaction(40, 45, Aij=111.8000, Aji=-32.1700)
        self.set_interaction(41, 47, Aij=122.4000, Aji=101.2000)
        self.set_interaction(42, 43, Aij=-2166.0000, Aji=745.3000)
        self.set_interaction(84, 85, Aij=1517.5000, Aji=-1869.9000)

