#  @file
#  @author Christian Diddens <c.diddens@utwente.nl>
#  @author Duarte Rocha <d.rocha@utwente.nl>
#  
#  @section LICENSE
# 
#  pyoomph - a multi-physics finite element framework based on oomph-lib and GiNaC 
#  Copyright (C) 2021-2025  Christian Diddens & Duarte Rocha
# 
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
# 
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>. 
#
#  The authors may be contacted at c.diddens@utwente.nl and d.rocha@utwente.nl
#
# ========================================================================


##########################################################################################
# These parameters stem from the AIOMFAC activity model
#		http://www.aiomfac.caltech.edu/
#
# For the source code and parameter tables (GPL v3), see
#		https://github.com/andizuend/AIOMFAC
#
# Cite the relevant publications when publishing results based on AIOMFAC
#		see: https://aiomfac.lab.mcgill.ca/citation.html
#
##########################################################################################


from ..activity import UNIFACLikeActivityModel,ActivityModel

@ActivityModel.register_activity_model()
class AIOMFAC(UNIFACLikeActivityModel):
    """
    AIOMFAC activity model (http://www.aiomfac.caltech.edu/), parameters were taken from the AIOMFAC source code (https://github.com/andizuend/AIOMFAC).
    
    Cite the publications when using this activity model (https://aiomfac.lab.mcgill.ca/citation.html).
    """
    name="AIOMFAC"
    def __init__(self):
        super(AIOMFAC, self).__init__()
#        self.modified_volume_fraction_exponent=1
        self.define_groups()
        self.define_interaction_table()

    def define_groups(self):
        with self.define_main_group("unnamed_1", index=1):
            self.define_sub_group("CH3", R=0.9011, Q=0.848, molar_mass=0.015034, index=1)
            self.define_sub_group("CH2", R=0.6744, Q=0.54, molar_mass=0.014026, index=2)
            self.define_sub_group("CH", R=0.4469, Q=0.228, molar_mass=0.013018, index=3)
            self.define_sub_group("C", R=0.2195, Q=0.0, molar_mass=0.01201, index=4)
        with self.define_main_group("unnamed_2", index=2):
            self.define_sub_group("CH=CH2", R=1.3454, Q=1.176, molar_mass=0.027044, index=5)
            self.define_sub_group("CH=CH", R=1.1167, Q=0.867, molar_mass=0.026, index=6)
            self.define_sub_group("CH2=C", R=1.1173, Q=0.988, molar_mass=0.026, index=7)
            self.define_sub_group("CH=C", R=0.8886, Q=0.676, molar_mass=0.025, index=8)
            self.define_sub_group("C=C (double bond)", R=0.6605, Q=0.485, molar_mass=0.031, index=70)
        with self.define_main_group("unnamed_3", index=3):
            self.define_sub_group("ACH", R=0.5313, Q=0.4, molar_mass=0.013018, index=9)
            self.define_sub_group("AC", R=0.3652, Q=0.12, molar_mass=0.01201, index=10)
        with self.define_main_group("unnamed_4", index=4):
            self.define_sub_group("ACCH3", R=1.2663, Q=0.968, molar_mass=0.027044, index=11)
            self.define_sub_group("ACCH2", R=1.0396, Q=0.66, molar_mass=0.026, index=12)
            self.define_sub_group("ACCH", R=0.8121, Q=0.348, molar_mass=0.025, index=13)
        with self.define_main_group("unnamed_5", index=5):
            self.define_sub_group("OH", R=1.0, Q=1.2, molar_mass=0.017008, index=14)
        with self.define_main_group("unnamed_6", index=6):
            self.define_sub_group("CH3OH", R=1.4311, Q=1.432, molar_mass=0.032, index=15)
        with self.define_main_group("unnamed_7", index=7):
            self.define_sub_group("H2O", R=0.92, Q=1.4, molar_mass=0.01801528, index=16)
        with self.define_main_group("unnamed_8", index=8):
            self.define_sub_group("ACOH", R=0.8952, Q=0.68, molar_mass=0.029, index=17)
        with self.define_main_group("unnamed_9", index=9):
            self.define_sub_group("CH3CO (ketone)", R=1.6724, Q=1.448, molar_mass=0.043, index=18)
            self.define_sub_group("CH2CO (ketone)", R=1.4457, Q=1.18, molar_mass=0.042, index=19)
        with self.define_main_group("unnamed_10", index=10):
            self.define_sub_group("CHO (aldehyde)", R=0.998, Q=0.948, molar_mass=0.029, index=20)
        with self.define_main_group("unnamed_11", index=11):
            self.define_sub_group("CH3COO (ester)", R=1.9031, Q=1.728, molar_mass=0.059, index=21)
            self.define_sub_group("CH2COO (ester)", R=1.6764, Q=1.42, molar_mass=0.058, index=22)
        with self.define_main_group("unnamed_12", index=12):
            self.define_sub_group("HCOO", R=1.242, Q=1.188, molar_mass=0.045, index=23)
        with self.define_main_group("unnamed_13", index=13):
            self.define_sub_group("CH3O (ether)", R=1.145, Q=1.088, molar_mass=0.031, index=24)
            self.define_sub_group("CH2O (ether)", R=0.9183, Q=0.78, molar_mass=0.03, index=25)
            self.define_sub_group("CHO (ether)", R=0.6908, Q=0.468, molar_mass=0.029, index=26)
            self.define_sub_group("THF (tetrahydrofuran)", R=0.9183, Q=1.1, molar_mass=0.03, index=27)
        with self.define_main_group("unnamed_14", index=14):
            self.define_sub_group("CH3NH2", R=1.5959, Q=1.544, molar_mass=0.03105, index=28)
            self.define_sub_group("CH2NH2", R=1.3692, Q=1.236, molar_mass=0.03, index=29)
            self.define_sub_group("CHNH2", R=1.1417, Q=0.924, molar_mass=0.029, index=30)
        with self.define_main_group("unnamed_15", index=15):
            self.define_sub_group("CH3NH", R=1.4337, Q=1.244, molar_mass=0.03, index=31)
            self.define_sub_group("CH2NH", R=1.207, Q=0.936, molar_mass=0.029, index=32)
            self.define_sub_group("CHNH", R=0.9795, Q=0.624, molar_mass=0.028, index=33)
        with self.define_main_group("unnamed_16", index=16):
            self.define_sub_group("CH3N", R=1.1865, Q=0.94, molar_mass=0.029, index=34)
            self.define_sub_group("CH2N", R=0.9597, Q=0.632, molar_mass=0.028, index=35)
        with self.define_main_group("unnamed_17", index=17):
            self.define_sub_group("ACNH2", R=1.06, Q=0.816, molar_mass=0.028, index=36)
        with self.define_main_group("unnamed_18", index=18):
            self.define_sub_group("C5H5N", R=2.9993, Q=2.113, molar_mass=0.079, index=37)
            self.define_sub_group("C5H4N", R=2.8332, Q=1.833, molar_mass=0.078, index=38)
            self.define_sub_group("C5H3N", R=2.667, Q=1.553, molar_mass=0.077, index=39)
        with self.define_main_group("unnamed_19", index=19):
            self.define_sub_group("CH3CN", R=1.8701, Q=1.724, molar_mass=0.041, index=40)
            self.define_sub_group("CH2CN", R=1.6434, Q=1.416, molar_mass=0.04, index=41)
        with self.define_main_group("unnamed_20", index=20):
            self.define_sub_group("COOH (carboxylic acid)", R=1.3013, Q=1.224, molar_mass=0.045, index=42)
            self.define_sub_group("HCOOH", R=1.528, Q=1.532, molar_mass=0.046, index=43)
        with self.define_main_group("unnamed_21", index=21):
            self.define_sub_group("CH2CL", R=1.4654, Q=1.264, molar_mass=0.0495, index=44)
            self.define_sub_group("CHCL", R=1.238, Q=0.952, molar_mass=0.0485, index=45)
            self.define_sub_group("CCL", R=1.0106, Q=0.724, molar_mass=0.0475, index=46)
        with self.define_main_group("unnamed_22", index=22):
            self.define_sub_group("CH2fCL2", R=2.2564, Q=1.998, molar_mass=0.085, index=47)
            self.define_sub_group("CHCL2", R=2.0606, Q=1.684, molar_mass=0.084, index=48)
            self.define_sub_group("CCL2", R=1.8016, Q=1.448, molar_mass=0.083, index=49)
        with self.define_main_group("unnamed_23", index=23):
            self.define_sub_group("CHCL3", R=2.87, Q=2.41, molar_mass=0.1195, index=50)
            self.define_sub_group("CCL3", R=2.6401, Q=2.184, molar_mass=0.1185, index=51)
        with self.define_main_group("unnamed_24", index=24):
            self.define_sub_group("CCL4", R=3.39, Q=2.91, molar_mass=0.154, index=52)
        with self.define_main_group("unnamed_25", index=25):
            self.define_sub_group("ACCL", R=1.1562, Q=0.844, molar_mass=0.0475, index=53)
        with self.define_main_group("unnamed_26", index=26):
            self.define_sub_group("CH3NO2", R=2.0086, Q=1.868, molar_mass=0.061, index=54)
            self.define_sub_group("CH2NO2", R=1.7818, Q=1.56, molar_mass=0.06, index=55)
            self.define_sub_group("CHNO2", R=1.5544, Q=1.248, molar_mass=0.059, index=56)
        with self.define_main_group("unnamed_27", index=27):
            self.define_sub_group("ACNO2", R=1.4199, Q=1.104, molar_mass=0.058, index=57)
        with self.define_main_group("unnamed_28", index=28):
            self.define_sub_group("CS2", R=2.057, Q=1.65, molar_mass=0.076, index=58)
        with self.define_main_group("unnamed_29", index=29):
            self.define_sub_group("CH3SH", R=1.877, Q=1.676, molar_mass=0.048, index=59)
            self.define_sub_group("CH2SH", R=1.651, Q=1.368, molar_mass=0.047, index=60)
        with self.define_main_group("unnamed_30", index=30):
            self.define_sub_group("FURFURAL", R=3.168, Q=2.484, molar_mass=0.09609, index=61)
        with self.define_main_group("unnamed_31", index=31):
            self.define_sub_group("DOH", R=2.4088, Q=2.248, molar_mass=0.019, index=62)
        with self.define_main_group("unnamed_32", index=32):
            self.define_sub_group("I (Iodide compound)", R=1.264, Q=0.992, molar_mass=0.1269, index=63)
        with self.define_main_group("unnamed_33", index=33):
            self.define_sub_group("Br (Bromide compound)", R=0.9492, Q=0.832, molar_mass=0.0799, index=64)
        with self.define_main_group("unnamed_34", index=34):
            self.define_sub_group("CH-=C (triple bond)", R=1.292, Q=1.088, molar_mass=0.025, index=65)
            self.define_sub_group("C-=C (triple bond)", R=1.0613, Q=0.784, molar_mass=0.024, index=66)
        with self.define_main_group("unnamed_35", index=35):
            self.define_sub_group("DMSO", R=2.8266, Q=2.472, molar_mass=0.07813, index=67)
        with self.define_main_group("unnamed_36", index=36):
            self.define_sub_group("Acrylonitril", R=2.3144, Q=2.052, molar_mass=0.05306, index=68)
        with self.define_main_group("unnamed_37", index=37):
            self.define_sub_group("Cl-(C=C)", R=0.791, Q=0.724, molar_mass=0.0595, index=69)
        with self.define_main_group("unnamed_38", index=38):
            self.define_sub_group("ACF", R=0.6948, Q=0.524, molar_mass=0.07309, index=71)
        with self.define_main_group("unnamed_39", index=39):
            self.define_sub_group("DMF", R=3.0856, Q=2.736, molar_mass=0.071, index=72)
            self.define_sub_group("HCON(CH2)2", R=2.6322, Q=2.12, molar_mass=0.069, index=73)
        with self.define_main_group("unnamed_40", index=40):
            self.define_sub_group("CF3", R=1.406, Q=1.38, molar_mass=0.05, index=74)
            self.define_sub_group("CF2", R=1.0105, Q=0.92, molar_mass=0.031, index=75)
            self.define_sub_group("CF", R=0.615, Q=0.46, molar_mass=0.044, index=76)
        with self.define_main_group("unnamed_41", index=41):
            self.define_sub_group("COO", R=1.38, Q=1.2, molar_mass=0.031, index=77)
        with self.define_main_group("unnamed_42", index=42):
            self.define_sub_group("SIH3", R=1.6035, Q=1.263, molar_mass=0.03, index=78)
            self.define_sub_group("SIH2", R=1.4443, Q=1.006, molar_mass=0.03, index=79)
            self.define_sub_group("SIH", R=1.2853, Q=0.749, molar_mass=0.029, index=80)
            self.define_sub_group("SI", R=1.047, Q=0.41, molar_mass=0.028, index=81)
        with self.define_main_group("unnamed_43", index=43):
            self.define_sub_group("SIH2O", R=1.4838, Q=1.062, molar_mass=0.046, index=82)
            self.define_sub_group("SIHO", R=1.303, Q=0.764, molar_mass=0.045, index=83)
            self.define_sub_group("SIO", R=1.1044, Q=0.466, molar_mass=0.044, index=84)
        with self.define_main_group("unnamed_44", index=44):
            self.define_sub_group("NMP", R=3.981, Q=3.2, molar_mass=0.09913, index=85)
        with self.define_main_group("unnamed_45", index=45):
            self.define_sub_group("CCL3F", R=3.0356, Q=2.644, molar_mass=0.1375, index=86)
            self.define_sub_group("CCL2F", R=2.2287, Q=1.916, molar_mass=0.102, index=87)
            self.define_sub_group("HCCL2F", R=2.406, Q=2.116, molar_mass=0.103, index=88)
            self.define_sub_group("HCCLF", R=1.6493, Q=1.416, molar_mass=0.0875, index=89)
            self.define_sub_group("CCLF2", R=1.8174, Q=1.648, molar_mass=0.0855, index=90)
            self.define_sub_group("HCCLF2", R=1.967, Q=1.828, molar_mass=0.0865, index=91)
            self.define_sub_group("CCLF3", R=2.1721, Q=2.1, molar_mass=0.1045, index=92)
            self.define_sub_group("CCK2F2", R=2.6243, Q=2.376, molar_mass=0.121, index=93)
        with self.define_main_group("unnamed_46", index=46):
            self.define_sub_group("CONH2", R=1.4515, Q=1.248, molar_mass=0.044, index=94)
            self.define_sub_group("CONHCH3", R=2.1905, Q=1.796, molar_mass=0.058, index=95)
            self.define_sub_group("CONHCH2", R=1.9637, Q=1.488, molar_mass=0.057, index=96)
            self.define_sub_group("CON(CH3)2", R=2.8589, Q=2.428, molar_mass=0.072, index=97)
            self.define_sub_group("CONCH3CH2", R=2.6322, Q=2.12, molar_mass=0.071, index=98)
            self.define_sub_group("CON(CH2)2", R=2.4054, Q=1.812, molar_mass=0.07, index=99)
        with self.define_main_group("unnamed_47", index=47):
            self.define_sub_group("C2H5O2", R=2.1226, Q=1.904, molar_mass=0.061, index=100)
            self.define_sub_group("C2H4O2", R=1.8952, Q=1.592, molar_mass=0.06, index=101)
        with self.define_main_group("unnamed_48", index=48):
            self.define_sub_group("CH3S", R=1.613, Q=1.368, molar_mass=0.047, index=102)
            self.define_sub_group("CH2S", R=1.3863, Q=1.06, molar_mass=0.046, index=103)
            self.define_sub_group("CHS", R=1.1589, Q=0.748, molar_mass=0.045, index=104)
        with self.define_main_group("unnamed_49", index=49):
            self.define_sub_group("MORPH", R=3.474, Q=2.796, molar_mass=0.08712, index=105)
        with self.define_main_group("unnamed_50", index=50):
            self.define_sub_group("C4H4S", R=2.8569, Q=2.14, molar_mass=0.084, index=106)
            self.define_sub_group("C4H3S", R=2.6908, Q=1.86, molar_mass=0.083, index=107)
            self.define_sub_group("C4H2S", R=2.5247, Q=1.58, molar_mass=0.082, index=108)
        with self.define_main_group("unnamed_52", index=52):
            self.define_sub_group("CH3 long-chain", R=0.9011, Q=0.848, molar_mass=0.015, index=109)
            self.define_sub_group("CH2 long-chain", R=0.6744, Q=0.54, molar_mass=0.014, index=110)
            self.define_sub_group("CH long-chain", R=0.4469, Q=0.228, molar_mass=0.013, index=111)
            self.define_sub_group("C long-chain", R=0.2195, Q=0.0, molar_mass=0.012, index=112)
        with self.define_main_group("unnamed_53", index=53):
            self.define_sub_group("OH long-chain", R=1.0, Q=1.2, molar_mass=0.017, index=113)
        with self.define_main_group("unnamed_54", index=54):
            self.define_sub_group("COOH long-chain", R=1.3013, Q=1.224, molar_mass=0.045, index=114)
        with self.define_main_group("unnamed_55", index=55):
            self.define_sub_group("CH3CO long-chain", R=1.6724, Q=1.448, molar_mass=0.043, index=115)
            self.define_sub_group("CH2CO long-chain", R=1.4457, Q=1.18, molar_mass=0.042, index=116)
        with self.define_main_group("unnamed_56", index=56):
            self.define_sub_group("CH3 monosaccharides", R=0.9011, Q=0.848, molar_mass=0.015, index=117)
            self.define_sub_group("CH2 monosaccharides", R=0.6744, Q=0.54, molar_mass=0.014, index=118)
            self.define_sub_group("CH mono-saccharides", R=0.4469, Q=0.228, molar_mass=0.013, index=119)
            self.define_sub_group("C monosaccharides", R=0.2195, Q=0.0, molar_mass=0.012, index=120)
        with self.define_main_group("unnamed_57", index=57):
            self.define_sub_group("CH3O mono-saccharides", R=1.145, Q=1.088, molar_mass=0.031, index=121)
            self.define_sub_group("CH2O monosaccharides", R=0.9183, Q=0.78, molar_mass=0.03, index=122)
            self.define_sub_group("CHO monosaccharides", R=0.6908, Q=0.468, molar_mass=0.029, index=123)
        with self.define_main_group("unnamed_58", index=58):
            self.define_sub_group("OH mono saccharides", R=1.0, Q=1.2, molar_mass=0.017, index=124)
        with self.define_main_group("unnamed_59", index=59):
            self.define_sub_group("CH3 hydroxy-acids", R=0.9011, Q=0.848, molar_mass=0.015, index=125)
            self.define_sub_group("CH2 hydroxy-acids", R=0.6744, Q=0.54, molar_mass=0.014, index=126)
            self.define_sub_group("CH hydroxy-acids", R=0.4469, Q=0.228, molar_mass=0.013, index=127)
            self.define_sub_group("C hydroxy-acids", R=0.2195, Q=0.0, molar_mass=0.012, index=128)
        with self.define_main_group("unnamed_60", index=60):
            self.define_sub_group("COOH hydroxy-acids", R=1.3013, Q=1.224, molar_mass=0.045, index=129)
        with self.define_main_group("unnamed_61", index=61):
            self.define_sub_group("OH hydroxy-acids", R=1.0, Q=1.2, molar_mass=0.017, index=130)
        with self.define_main_group("unnamed_62", index=62):
            self.define_sub_group("CH3 diacids", R=0.9011, Q=0.848, molar_mass=0.015, index=131)
            self.define_sub_group("CH2 diacids", R=0.6744, Q=0.54, molar_mass=0.014, index=132)
            self.define_sub_group("CH diacids", R=0.4469, Q=0.228, molar_mass=0.013, index=133)
            self.define_sub_group("C diacids", R=0.2195, Q=0.0, molar_mass=0.012, index=134)
        with self.define_main_group("unnamed_63", index=63):
            self.define_sub_group("COOH diacids (carboxylic acid)", R=1.3013, Q=1.224, molar_mass=0.045, index=135)
        with self.define_main_group("unnamed_64", index=64):
            self.define_sub_group("OH(?)", R=1.0, Q=1.2, molar_mass=0.017, index=136) ###TODO : What is this? DOUBLE DEFINED NAME
        with self.define_main_group("unnamed_65", index=65):
            self.define_sub_group("COOH (carboxylic acid)(?)", R=1.3013, Q=1.224, molar_mass=0.045, index=137) ###TODO : What is this? DOUBLE DEFINED NAME
        with self.define_main_group("unnamed_66", index=66):
            self.define_sub_group("CH3(alk)", R=0.9011, Q=0.848, molar_mass=0.015034, index=141)
            self.define_sub_group("CH2(alk)", R=0.6744, Q=0.54, molar_mass=0.014026, index=142)
            self.define_sub_group("CH(alk)", R=0.4469, Q=0.228, molar_mass=0.013018, index=143)
            self.define_sub_group("C(alk)", R=0.2195, Q=0.0, molar_mass=0.01201, index=144)
        with self.define_main_group("unnamed_67", index=67):
            self.define_sub_group("CH3(long-chain)", R=0.9011, Q=0.848, molar_mass=0.015034, index=145)
            self.define_sub_group("CH2(long-chain)", R=0.6744, Q=0.54, molar_mass=0.014026, index=146)
            self.define_sub_group("CH(long-chain)", R=0.4469, Q=0.228, molar_mass=0.013018, index=147)
            self.define_sub_group("C(long-chain)", R=0.2195, Q=0.0, molar_mass=0.01201, index=148)
        with self.define_main_group("unnamed_68", index=68):
            self.define_sub_group("CH3(hydroxy)", R=0.9011, Q=0.848, molar_mass=0.015034, index=149)
            self.define_sub_group("CH2(hydroxy)", R=0.6744, Q=0.54, molar_mass=0.014026, index=150)
            self.define_sub_group("CH(hydroxy)", R=0.4469, Q=0.228, molar_mass=0.013018, index=151)
            self.define_sub_group("C(hydroxy)", R=0.2195, Q=0.0, molar_mass=0.01201, index=152)
        with self.define_main_group("unnamed_69", index=69):
            self.define_sub_group("OH(new)", R=1.0, Q=1.2, molar_mass=0.017008, index=153)
        with self.define_main_group("unnamed_70", index=70):
            self.define_sub_group("CH2OCH2", R=1.5927, Q=1.32, molar_mass=0.0440528, index=154)
        with self.define_main_group("unnamed_71", index=71):
            self.define_sub_group("CH2ONO2", R=2.1246, Q=1.8682, molar_mass=0.076026, index=155)
            self.define_sub_group("CHONO2", R=1.8971, Q=1.5562, molar_mass=0.075018, index=156)
            self.define_sub_group("CONO2", R=1.6697, Q=1.3282, molar_mass=0.07401, index=157)
        with self.define_main_group("unnamed_72", index=72):
            self.define_sub_group("CH2OOH", R=1.5869, Q=1.437, molar_mass=0.047034, index=158)
            self.define_sub_group("CHOOH", R=1.3594, Q=1.125, molar_mass=0.046026, index=159)
            self.define_sub_group("COOH (hydroperoxide)", R=1.132, Q=0.897, molar_mass=0.045018, index=160)
        with self.define_main_group("unnamed_73", index=73):
            self.define_sub_group("C(=O)OOH (peroxyacid)", R=1.7025, Q=1.5217, molar_mass=0.061018, index=161)
        with self.define_main_group("unnamed_74", index=74):
            self.define_sub_group("CH3OOCH3", R=2.4926, Q=2.288, molar_mass=0.062068, index=162)
            self.define_sub_group("CH3OOCH2", R=2.2659, Q=1.98, molar_mass=0.06106, index=163)
            self.define_sub_group("CH3OOCH", R=2.0384, Q=1.668, molar_mass=0.060052, index=164)
            self.define_sub_group("CH3OOC", R=1.811, Q=1.44, molar_mass=0.059044, index=165)
            self.define_sub_group("CH2OOCH2", R=2.0392, Q=1.672, molar_mass=0.060052, index=166)
            self.define_sub_group("CH2OOCH", R=1.8117, Q=1.36, molar_mass=0.059044, index=167)
            self.define_sub_group("CH2OOC", R=1.5843, Q=1.132, molar_mass=0.058036, index=168)
            self.define_sub_group("CHOOCH", R=1.5842, Q=1.048, molar_mass=0.058036, index=169)
            self.define_sub_group("CHOOC", R=1.3568, Q=0.82, molar_mass=0.057028, index=170)
            self.define_sub_group("COOC", R=1.1294, Q=0.592, molar_mass=0.05602, index=171)
        with self.define_main_group("unnamed_75", index=75):
            self.define_sub_group("C(=O)OONO2", R=2.6217, Q=2.2887, molar_mass=0.10601, index=172)
        with self.define_main_group("unnamed_51", index=51):
            self.define_sub_group("Li+", R=0.61, Q=0.99, molar_mass=6.941, index=201)
            self.define_sub_group("Na+", R=0.38, Q=0.62, molar_mass=22.99, index=202)
            self.define_sub_group("K+", R=0.44, Q=0.58, molar_mass=39.098, index=203)
            self.define_sub_group("NH4+", R=0.69, Q=0.78, molar_mass=18.038, index=204)
            self.define_sub_group("H+", R=1.78, Q=2.7, molar_mass=1.008, index=205)
            self.define_sub_group("Ca2+", R=2.24, Q=3.4, molar_mass=40.078, index=221)
            self.define_sub_group("Ba2+", R=3.0, Q=3.0, molar_mass=137.33, index=222)
            self.define_sub_group("Mg2+", R=5.44, Q=8.35, molar_mass=24.305, index=223)
            self.define_sub_group("Sr2+", R=1.0, Q=1.0, molar_mass=87.62, index=224)
            self.define_sub_group("Co2+", R=1.0, Q=1.0, molar_mass=58.933, index=225)
            self.define_sub_group("Ni2+", R=1.0, Q=1.0, molar_mass=58.693, index=226)
            self.define_sub_group("Cu2+", R=1.0, Q=1.0, molar_mass=63.546, index=227)
            self.define_sub_group("Zn2+", R=1.0, Q=1.0, molar_mass=65.39, index=228)
            self.define_sub_group("Hg2+", R=3.0, Q=3.0, molar_mass=200.59, index=229)
            self.define_sub_group("F-", R=0.4195, Q=0.5597, molar_mass=18.998, index=241)
            self.define_sub_group("Cl-", R=0.99, Q=0.99, molar_mass=35.453, index=242)
            self.define_sub_group("Br-", R=1.25, Q=1.16, molar_mass=79.904, index=243)
            self.define_sub_group("I-", R=1.6807, Q=1.4118, molar_mass=126.9, index=244)
            self.define_sub_group("NO3-", R=0.95, Q=0.97, molar_mass=62.004, index=245)
            self.define_sub_group("CH3COO-", R=2.05, Q=1.9, molar_mass=59.044, index=246)
            self.define_sub_group("SCN-", R=1.2325, Q=1.1506, molar_mass=58.084, index=247)
            self.define_sub_group("HSO4-", R=1.65, Q=1.4, molar_mass=97.071, index=248)
            self.define_sub_group("SO4--", R=3.34, Q=3.96, molar_mass=96.063, index=261)


    def define_interaction_table(self):
        self.set_interaction(1, 2, Aij=86.02)
        self.set_interaction(1, 3, Aij=61.13)
        self.set_interaction(1, 4, Aij=76.5)
        self.set_interaction(1, 5, Aij=986.5)
        self.set_interaction(1, 6, Aij=697.20001221)
        self.set_interaction(1, 7, Aij=1318.0)
        self.set_interaction(1, 8, Aij=1333.0)
        self.set_interaction(1, 9, Aij=476.3999939)
        self.set_interaction(1, 10, Aij=677.0)
        self.set_interaction(1, 11, Aij=232.1000061)
        self.set_interaction(1, 12, Aij=507.0)
        self.set_interaction(1, 13, Aij=251.5)
        self.set_interaction(1, 14, Aij=391.5)
        self.set_interaction(1, 15, Aij=255.69999695)
        self.set_interaction(1, 16, Aij=206.6000061)
        self.set_interaction(1, 17, Aij=920.70001221)
        self.set_interaction(1, 18, Aij=287.79998779)
        self.set_interaction(1, 19, Aij=597.0)
        self.set_interaction(1, 20, Aij=663.5)
        self.set_interaction(1, 21, Aij=35.93000031)
        self.set_interaction(1, 22, Aij=53.75999832)
        self.set_interaction(1, 23, Aij=24.89999962)
        self.set_interaction(1, 24, Aij=104.30000305)
        self.set_interaction(1, 25, Aij=11.43999958)
        self.set_interaction(1, 26, Aij=661.5)
        self.set_interaction(1, 27, Aij=543.0)
        self.set_interaction(1, 28, Aij=153.6000061)
        self.set_interaction(1, 29, Aij=184.3999939)
        self.set_interaction(1, 30, Aij=354.6000061)
        self.set_interaction(1, 31, Aij=3025.0)
        self.set_interaction(1, 32, Aij=335.79998779)
        self.set_interaction(1, 33, Aij=479.5)
        self.set_interaction(1, 34, Aij=298.8999939)
        self.set_interaction(1, 35, Aij=526.5)
        self.set_interaction(1, 36, Aij=689.0)
        self.set_interaction(1, 37, Aij=-4.18900013)
        self.set_interaction(1, 38, Aij=125.80000305)
        self.set_interaction(1, 39, Aij=485.29998779)
        self.set_interaction(1, 40, Aij=-2.85899997)
        self.set_interaction(1, 41, Aij=387.1000061)
        self.set_interaction(1, 42, Aij=-450.3999939)
        self.set_interaction(1, 43, Aij=252.69999695)
        self.set_interaction(1, 44, Aij=220.30000305)
        self.set_interaction(1, 45, Aij=-5.86899996)
        self.set_interaction(1, 46, Aij=390.8999939)
        self.set_interaction(1, 47, Aij=553.29998779)
        self.set_interaction(1, 48, Aij=187.0)
        self.set_interaction(1, 49, Aij=216.1000061)
        self.set_interaction(1, 50, Aij=92.98999786)
        self.set_interaction(1, 53, Aij=986.5)
        self.set_interaction(1, 54, Aij=663.5)
        self.set_interaction(1, 55, Aij=476.3999939)
        self.set_interaction(1, 57, Aij=251.5)
        self.set_interaction(1, 58, Aij=986.5)
        self.set_interaction(1, 60, Aij=663.5)
        self.set_interaction(1, 61, Aij=986.5)
        self.set_interaction(1, 63, Aij=663.5)
        self.set_interaction(1, 64, Aij=986.5)
        self.set_interaction(1, 65, Aij=663.5)
        self.set_interaction(1, 69, Aij=986.5)
        self.set_interaction(1, 70, Aij=66.122)
        self.set_interaction(1, 71, Aij=500.95)
        self.set_interaction(1, 72, Aij=977.56)
        self.set_interaction(1, 73, Aij=1331.0)
        self.set_interaction(1, 74, Aij=297.24)
        self.set_interaction(1, 75, Aij=528.5)
        self.set_interaction(2, 1, Aij=-35.36000061)
        self.set_interaction(2, 3, Aij=38.81000137)
        self.set_interaction(2, 4, Aij=74.15000153)
        self.set_interaction(2, 5, Aij=524.09997559)
        self.set_interaction(2, 6, Aij=787.59997559)
        self.set_interaction(2, 7, Aij=270.6000061)
        self.set_interaction(2, 8, Aij=526.09997559)
        self.set_interaction(2, 9, Aij=182.6000061)
        self.set_interaction(2, 10, Aij=448.79998779)
        self.set_interaction(2, 11, Aij=37.84999847)
        self.set_interaction(2, 12, Aij=333.5)
        self.set_interaction(2, 13, Aij=214.5)
        self.set_interaction(2, 14, Aij=240.8999939)
        self.set_interaction(2, 15, Aij=163.8999939)
        self.set_interaction(2, 16, Aij=61.11000061)
        self.set_interaction(2, 17, Aij=749.29998779)
        self.set_interaction(2, 18, Aij=280.5)
        self.set_interaction(2, 19, Aij=336.8999939)
        self.set_interaction(2, 20, Aij=318.8999939)
        self.set_interaction(2, 21, Aij=-36.86999893)
        self.set_interaction(2, 22, Aij=58.54999924)
        self.set_interaction(2, 23, Aij=-13.98999977)
        self.set_interaction(2, 24, Aij=-109.69999695)
        self.set_interaction(2, 25, Aij=100.09999847)
        self.set_interaction(2, 26, Aij=357.5)
        self.set_interaction(2, 28, Aij=76.30000305)
        self.set_interaction(2, 30, Aij=262.8999939)
        self.set_interaction(2, 33, Aij=183.80000305)
        self.set_interaction(2, 34, Aij=31.13999939)
        self.set_interaction(2, 35, Aij=179.0)
        self.set_interaction(2, 36, Aij=-52.86999893)
        self.set_interaction(2, 37, Aij=-66.45999908)
        self.set_interaction(2, 38, Aij=359.29998779)
        self.set_interaction(2, 39, Aij=-70.44999695)
        self.set_interaction(2, 40, Aij=449.3999939)
        self.set_interaction(2, 41, Aij=48.33000183)
        self.set_interaction(2, 44, Aij=86.45999908)
        self.set_interaction(2, 46, Aij=200.19999695)
        self.set_interaction(2, 47, Aij=268.1000061)
        self.set_interaction(2, 48, Aij=-617.0)
        self.set_interaction(2, 49, Aij=62.56000137)
        self.set_interaction(2, 52, Aij=-35.36000061)
        self.set_interaction(2, 53, Aij=524.09997559)
        self.set_interaction(2, 54, Aij=318.8999939)
        self.set_interaction(2, 55, Aij=182.6000061)
        self.set_interaction(2, 56, Aij=-35.36000061)
        self.set_interaction(2, 57, Aij=214.5)
        self.set_interaction(2, 58, Aij=524.09997559)
        self.set_interaction(2, 59, Aij=-35.36000061)
        self.set_interaction(2, 60, Aij=318.8999939)
        self.set_interaction(2, 61, Aij=524.09997559)
        self.set_interaction(2, 62, Aij=-35.36000061)
        self.set_interaction(2, 63, Aij=318.8999939)
        self.set_interaction(2, 64, Aij=524.09997559)
        self.set_interaction(2, 65, Aij=318.8999939)
        self.set_interaction(2, 66, Aij=-35.36000061)
        self.set_interaction(2, 67, Aij=-35.36000061)
        self.set_interaction(2, 68, Aij=-35.36000061)
        self.set_interaction(2, 69, Aij=524.09997559)
        self.set_interaction(2, 71, Aij=10326.0)
        self.set_interaction(2, 72, Aij=475.91)
        self.set_interaction(2, 73, Aij=742.38)
        self.set_interaction(2, 74, Aij=606.71)
        self.set_interaction(2, 75, Aij=469.27)
        self.set_interaction(3, 1, Aij=-11.11999989)
        self.set_interaction(3, 2, Aij=3.4460001)
        self.set_interaction(3, 4, Aij=167.0)
        self.set_interaction(3, 5, Aij=636.09997559)
        self.set_interaction(3, 6, Aij=637.40002441)
        self.set_interaction(3, 7, Aij=903.79998779)
        self.set_interaction(3, 8, Aij=1329.0)
        self.set_interaction(3, 9, Aij=25.77000046)
        self.set_interaction(3, 10, Aij=347.29998779)
        self.set_interaction(3, 11, Aij=5.99399996)
        self.set_interaction(3, 12, Aij=287.1000061)
        self.set_interaction(3, 13, Aij=32.13999939)
        self.set_interaction(3, 14, Aij=161.69999695)
        self.set_interaction(3, 15, Aij=122.80000305)
        self.set_interaction(3, 16, Aij=90.45999908)
        self.set_interaction(3, 17, Aij=648.20001221)
        self.set_interaction(3, 18, Aij=-4.44899988)
        self.set_interaction(3, 19, Aij=212.5)
        self.set_interaction(3, 20, Aij=537.40002441)
        self.set_interaction(3, 21, Aij=-18.80999947)
        self.set_interaction(3, 22, Aij=-144.3999939)
        self.set_interaction(3, 23, Aij=-231.8999939)
        self.set_interaction(3, 24, Aij=3.0)
        self.set_interaction(3, 25, Aij=187.0)
        self.set_interaction(3, 26, Aij=168.0)
        self.set_interaction(3, 27, Aij=194.8999939)
        self.set_interaction(3, 28, Aij=52.06999969)
        self.set_interaction(3, 29, Aij=-10.43000031)
        self.set_interaction(3, 30, Aij=-64.69000244)
        self.set_interaction(3, 31, Aij=210.69999695)
        self.set_interaction(3, 32, Aij=113.30000305)
        self.set_interaction(3, 33, Aij=261.29998779)
        self.set_interaction(3, 35, Aij=169.8999939)
        self.set_interaction(3, 36, Aij=383.8999939)
        self.set_interaction(3, 37, Aij=-259.1000061)
        self.set_interaction(3, 38, Aij=389.29998779)
        self.set_interaction(3, 39, Aij=245.6000061)
        self.set_interaction(3, 40, Aij=22.67000008)
        self.set_interaction(3, 41, Aij=103.5)
        self.set_interaction(3, 42, Aij=-432.29998779)
        self.set_interaction(3, 43, Aij=238.8999939)
        self.set_interaction(3, 44, Aij=30.04000092)
        self.set_interaction(3, 45, Aij=-88.11000061)
        self.set_interaction(3, 47, Aij=333.29998779)
        self.set_interaction(3, 49, Aij=-59.58000183)
        self.set_interaction(3, 50, Aij=-39.15999985)
        self.set_interaction(3, 52, Aij=-11.11999989)
        self.set_interaction(3, 53, Aij=636.09997559)
        self.set_interaction(3, 54, Aij=537.40002441)
        self.set_interaction(3, 55, Aij=25.77000046)
        self.set_interaction(3, 56, Aij=-11.11999989)
        self.set_interaction(3, 57, Aij=32.13999939)
        self.set_interaction(3, 58, Aij=636.09997559)
        self.set_interaction(3, 59, Aij=-11.11999989)
        self.set_interaction(3, 60, Aij=537.40002441)
        self.set_interaction(3, 61, Aij=636.09997559)
        self.set_interaction(3, 62, Aij=-11.11999989)
        self.set_interaction(3, 63, Aij=537.40002441)
        self.set_interaction(3, 64, Aij=636.09997559)
        self.set_interaction(3, 65, Aij=537.40002441)
        self.set_interaction(3, 66, Aij=-11.11999989)
        self.set_interaction(3, 67, Aij=-11.11999989)
        self.set_interaction(3, 68, Aij=-11.11999989)
        self.set_interaction(3, 69, Aij=636.09997559)
        self.set_interaction(4, 1, Aij=-69.69999695)
        self.set_interaction(4, 2, Aij=-113.59999847)
        self.set_interaction(4, 3, Aij=-146.80000305)
        self.set_interaction(4, 5, Aij=803.20001221)
        self.set_interaction(4, 6, Aij=603.29998779)
        self.set_interaction(4, 7, Aij=5695.0)
        self.set_interaction(4, 8, Aij=884.90002441)
        self.set_interaction(4, 9, Aij=-52.09999847)
        self.set_interaction(4, 10, Aij=586.79998779)
        self.set_interaction(4, 11, Aij=5688.0)
        self.set_interaction(4, 12, Aij=197.80000305)
        self.set_interaction(4, 13, Aij=213.1000061)
        self.set_interaction(4, 14, Aij=19.02000046)
        self.set_interaction(4, 15, Aij=-49.29000092)
        self.set_interaction(4, 16, Aij=23.5)
        self.set_interaction(4, 17, Aij=664.20001221)
        self.set_interaction(4, 18, Aij=52.79999924)
        self.set_interaction(4, 19, Aij=6096.0)
        self.set_interaction(4, 20, Aij=872.29998779)
        self.set_interaction(4, 21, Aij=-114.09999847)
        self.set_interaction(4, 22, Aij=-111.0)
        self.set_interaction(4, 23, Aij=-80.25)
        self.set_interaction(4, 24, Aij=-141.30000305)
        self.set_interaction(4, 25, Aij=-211.0)
        self.set_interaction(4, 26, Aij=3629.0)
        self.set_interaction(4, 27, Aij=4448.0)
        self.set_interaction(4, 28, Aij=-9.45100021)
        self.set_interaction(4, 29, Aij=393.6000061)
        self.set_interaction(4, 30, Aij=48.49000168)
        self.set_interaction(4, 31, Aij=4975.0)
        self.set_interaction(4, 32, Aij=259.0)
        self.set_interaction(4, 33, Aij=210.0)
        self.set_interaction(4, 35, Aij=4284.0)
        self.set_interaction(4, 36, Aij=-119.19999695)
        self.set_interaction(4, 37, Aij=-282.5)
        self.set_interaction(4, 38, Aij=101.40000153)
        self.set_interaction(4, 39, Aij=5629.0)
        self.set_interaction(4, 41, Aij=69.26000214)
        self.set_interaction(4, 44, Aij=46.38000107)
        self.set_interaction(4, 47, Aij=421.8999939)
        self.set_interaction(4, 49, Aij=-203.6000061)
        self.set_interaction(4, 50, Aij=184.8999939)
        self.set_interaction(4, 52, Aij=-69.69999695)
        self.set_interaction(4, 53, Aij=803.20001221)
        self.set_interaction(4, 54, Aij=872.29998779)
        self.set_interaction(4, 55, Aij=-52.09999847)
        self.set_interaction(4, 56, Aij=-69.69999695)
        self.set_interaction(4, 57, Aij=213.1000061)
        self.set_interaction(4, 58, Aij=803.20001221)
        self.set_interaction(4, 59, Aij=-69.69999695)
        self.set_interaction(4, 60, Aij=872.29998779)
        self.set_interaction(4, 61, Aij=803.20001221)
        self.set_interaction(4, 62, Aij=-69.69999695)
        self.set_interaction(4, 63, Aij=872.29998779)
        self.set_interaction(4, 64, Aij=803.20001221)
        self.set_interaction(4, 65, Aij=872.29998779)
        self.set_interaction(4, 66, Aij=-69.69999695)
        self.set_interaction(4, 67, Aij=-69.69999695)
        self.set_interaction(4, 68, Aij=-69.69999695)
        self.set_interaction(4, 69, Aij=803.20001221)
        self.set_interaction(5, 1, Aij=156.3999939)
        self.set_interaction(5, 2, Aij=457.0)
        self.set_interaction(5, 3, Aij=89.59999847)
        self.set_interaction(5, 4, Aij=25.81999969)
        self.set_interaction(5, 6, Aij=-137.1000061)
        self.set_interaction(5, 7, Aij=265.86999512)
        self.set_interaction(5, 8, Aij=-259.70001221)
        self.set_interaction(5, 9, Aij=84.0)
        self.set_interaction(5, 10, Aij=-203.6000061)
        self.set_interaction(5, 11, Aij=101.09999847)
        self.set_interaction(5, 12, Aij=267.79998779)
        self.set_interaction(5, 13, Aij=28.05999947)
        self.set_interaction(5, 14, Aij=8.6420002)
        self.set_interaction(5, 15, Aij=42.70000076)
        self.set_interaction(5, 16, Aij=-323.0)
        self.set_interaction(5, 17, Aij=-52.38999939)
        self.set_interaction(5, 18, Aij=170.0)
        self.set_interaction(5, 19, Aij=6.71199989)
        self.set_interaction(5, 20, Aij=224.38999939)
        self.set_interaction(5, 21, Aij=75.62000275)
        self.set_interaction(5, 22, Aij=65.27999878)
        self.set_interaction(5, 23, Aij=-98.12000275)
        self.set_interaction(5, 24, Aij=143.1000061)
        self.set_interaction(5, 25, Aij=123.5)
        self.set_interaction(5, 26, Aij=256.5)
        self.set_interaction(5, 27, Aij=157.1000061)
        self.set_interaction(5, 28, Aij=488.8999939)
        self.set_interaction(5, 29, Aij=147.5)
        self.set_interaction(5, 30, Aij=-120.5)
        self.set_interaction(5, 31, Aij=-318.8999939)
        self.set_interaction(5, 32, Aij=313.5)
        self.set_interaction(5, 33, Aij=202.1000061)
        self.set_interaction(5, 34, Aij=727.79998779)
        self.set_interaction(5, 35, Aij=-202.1000061)
        self.set_interaction(5, 36, Aij=74.26999664)
        self.set_interaction(5, 37, Aij=225.80000305)
        self.set_interaction(5, 38, Aij=44.77999878)
        self.set_interaction(5, 39, Aij=-143.8999939)
        self.set_interaction(5, 41, Aij=190.30000305)
        self.set_interaction(5, 42, Aij=-817.70001221)
        self.set_interaction(5, 44, Aij=-504.20001221)
        self.set_interaction(5, 45, Aij=72.95999908)
        self.set_interaction(5, 46, Aij=-382.70001221)
        self.set_interaction(5, 47, Aij=-248.30000305)
        self.set_interaction(5, 49, Aij=104.69999695)
        self.set_interaction(5, 50, Aij=57.65000153)
        self.set_interaction(5, 52, Aij=156.3999939)
        self.set_interaction(5, 54, Aij=199.0)
        self.set_interaction(5, 55, Aij=84.0)
        self.set_interaction(5, 56, Aij=156.3999939)
        self.set_interaction(5, 57, Aij=28.05999947)
        self.set_interaction(5, 59, Aij=156.3999939)
        self.set_interaction(5, 60, Aij=199.0)
        self.set_interaction(5, 62, Aij=156.3999939)
        self.set_interaction(5, 63, Aij=199.0)
        self.set_interaction(5, 65, Aij=224.39)
        self.set_interaction(5, 66, Aij=156.3999939)
        self.set_interaction(5, 67, Aij=156.3999939)
        self.set_interaction(5, 68, Aij=156.3999939)
        self.set_interaction(5, 70, Aij=-843.09002686)
        self.set_interaction(5, 71, Aij=37.631)
        self.set_interaction(5, 72, Aij=-330.28)
        self.set_interaction(5, 73, Aij=1789.0)
        self.set_interaction(5, 74, Aij=221.38)
        self.set_interaction(5, 75, Aij=-77.526)
        self.set_interaction(6, 1, Aij=16.51000023)
        self.set_interaction(6, 2, Aij=-12.52000046)
        self.set_interaction(6, 3, Aij=-50.0)
        self.set_interaction(6, 4, Aij=-44.5)
        self.set_interaction(6, 5, Aij=249.1000061)
        self.set_interaction(6, 7, Aij=-181.0)
        self.set_interaction(6, 8, Aij=-101.69999695)
        self.set_interaction(6, 9, Aij=23.38999939)
        self.set_interaction(6, 10, Aij=306.3999939)
        self.set_interaction(6, 11, Aij=-10.72000027)
        self.set_interaction(6, 12, Aij=179.69999695)
        self.set_interaction(6, 13, Aij=-128.6000061)
        self.set_interaction(6, 14, Aij=359.29998779)
        self.set_interaction(6, 15, Aij=-20.97999954)
        self.set_interaction(6, 16, Aij=53.90000153)
        self.set_interaction(6, 17, Aij=489.70001221)
        self.set_interaction(6, 18, Aij=580.5)
        self.set_interaction(6, 19, Aij=53.27999878)
        self.set_interaction(6, 20, Aij=-202.0)
        self.set_interaction(6, 21, Aij=-38.31999969)
        self.set_interaction(6, 22, Aij=-102.5)
        self.set_interaction(6, 23, Aij=-139.3999939)
        self.set_interaction(6, 24, Aij=-44.75999832)
        self.set_interaction(6, 25, Aij=-28.25)
        self.set_interaction(6, 26, Aij=75.13999939)
        self.set_interaction(6, 28, Aij=-31.09000015)
        self.set_interaction(6, 29, Aij=17.5)
        self.set_interaction(6, 31, Aij=-119.19999695)
        self.set_interaction(6, 32, Aij=212.1000061)
        self.set_interaction(6, 33, Aij=106.30000305)
        self.set_interaction(6, 35, Aij=-399.29998779)
        self.set_interaction(6, 36, Aij=-5.22399998)
        self.set_interaction(6, 37, Aij=33.47000122)
        self.set_interaction(6, 38, Aij=-48.25)
        self.set_interaction(6, 39, Aij=-172.3999939)
        self.set_interaction(6, 41, Aij=165.69999695)
        self.set_interaction(6, 45, Aij=-52.09999847)
        self.set_interaction(6, 48, Aij=37.63000107)
        self.set_interaction(6, 49, Aij=-59.40000153)
        self.set_interaction(6, 50, Aij=-46.00999832)
        self.set_interaction(6, 52, Aij=16.51000023)
        self.set_interaction(6, 53, Aij=249.1000061)
        self.set_interaction(6, 54, Aij=-202.0)
        self.set_interaction(6, 55, Aij=23.38999939)
        self.set_interaction(6, 56, Aij=16.51000023)
        self.set_interaction(6, 57, Aij=-128.6000061)
        self.set_interaction(6, 58, Aij=249.1000061)
        self.set_interaction(6, 59, Aij=16.51000023)
        self.set_interaction(6, 60, Aij=-202.0)
        self.set_interaction(6, 61, Aij=249.1000061)
        self.set_interaction(6, 62, Aij=16.51000023)
        self.set_interaction(6, 63, Aij=-202.0)
        self.set_interaction(6, 64, Aij=249.1000061)
        self.set_interaction(6, 65, Aij=-202.0)
        self.set_interaction(6, 66, Aij=16.51000023)
        self.set_interaction(6, 67, Aij=16.51000023)
        self.set_interaction(6, 68, Aij=16.51000023)
        self.set_interaction(6, 69, Aij=249.1000061)
        self.set_interaction(7, 1, Aij=300.0)
        self.set_interaction(7, 2, Aij=496.1000061)
        self.set_interaction(7, 3, Aij=362.29998779)
        self.set_interaction(7, 4, Aij=377.6000061)
        self.set_interaction(7, 5, Aij=-467.42001343)
        self.set_interaction(7, 6, Aij=289.6000061)
        self.set_interaction(7, 8, Aij=324.5)
        self.set_interaction(7, 9, Aij=-195.3999939)
        self.set_interaction(7, 10, Aij=-116.0)
        self.set_interaction(7, 11, Aij=72.87000275)
        self.set_interaction(7, 13, Aij=540.5)
        self.set_interaction(7, 14, Aij=48.88999939)
        self.set_interaction(7, 15, Aij=168.0)
        self.set_interaction(7, 16, Aij=304.0)
        self.set_interaction(7, 17, Aij=243.19999695)
        self.set_interaction(7, 18, Aij=459.0)
        self.set_interaction(7, 19, Aij=112.59999847)
        self.set_interaction(7, 20, Aij=-69.29000092)
        self.set_interaction(7, 21, Aij=325.3999939)
        self.set_interaction(7, 22, Aij=370.3999939)
        self.set_interaction(7, 23, Aij=353.70001221)
        self.set_interaction(7, 24, Aij=497.5)
        self.set_interaction(7, 25, Aij=133.8999939)
        self.set_interaction(7, 26, Aij=220.6000061)
        self.set_interaction(7, 27, Aij=399.5)
        self.set_interaction(7, 28, Aij=887.09997559)
        self.set_interaction(7, 30, Aij=188.0)
        self.set_interaction(7, 31, Aij=12.72000027)
        self.set_interaction(7, 35, Aij=-139.0)
        self.set_interaction(7, 36, Aij=160.80000305)
        self.set_interaction(7, 39, Aij=319.0)
        self.set_interaction(7, 41, Aij=-197.5)
        self.set_interaction(7, 42, Aij=-363.79998779)
        self.set_interaction(7, 44, Aij=-452.20001221)
        self.set_interaction(7, 46, Aij=835.59997559)
        self.set_interaction(7, 47, Aij=139.6000061)
        self.set_interaction(7, 49, Aij=407.8999939)
        self.set_interaction(7, 52, Aij=541.29998779)
        self.set_interaction(7, 53, Aij=-959.0)
        self.set_interaction(7, 54, Aij=-530.0)
        self.set_interaction(7, 55, Aij=-471.0)
        self.set_interaction(7, 56, Aij=300.0)
        self.set_interaction(7, 57, Aij=193.1000061)
        self.set_interaction(7, 58, Aij=171.30000305)
        self.set_interaction(7, 59, Aij=300.0)
        self.set_interaction(7, 60, Aij=64.59999847)
        self.set_interaction(7, 61, Aij=224.3999939)
        self.set_interaction(7, 62, Aij=304.0)
        self.set_interaction(7, 63, Aij=86.0)
        self.set_interaction(7, 64, Aij=-467.42001343)
        self.set_interaction(7, 65, Aij=-69.29000092)
        self.set_interaction(7, 66, Aij=162.30000305)
        self.set_interaction(7, 67, Aij=362.1000061)
        self.set_interaction(7, 68, Aij=-89.70999908)
        self.set_interaction(7, 69, Aij=-153.0)
        self.set_interaction(7, 70, Aij=-632.54998779)
        self.set_interaction(7, 71, Aij=142.65)
        self.set_interaction(7, 72, Aij=-341.18)
        self.set_interaction(7, 73, Aij=-329.81)
        self.set_interaction(7, 74, Aij=-7.2937)
        self.set_interaction(7, 75, Aij=76.211)
        self.set_interaction(8, 1, Aij=275.79998779)
        self.set_interaction(8, 2, Aij=217.5)
        self.set_interaction(8, 3, Aij=25.34000015)
        self.set_interaction(8, 4, Aij=244.19999695)
        self.set_interaction(8, 5, Aij=-451.6000061)
        self.set_interaction(8, 6, Aij=-265.20001221)
        self.set_interaction(8, 7, Aij=-601.79998779)
        self.set_interaction(8, 9, Aij=-356.1000061)
        self.set_interaction(8, 10, Aij=-271.1000061)
        self.set_interaction(8, 11, Aij=-449.3999939)
        self.set_interaction(8, 13, Aij=-162.8999939)
        self.set_interaction(8, 17, Aij=119.90000153)
        self.set_interaction(8, 18, Aij=-305.5)
        self.set_interaction(8, 20, Aij=408.8999939)
        self.set_interaction(8, 24, Aij=1827.0)
        self.set_interaction(8, 25, Aij=6915.0)
        self.set_interaction(8, 28, Aij=8484.0)
        self.set_interaction(8, 31, Aij=-687.09997559)
        self.set_interaction(8, 41, Aij=-494.20001221)
        self.set_interaction(8, 44, Aij=-659.0)
        self.set_interaction(8, 50, Aij=1005.0)
        self.set_interaction(8, 52, Aij=275.79998779)
        self.set_interaction(8, 53, Aij=-451.6000061)
        self.set_interaction(8, 54, Aij=408.8999939)
        self.set_interaction(8, 55, Aij=-356.1000061)
        self.set_interaction(8, 56, Aij=275.79998779)
        self.set_interaction(8, 57, Aij=-162.8999939)
        self.set_interaction(8, 58, Aij=-451.6000061)
        self.set_interaction(8, 59, Aij=275.79998779)
        self.set_interaction(8, 60, Aij=408.8999939)
        self.set_interaction(8, 61, Aij=451.6000061)
        self.set_interaction(8, 62, Aij=275.79998779)
        self.set_interaction(8, 63, Aij=408.8999939)
        self.set_interaction(8, 64, Aij=-451.6000061)
        self.set_interaction(8, 65, Aij=408.8999939)
        self.set_interaction(8, 66, Aij=275.79998779)
        self.set_interaction(8, 67, Aij=275.79998779)
        self.set_interaction(8, 68, Aij=275.79998779)
        self.set_interaction(8, 69, Aij=-451.6000061)
        self.set_interaction(9, 1, Aij=26.76000023)
        self.set_interaction(9, 2, Aij=42.91999817)
        self.set_interaction(9, 3, Aij=140.1000061)
        self.set_interaction(9, 4, Aij=365.79998779)
        self.set_interaction(9, 5, Aij=164.5)
        self.set_interaction(9, 6, Aij=108.69999695)
        self.set_interaction(9, 7, Aij=472.5)
        self.set_interaction(9, 8, Aij=-133.1000061)
        self.set_interaction(9, 10, Aij=-37.36000061)
        self.set_interaction(9, 11, Aij=-213.69999695)
        self.set_interaction(9, 12, Aij=-190.3999939)
        self.set_interaction(9, 13, Aij=-103.59999847)
        self.set_interaction(9, 15, Aij=-174.19999695)
        self.set_interaction(9, 16, Aij=-169.0)
        self.set_interaction(9, 17, Aij=6201.0)
        self.set_interaction(9, 18, Aij=7.34100008)
        self.set_interaction(9, 19, Aij=481.70001221)
        self.set_interaction(9, 20, Aij=669.40002441)
        self.set_interaction(9, 21, Aij=-191.69999695)
        self.set_interaction(9, 22, Aij=-130.30000305)
        self.set_interaction(9, 23, Aij=-354.6000061)
        self.set_interaction(9, 24, Aij=-39.20000076)
        self.set_interaction(9, 25, Aij=-119.80000305)
        self.set_interaction(9, 26, Aij=137.5)
        self.set_interaction(9, 27, Aij=548.5)
        self.set_interaction(9, 28, Aij=216.1000061)
        self.set_interaction(9, 29, Aij=-46.27999878)
        self.set_interaction(9, 30, Aij=-163.69999695)
        self.set_interaction(9, 31, Aij=71.45999908)
        self.set_interaction(9, 32, Aij=53.59000015)
        self.set_interaction(9, 33, Aij=245.19999695)
        self.set_interaction(9, 34, Aij=-246.6000061)
        self.set_interaction(9, 35, Aij=-44.58000183)
        self.set_interaction(9, 37, Aij=-34.56999969)
        self.set_interaction(9, 39, Aij=-61.70000076)
        self.set_interaction(9, 41, Aij=-18.79999924)
        self.set_interaction(9, 42, Aij=-588.90002441)
        self.set_interaction(9, 47, Aij=37.54000092)
        self.set_interaction(9, 50, Aij=-162.6000061)
        self.set_interaction(9, 52, Aij=26.76000023)
        self.set_interaction(9, 53, Aij=164.5)
        self.set_interaction(9, 54, Aij=669.40002441)
        self.set_interaction(9, 56, Aij=26.76000023)
        self.set_interaction(9, 57, Aij=-103.59999847)
        self.set_interaction(9, 58, Aij=164.5)
        self.set_interaction(9, 59, Aij=26.76000023)
        self.set_interaction(9, 60, Aij=669.40002441)
        self.set_interaction(9, 61, Aij=164.5)
        self.set_interaction(9, 62, Aij=26.76000023)
        self.set_interaction(9, 63, Aij=669.40002441)
        self.set_interaction(9, 64, Aij=164.5)
        self.set_interaction(9, 65, Aij=669.40002441)
        self.set_interaction(9, 66, Aij=26.76000023)
        self.set_interaction(9, 67, Aij=26.76000023)
        self.set_interaction(9, 68, Aij=26.76000023)
        self.set_interaction(9, 69, Aij=164.5)
        self.set_interaction(9, 71, Aij=-197.93)
        self.set_interaction(9, 72, Aij=-350.58)
        self.set_interaction(9, 73, Aij=252.05)
        self.set_interaction(9, 74, Aij=-286.39)
        self.set_interaction(9, 75, Aij=-3.8839)
        self.set_interaction(10, 1, Aij=505.70001221)
        self.set_interaction(10, 2, Aij=56.29999924)
        self.set_interaction(10, 3, Aij=23.38999939)
        self.set_interaction(10, 4, Aij=106.0)
        self.set_interaction(10, 5, Aij=529.0)
        self.set_interaction(10, 6, Aij=-340.20001221)
        self.set_interaction(10, 7, Aij=480.79998779)
        self.set_interaction(10, 8, Aij=-155.6000061)
        self.set_interaction(10, 9, Aij=128.0)
        self.set_interaction(10, 11, Aij=-110.30000305)
        self.set_interaction(10, 12, Aij=766.0)
        self.set_interaction(10, 13, Aij=304.1000061)
        self.set_interaction(10, 20, Aij=497.5)
        self.set_interaction(10, 21, Aij=751.90002441)
        self.set_interaction(10, 22, Aij=67.51999664)
        self.set_interaction(10, 23, Aij=-483.70001221)
        self.set_interaction(10, 32, Aij=117.0)
        self.set_interaction(10, 36, Aij=-339.20001221)
        self.set_interaction(10, 37, Aij=172.3999939)
        self.set_interaction(10, 39, Aij=-268.79998779)
        self.set_interaction(10, 41, Aij=-275.5)
        self.set_interaction(10, 52, Aij=505.70001221)
        self.set_interaction(10, 53, Aij=529.0)
        self.set_interaction(10, 54, Aij=497.5)
        self.set_interaction(10, 55, Aij=128.0)
        self.set_interaction(10, 56, Aij=505.70001221)
        self.set_interaction(10, 57, Aij=304.1000061)
        self.set_interaction(10, 58, Aij=529.0)
        self.set_interaction(10, 59, Aij=505.70001221)
        self.set_interaction(10, 60, Aij=497.5)
        self.set_interaction(10, 61, Aij=529.0)
        self.set_interaction(10, 62, Aij=505.70001221)
        self.set_interaction(10, 63, Aij=497.5)
        self.set_interaction(10, 64, Aij=529.0)
        self.set_interaction(10, 65, Aij=497.5)
        self.set_interaction(10, 66, Aij=505.70001221)
        self.set_interaction(10, 67, Aij=505.70001221)
        self.set_interaction(10, 68, Aij=505.70001221)
        self.set_interaction(10, 69, Aij=529.0)
        self.set_interaction(10, 71, Aij=402.0)
        self.set_interaction(10, 72, Aij=-387.63)
        self.set_interaction(10, 73, Aij=12274.0)
        self.set_interaction(10, 74, Aij=-18.524)
        self.set_interaction(10, 75, Aij=308.97)
        self.set_interaction(11, 1, Aij=114.80000305)
        self.set_interaction(11, 2, Aij=132.1000061)
        self.set_interaction(11, 3, Aij=85.83999634)
        self.set_interaction(11, 4, Aij=-170.0)
        self.set_interaction(11, 5, Aij=245.3999939)
        self.set_interaction(11, 6, Aij=249.6000061)
        self.set_interaction(11, 7, Aij=200.80000305)
        self.set_interaction(11, 8, Aij=-36.72000122)
        self.set_interaction(11, 9, Aij=372.20001221)
        self.set_interaction(11, 10, Aij=185.1000061)
        self.set_interaction(11, 12, Aij=-241.80000305)
        self.set_interaction(11, 13, Aij=-235.69999695)
        self.set_interaction(11, 15, Aij=-73.5)
        self.set_interaction(11, 16, Aij=-196.69999695)
        self.set_interaction(11, 17, Aij=475.5)
        self.set_interaction(11, 19, Aij=494.6000061)
        self.set_interaction(11, 20, Aij=660.20001221)
        self.set_interaction(11, 21, Aij=-34.74000168)
        self.set_interaction(11, 22, Aij=108.90000153)
        self.set_interaction(11, 23, Aij=-209.69999695)
        self.set_interaction(11, 24, Aij=54.56999969)
        self.set_interaction(11, 25, Aij=442.3999939)
        self.set_interaction(11, 26, Aij=-81.12999725)
        self.set_interaction(11, 28, Aij=183.0)
        self.set_interaction(11, 30, Aij=202.30000305)
        self.set_interaction(11, 31, Aij=-101.69999695)
        self.set_interaction(11, 32, Aij=148.30000305)
        self.set_interaction(11, 33, Aij=18.87999916)
        self.set_interaction(11, 35, Aij=52.08000183)
        self.set_interaction(11, 36, Aij=-28.61000061)
        self.set_interaction(11, 37, Aij=-275.20001221)
        self.set_interaction(11, 39, Aij=85.33000183)
        self.set_interaction(11, 41, Aij=560.20001221)
        self.set_interaction(11, 47, Aij=151.80000305)
        self.set_interaction(11, 52, Aij=114.80000305)
        self.set_interaction(11, 53, Aij=245.3999939)
        self.set_interaction(11, 54, Aij=660.20001221)
        self.set_interaction(11, 55, Aij=372.20001221)
        self.set_interaction(11, 56, Aij=114.80000305)
        self.set_interaction(11, 57, Aij=-235.69999695)
        self.set_interaction(11, 58, Aij=245.3999939)
        self.set_interaction(11, 59, Aij=114.80000305)
        self.set_interaction(11, 60, Aij=660.20001221)
        self.set_interaction(11, 61, Aij=245.3999939)
        self.set_interaction(11, 62, Aij=114.80000305)
        self.set_interaction(11, 63, Aij=660.20001221)
        self.set_interaction(11, 64, Aij=245.3999939)
        self.set_interaction(11, 65, Aij=660.20001221)
        self.set_interaction(11, 66, Aij=114.80000305)
        self.set_interaction(11, 67, Aij=114.80000305)
        self.set_interaction(11, 68, Aij=114.80000305)
        self.set_interaction(11, 69, Aij=245.3999939)
        self.set_interaction(11, 71, Aij=1273.8)
        self.set_interaction(11, 72, Aij=928.33)
        self.set_interaction(11, 73, Aij=416.0)
        self.set_interaction(11, 74, Aij=-252.22)
        self.set_interaction(11, 75, Aij=426.52)
        self.set_interaction(12, 1, Aij=329.29998779)
        self.set_interaction(12, 2, Aij=110.40000153)
        self.set_interaction(12, 3, Aij=18.12000084)
        self.set_interaction(12, 4, Aij=428.0)
        self.set_interaction(12, 5, Aij=139.3999939)
        self.set_interaction(12, 6, Aij=227.80000305)
        self.set_interaction(12, 9, Aij=385.3999939)
        self.set_interaction(12, 10, Aij=-236.5)
        self.set_interaction(12, 11, Aij=1167.0)
        self.set_interaction(12, 13, Aij=-234.0)
        self.set_interaction(12, 18, Aij=-233.3999939)
        self.set_interaction(12, 19, Aij=-47.25)
        self.set_interaction(12, 20, Aij=-268.1000061)
        self.set_interaction(12, 23, Aij=-126.19999695)
        self.set_interaction(12, 24, Aij=179.69999695)
        self.set_interaction(12, 25, Aij=24.28000069)
        self.set_interaction(12, 29, Aij=103.90000153)
        self.set_interaction(12, 37, Aij=-11.39999962)
        self.set_interaction(12, 39, Aij=308.8999939)
        self.set_interaction(12, 41, Aij=-122.30000305)
        self.set_interaction(12, 52, Aij=329.29998779)
        self.set_interaction(12, 53, Aij=139.3999939)
        self.set_interaction(12, 54, Aij=-268.1000061)
        self.set_interaction(12, 55, Aij=385.3999939)
        self.set_interaction(12, 56, Aij=329.29998779)
        self.set_interaction(12, 57, Aij=-234.0)
        self.set_interaction(12, 58, Aij=139.3999939)
        self.set_interaction(12, 59, Aij=329.29998779)
        self.set_interaction(12, 60, Aij=-268.1000061)
        self.set_interaction(12, 61, Aij=139.3999939)
        self.set_interaction(12, 62, Aij=329.29998779)
        self.set_interaction(12, 63, Aij=-268.1000061)
        self.set_interaction(12, 64, Aij=139.3999939)
        self.set_interaction(12, 65, Aij=-268.1000061)
        self.set_interaction(12, 66, Aij=329.29998779)
        self.set_interaction(12, 67, Aij=329.29998779)
        self.set_interaction(12, 68, Aij=329.29998779)
        self.set_interaction(12, 69, Aij=139.3999939)
        self.set_interaction(13, 1, Aij=83.36000061)
        self.set_interaction(13, 2, Aij=26.51000023)
        self.set_interaction(13, 3, Aij=52.13000107)
        self.set_interaction(13, 4, Aij=65.69000244)
        self.set_interaction(13, 5, Aij=237.69999695)
        self.set_interaction(13, 6, Aij=238.3999939)
        self.set_interaction(13, 7, Aij=-314.70001221)
        self.set_interaction(13, 8, Aij=-178.5)
        self.set_interaction(13, 9, Aij=191.1000061)
        self.set_interaction(13, 10, Aij=-7.83799982)
        self.set_interaction(13, 11, Aij=461.29998779)
        self.set_interaction(13, 12, Aij=457.29998779)
        self.set_interaction(13, 14, Aij=-78.36000061)
        self.set_interaction(13, 15, Aij=251.5)
        self.set_interaction(13, 16, Aij=5422.0)
        self.set_interaction(13, 18, Aij=213.19999695)
        self.set_interaction(13, 19, Aij=-18.51000023)
        self.set_interaction(13, 20, Aij=664.59997559)
        self.set_interaction(13, 21, Aij=301.1000061)
        self.set_interaction(13, 22, Aij=137.80000305)
        self.set_interaction(13, 23, Aij=-154.30000305)
        self.set_interaction(13, 24, Aij=47.66999817)
        self.set_interaction(13, 25, Aij=134.80000305)
        self.set_interaction(13, 26, Aij=95.18000031)
        self.set_interaction(13, 28, Aij=140.8999939)
        self.set_interaction(13, 29, Aij=-8.53800011)
        self.set_interaction(13, 30, Aij=170.1000061)
        self.set_interaction(13, 31, Aij=-20.11000061)
        self.set_interaction(13, 32, Aij=-149.5)
        self.set_interaction(13, 33, Aij=-202.30000305)
        self.set_interaction(13, 35, Aij=128.80000305)
        self.set_interaction(13, 37, Aij=240.19999695)
        self.set_interaction(13, 38, Aij=-274.0)
        self.set_interaction(13, 39, Aij=254.80000305)
        self.set_interaction(13, 41, Aij=417.0)
        self.set_interaction(13, 42, Aij=1338.0)
        self.set_interaction(13, 52, Aij=83.36000061)
        self.set_interaction(13, 53, Aij=237.69999695)
        self.set_interaction(13, 54, Aij=664.59997559)
        self.set_interaction(13, 55, Aij=191.1000061)
        self.set_interaction(13, 56, Aij=83.36000061)
        self.set_interaction(13, 58, Aij=237.69999695)
        self.set_interaction(13, 59, Aij=83.36000061)
        self.set_interaction(13, 60, Aij=664.59997559)
        self.set_interaction(13, 61, Aij=237.69999695)
        self.set_interaction(13, 62, Aij=83.36000061)
        self.set_interaction(13, 63, Aij=664.59997559)
        self.set_interaction(13, 64, Aij=237.69999695)
        self.set_interaction(13, 65, Aij=664.59997559)
        self.set_interaction(13, 66, Aij=83.36000061)
        self.set_interaction(13, 67, Aij=83.36000061)
        self.set_interaction(13, 68, Aij=83.36000061)
        self.set_interaction(13, 69, Aij=237.69999695)
        self.set_interaction(13, 71, Aij=1133.1)
        self.set_interaction(13, 72, Aij=-438.74)
        self.set_interaction(13, 73, Aij=2221.9)
        self.set_interaction(13, 74, Aij=-130.54)
        self.set_interaction(13, 75, Aij=1166.7)
        self.set_interaction(14, 1, Aij=-30.47999954)
        self.set_interaction(14, 2, Aij=1.16299999)
        self.set_interaction(14, 3, Aij=-44.84999847)
        self.set_interaction(14, 4, Aij=296.3999939)
        self.set_interaction(14, 5, Aij=-242.80000305)
        self.set_interaction(14, 6, Aij=-481.70001221)
        self.set_interaction(14, 7, Aij=-330.3999939)
        self.set_interaction(14, 13, Aij=222.1000061)
        self.set_interaction(14, 15, Aij=-107.19999695)
        self.set_interaction(14, 16, Aij=-41.11000061)
        self.set_interaction(14, 17, Aij=-200.69999695)
        self.set_interaction(14, 19, Aij=358.8999939)
        self.set_interaction(14, 21, Aij=-82.91999817)
        self.set_interaction(14, 24, Aij=-99.80999756)
        self.set_interaction(14, 25, Aij=30.04999924)
        self.set_interaction(14, 29, Aij=-70.13999939)
        self.set_interaction(14, 39, Aij=-164.0)
        self.set_interaction(14, 42, Aij=-664.40002441)
        self.set_interaction(14, 43, Aij=275.8999939)
        self.set_interaction(14, 52, Aij=-30.47999954)
        self.set_interaction(14, 53, Aij=-242.80000305)
        self.set_interaction(14, 56, Aij=-30.47999954)
        self.set_interaction(14, 57, Aij=222.1000061)
        self.set_interaction(14, 58, Aij=-242.80000305)
        self.set_interaction(14, 59, Aij=-30.47999954)
        self.set_interaction(14, 61, Aij=-242.80000305)
        self.set_interaction(14, 62, Aij=-30.47999954)
        self.set_interaction(14, 64, Aij=-242.80000305)
        self.set_interaction(14, 66, Aij=-30.47999954)
        self.set_interaction(14, 67, Aij=-30.47999954)
        self.set_interaction(14, 68, Aij=-30.47999954)
        self.set_interaction(14, 69, Aij=-242.80000305)
        self.set_interaction(15, 1, Aij=65.33000183)
        self.set_interaction(15, 2, Aij=-28.70000076)
        self.set_interaction(15, 3, Aij=-22.30999947)
        self.set_interaction(15, 4, Aij=223.0)
        self.set_interaction(15, 5, Aij=-150.0)
        self.set_interaction(15, 6, Aij=-370.29998779)
        self.set_interaction(15, 7, Aij=-448.20001221)
        self.set_interaction(15, 9, Aij=394.6000061)
        self.set_interaction(15, 11, Aij=136.0)
        self.set_interaction(15, 13, Aij=-56.08000183)
        self.set_interaction(15, 14, Aij=127.40000153)
        self.set_interaction(15, 16, Aij=-189.19999695)
        self.set_interaction(15, 19, Aij=147.1000061)
        self.set_interaction(15, 24, Aij=71.23000336)
        self.set_interaction(15, 25, Aij=-18.93000031)
        self.set_interaction(15, 38, Aij=570.90002441)
        self.set_interaction(15, 41, Aij=-38.77000046)
        self.set_interaction(15, 42, Aij=448.1000061)
        self.set_interaction(15, 43, Aij=-1327.0)
        self.set_interaction(15, 52, Aij=65.33000183)
        self.set_interaction(15, 53, Aij=-150.0)
        self.set_interaction(15, 55, Aij=394.6000061)
        self.set_interaction(15, 56, Aij=65.33000183)
        self.set_interaction(15, 57, Aij=-56.08000183)
        self.set_interaction(15, 58, Aij=-150.0)
        self.set_interaction(15, 59, Aij=65.33000183)
        self.set_interaction(15, 61, Aij=-150.0)
        self.set_interaction(15, 62, Aij=65.33000183)
        self.set_interaction(15, 64, Aij=-150.0)
        self.set_interaction(15, 66, Aij=65.33000183)
        self.set_interaction(15, 67, Aij=65.33000183)
        self.set_interaction(15, 68, Aij=65.33000183)
        self.set_interaction(15, 69, Aij=-150.0)
        self.set_interaction(16, 1, Aij=-83.98000336)
        self.set_interaction(16, 2, Aij=-25.37999916)
        self.set_interaction(16, 3, Aij=-223.8999939)
        self.set_interaction(16, 4, Aij=109.90000153)
        self.set_interaction(16, 5, Aij=28.60000038)
        self.set_interaction(16, 6, Aij=-406.79998779)
        self.set_interaction(16, 7, Aij=-598.79998779)
        self.set_interaction(16, 9, Aij=225.30000305)
        self.set_interaction(16, 11, Aij=2889.0)
        self.set_interaction(16, 13, Aij=-194.1000061)
        self.set_interaction(16, 14, Aij=38.88999939)
        self.set_interaction(16, 15, Aij=865.90002441)
        self.set_interaction(16, 22, Aij=-73.84999847)
        self.set_interaction(16, 23, Aij=-352.8999939)
        self.set_interaction(16, 24, Aij=-262.0)
        self.set_interaction(16, 25, Aij=-181.8999939)
        self.set_interaction(16, 35, Aij=243.1000061)
        self.set_interaction(16, 38, Aij=-196.30000305)
        self.set_interaction(16, 39, Aij=22.04999924)
        self.set_interaction(16, 52, Aij=-83.98000336)
        self.set_interaction(16, 53, Aij=28.60000038)
        self.set_interaction(16, 55, Aij=225.30000305)
        self.set_interaction(16, 56, Aij=-83.98000336)
        self.set_interaction(16, 57, Aij=-194.1000061)
        self.set_interaction(16, 58, Aij=28.60000038)
        self.set_interaction(16, 59, Aij=-83.98000336)
        self.set_interaction(16, 61, Aij=28.60000038)
        self.set_interaction(16, 62, Aij=-83.98000336)
        self.set_interaction(16, 64, Aij=28.60000038)
        self.set_interaction(16, 66, Aij=-83.98000336)
        self.set_interaction(16, 67, Aij=-83.98000336)
        self.set_interaction(16, 68, Aij=-83.98000336)
        self.set_interaction(16, 69, Aij=28.60000038)
        self.set_interaction(17, 1, Aij=1139.0)
        self.set_interaction(17, 2, Aij=2000.0)
        self.set_interaction(17, 3, Aij=247.5)
        self.set_interaction(17, 4, Aij=762.79998779)
        self.set_interaction(17, 5, Aij=-17.39999962)
        self.set_interaction(17, 6, Aij=-118.09999847)
        self.set_interaction(17, 7, Aij=-341.6000061)
        self.set_interaction(17, 8, Aij=-253.1000061)
        self.set_interaction(17, 9, Aij=-450.29998779)
        self.set_interaction(17, 11, Aij=-294.79998779)
        self.set_interaction(17, 14, Aij=-15.06999969)
        self.set_interaction(17, 18, Aij=89.69999695)
        self.set_interaction(17, 19, Aij=-281.6000061)
        self.set_interaction(17, 20, Aij=-396.0)
        self.set_interaction(17, 21, Aij=287.0)
        self.set_interaction(17, 22, Aij=-111.0)
        self.set_interaction(17, 24, Aij=882.0)
        self.set_interaction(17, 25, Aij=617.5)
        self.set_interaction(17, 27, Aij=-139.30000305)
        self.set_interaction(17, 31, Aij=0.1004)
        self.set_interaction(17, 39, Aij=-334.3999939)
        self.set_interaction(17, 41, Aij=-89.41999817)
        self.set_interaction(17, 52, Aij=1139.0)
        self.set_interaction(17, 53, Aij=-17.39999962)
        self.set_interaction(17, 54, Aij=-396.0)
        self.set_interaction(17, 55, Aij=-450.29998779)
        self.set_interaction(17, 56, Aij=1139.0)
        self.set_interaction(17, 58, Aij=-17.39999962)
        self.set_interaction(17, 59, Aij=1139.0)
        self.set_interaction(17, 60, Aij=-396.0)
        self.set_interaction(17, 61, Aij=-17.39999962)
        self.set_interaction(17, 62, Aij=1139.0)
        self.set_interaction(17, 63, Aij=-396.0)
        self.set_interaction(17, 64, Aij=-17.39999962)
        self.set_interaction(17, 65, Aij=-396.0)
        self.set_interaction(17, 66, Aij=1139.0)
        self.set_interaction(17, 67, Aij=1139.0)
        self.set_interaction(17, 68, Aij=1139.0)
        self.set_interaction(17, 69, Aij=-17.39999962)
        self.set_interaction(18, 1, Aij=-101.59999847)
        self.set_interaction(18, 2, Aij=-47.63000107)
        self.set_interaction(18, 3, Aij=31.87000084)
        self.set_interaction(18, 4, Aij=49.79999924)
        self.set_interaction(18, 5, Aij=-132.30000305)
        self.set_interaction(18, 6, Aij=-378.20001221)
        self.set_interaction(18, 7, Aij=-332.8999939)
        self.set_interaction(18, 8, Aij=-341.6000061)
        self.set_interaction(18, 9, Aij=29.10000038)
        self.set_interaction(18, 12, Aij=554.40002441)
        self.set_interaction(18, 13, Aij=-156.1000061)
        self.set_interaction(18, 17, Aij=117.40000153)
        self.set_interaction(18, 19, Aij=-169.69999695)
        self.set_interaction(18, 20, Aij=-153.69999695)
        self.set_interaction(18, 22, Aij=-351.6000061)
        self.set_interaction(18, 23, Aij=-114.69999695)
        self.set_interaction(18, 24, Aij=-205.30000305)
        self.set_interaction(18, 27, Aij=2845.0)
        self.set_interaction(18, 33, Aij=-60.77999878)
        self.set_interaction(18, 37, Aij=160.69999695)
        self.set_interaction(18, 38, Aij=-158.80000305)
        self.set_interaction(18, 50, Aij=-136.6000061)
        self.set_interaction(18, 52, Aij=-101.59999847)
        self.set_interaction(18, 53, Aij=-132.30000305)
        self.set_interaction(18, 54, Aij=-153.69999695)
        self.set_interaction(18, 55, Aij=29.10000038)
        self.set_interaction(18, 56, Aij=-101.59999847)
        self.set_interaction(18, 57, Aij=-156.1000061)
        self.set_interaction(18, 58, Aij=-132.30000305)
        self.set_interaction(18, 59, Aij=-101.59999847)
        self.set_interaction(18, 60, Aij=-153.69999695)
        self.set_interaction(18, 61, Aij=-132.30000305)
        self.set_interaction(18, 62, Aij=-101.59999847)
        self.set_interaction(18, 63, Aij=-153.69999695)
        self.set_interaction(18, 64, Aij=-132.30000305)
        self.set_interaction(18, 65, Aij=-153.69999695)
        self.set_interaction(18, 66, Aij=-101.59999847)
        self.set_interaction(18, 67, Aij=-101.59999847)
        self.set_interaction(18, 68, Aij=-101.59999847)
        self.set_interaction(18, 69, Aij=-132.30000305)
        self.set_interaction(19, 1, Aij=24.81999969)
        self.set_interaction(19, 2, Aij=-40.61999893)
        self.set_interaction(19, 3, Aij=-22.96999931)
        self.set_interaction(19, 4, Aij=-138.3999939)
        self.set_interaction(19, 5, Aij=185.3999939)
        self.set_interaction(19, 6, Aij=162.6000061)
        self.set_interaction(19, 7, Aij=242.80000305)
        self.set_interaction(19, 9, Aij=-287.5)
        self.set_interaction(19, 11, Aij=-266.6000061)
        self.set_interaction(19, 12, Aij=99.37000275)
        self.set_interaction(19, 13, Aij=38.81000137)
        self.set_interaction(19, 14, Aij=-157.30000305)
        self.set_interaction(19, 15, Aij=-108.5)
        self.set_interaction(19, 17, Aij=777.40002441)
        self.set_interaction(19, 18, Aij=134.30000305)
        self.set_interaction(19, 21, Aij=4.93300009)
        self.set_interaction(19, 22, Aij=-152.69999695)
        self.set_interaction(19, 23, Aij=-15.61999989)
        self.set_interaction(19, 24, Aij=-54.86000061)
        self.set_interaction(19, 25, Aij=-4.62400007)
        self.set_interaction(19, 26, Aij=-0.51499999)
        self.set_interaction(19, 28, Aij=230.8999939)
        self.set_interaction(19, 29, Aij=0.46039999)
        self.set_interaction(19, 31, Aij=177.5)
        self.set_interaction(19, 33, Aij=-62.16999817)
        self.set_interaction(19, 34, Aij=-203.0)
        self.set_interaction(19, 36, Aij=81.56999969)
        self.set_interaction(19, 37, Aij=-55.77000046)
        self.set_interaction(19, 39, Aij=-151.5)
        self.set_interaction(19, 41, Aij=120.30000305)
        self.set_interaction(19, 47, Aij=16.22999954)
        self.set_interaction(19, 52, Aij=24.81999969)
        self.set_interaction(19, 53, Aij=185.3999939)
        self.set_interaction(19, 55, Aij=-287.5)
        self.set_interaction(19, 56, Aij=24.81999969)
        self.set_interaction(19, 57, Aij=38.81000137)
        self.set_interaction(19, 58, Aij=185.3999939)
        self.set_interaction(19, 59, Aij=24.81999969)
        self.set_interaction(19, 61, Aij=185.3999939)
        self.set_interaction(19, 62, Aij=24.81999969)
        self.set_interaction(19, 64, Aij=185.3999939)
        self.set_interaction(19, 66, Aij=24.81999969)
        self.set_interaction(19, 67, Aij=24.81999969)
        self.set_interaction(19, 68, Aij=24.81999969)
        self.set_interaction(19, 69, Aij=185.3999939)
        self.set_interaction(20, 1, Aij=315.29998779)
        self.set_interaction(20, 2, Aij=1264.0)
        self.set_interaction(20, 3, Aij=62.31999969)
        self.set_interaction(20, 4, Aij=89.86000061)
        self.set_interaction(20, 5, Aij=-103.02999878)
        self.set_interaction(20, 6, Aij=339.79998779)
        self.set_interaction(20, 7, Aij=-145.88000488)
        self.set_interaction(20, 8, Aij=-11.0)
        self.set_interaction(20, 9, Aij=-297.79998779)
        self.set_interaction(20, 10, Aij=-165.5)
        self.set_interaction(20, 11, Aij=-256.29998779)
        self.set_interaction(20, 12, Aij=193.8999939)
        self.set_interaction(20, 13, Aij=-338.5)
        self.set_interaction(20, 17, Aij=493.79998779)
        self.set_interaction(20, 18, Aij=-313.5)
        self.set_interaction(20, 21, Aij=13.40999985)
        self.set_interaction(20, 22, Aij=-44.70000076)
        self.set_interaction(20, 23, Aij=39.63000107)
        self.set_interaction(20, 24, Aij=183.3999939)
        self.set_interaction(20, 25, Aij=-79.08000183)
        self.set_interaction(20, 30, Aij=-208.8999939)
        self.set_interaction(20, 32, Aij=228.3999939)
        self.set_interaction(20, 33, Aij=-95.0)
        self.set_interaction(20, 35, Aij=-463.6000061)
        self.set_interaction(20, 37, Aij=-11.15999985)
        self.set_interaction(20, 39, Aij=-228.0)
        self.set_interaction(20, 41, Aij=-337.0)
        self.set_interaction(20, 46, Aij=-322.29998779)
        self.set_interaction(20, 52, Aij=315.29998779)
        self.set_interaction(20, 53, Aij=-151.0)
        self.set_interaction(20, 55, Aij=-297.79998779)
        self.set_interaction(20, 56, Aij=315.29998779)
        self.set_interaction(20, 57, Aij=-338.5)
        self.set_interaction(20, 58, Aij=-151.0)
        self.set_interaction(20, 59, Aij=315.29998779)
        self.set_interaction(20, 61, Aij=-151.0)
        self.set_interaction(20, 62, Aij=315.29998779)
        self.set_interaction(20, 64, Aij=-103.03)
        self.set_interaction(20, 66, Aij=315.29998779)
        self.set_interaction(20, 67, Aij=315.29998779)
        self.set_interaction(20, 68, Aij=315.29998779)
        self.set_interaction(20, 69, Aij=-103.02999878)
        self.set_interaction(20, 71, Aij=-100.17)
        self.set_interaction(20, 72, Aij=-501.23)
        self.set_interaction(20, 73, Aij=-579.8)
        self.set_interaction(20, 74, Aij=79.052)
        self.set_interaction(20, 75, Aij=-340.95)
        self.set_interaction(21, 1, Aij=91.45999908)
        self.set_interaction(21, 2, Aij=40.25)
        self.set_interaction(21, 3, Aij=4.67999983)
        self.set_interaction(21, 4, Aij=122.90000153)
        self.set_interaction(21, 5, Aij=562.20001221)
        self.set_interaction(21, 6, Aij=529.0)
        self.set_interaction(21, 7, Aij=698.20001221)
        self.set_interaction(21, 9, Aij=286.29998779)
        self.set_interaction(21, 10, Aij=-47.50999832)
        self.set_interaction(21, 11, Aij=35.38000107)
        self.set_interaction(21, 13, Aij=225.3999939)
        self.set_interaction(21, 14, Aij=131.19999695)
        self.set_interaction(21, 17, Aij=429.70001221)
        self.set_interaction(21, 19, Aij=54.31999969)
        self.set_interaction(21, 20, Aij=519.09997559)
        self.set_interaction(21, 22, Aij=108.30000305)
        self.set_interaction(21, 23, Aij=249.6000061)
        self.set_interaction(21, 24, Aij=62.41999817)
        self.set_interaction(21, 25, Aij=153.0)
        self.set_interaction(21, 26, Aij=32.72999954)
        self.set_interaction(21, 27, Aij=86.19999695)
        self.set_interaction(21, 28, Aij=450.1000061)
        self.set_interaction(21, 29, Aij=59.02000046)
        self.set_interaction(21, 33, Aij=344.3999939)
        self.set_interaction(21, 37, Aij=-168.19999695)
        self.set_interaction(21, 41, Aij=63.66999817)
        self.set_interaction(21, 52, Aij=91.45999908)
        self.set_interaction(21, 53, Aij=562.20001221)
        self.set_interaction(21, 54, Aij=519.09997559)
        self.set_interaction(21, 55, Aij=286.29998779)
        self.set_interaction(21, 56, Aij=91.45999908)
        self.set_interaction(21, 57, Aij=225.3999939)
        self.set_interaction(21, 58, Aij=562.20001221)
        self.set_interaction(21, 59, Aij=91.45999908)
        self.set_interaction(21, 60, Aij=519.09997559)
        self.set_interaction(21, 61, Aij=562.20001221)
        self.set_interaction(21, 62, Aij=91.45999908)
        self.set_interaction(21, 63, Aij=519.09997559)
        self.set_interaction(21, 64, Aij=562.20001221)
        self.set_interaction(21, 65, Aij=519.09997559)
        self.set_interaction(21, 66, Aij=91.45999908)
        self.set_interaction(21, 67, Aij=91.45999908)
        self.set_interaction(21, 68, Aij=91.45999908)
        self.set_interaction(21, 69, Aij=562.20001221)
        self.set_interaction(22, 1, Aij=34.00999832)
        self.set_interaction(22, 2, Aij=-23.5)
        self.set_interaction(22, 3, Aij=121.30000305)
        self.set_interaction(22, 4, Aij=140.80000305)
        self.set_interaction(22, 5, Aij=527.59997559)
        self.set_interaction(22, 6, Aij=669.90002441)
        self.set_interaction(22, 7, Aij=708.70001221)
        self.set_interaction(22, 9, Aij=82.86000061)
        self.set_interaction(22, 10, Aij=190.6000061)
        self.set_interaction(22, 11, Aij=-133.0)
        self.set_interaction(22, 13, Aij=-197.69999695)
        self.set_interaction(22, 16, Aij=-141.3999939)
        self.set_interaction(22, 17, Aij=140.80000305)
        self.set_interaction(22, 18, Aij=587.29998779)
        self.set_interaction(22, 19, Aij=258.6000061)
        self.set_interaction(22, 20, Aij=543.29998779)
        self.set_interaction(22, 21, Aij=-84.52999878)
        self.set_interaction(22, 24, Aij=56.33000183)
        self.set_interaction(22, 25, Aij=223.1000061)
        self.set_interaction(22, 26, Aij=108.90000153)
        self.set_interaction(22, 32, Aij=177.6000061)
        self.set_interaction(22, 33, Aij=315.8999939)
        self.set_interaction(22, 35, Aij=215.0)
        self.set_interaction(22, 37, Aij=-91.80000305)
        self.set_interaction(22, 41, Aij=-96.87000275)
        self.set_interaction(22, 47, Aij=361.1000061)
        self.set_interaction(22, 52, Aij=34.00999832)
        self.set_interaction(22, 53, Aij=527.59997559)
        self.set_interaction(22, 54, Aij=543.29998779)
        self.set_interaction(22, 55, Aij=82.86000061)
        self.set_interaction(22, 56, Aij=34.00999832)
        self.set_interaction(22, 57, Aij=-197.69999695)
        self.set_interaction(22, 58, Aij=527.59997559)
        self.set_interaction(22, 59, Aij=34.00999832)
        self.set_interaction(22, 60, Aij=543.29998779)
        self.set_interaction(22, 61, Aij=527.59997559)
        self.set_interaction(22, 62, Aij=34.00999832)
        self.set_interaction(22, 63, Aij=543.29998779)
        self.set_interaction(22, 64, Aij=527.59997559)
        self.set_interaction(22, 65, Aij=543.29998779)
        self.set_interaction(22, 66, Aij=34.00999832)
        self.set_interaction(22, 67, Aij=34.00999832)
        self.set_interaction(22, 68, Aij=34.00999832)
        self.set_interaction(22, 69, Aij=527.59997559)
        self.set_interaction(23, 1, Aij=36.70000076)
        self.set_interaction(23, 2, Aij=51.06000137)
        self.set_interaction(23, 3, Aij=288.5)
        self.set_interaction(23, 4, Aij=69.90000153)
        self.set_interaction(23, 5, Aij=742.09997559)
        self.set_interaction(23, 6, Aij=649.09997559)
        self.set_interaction(23, 7, Aij=826.79998779)
        self.set_interaction(23, 9, Aij=552.09997559)
        self.set_interaction(23, 10, Aij=242.80000305)
        self.set_interaction(23, 11, Aij=176.5)
        self.set_interaction(23, 12, Aij=235.6000061)
        self.set_interaction(23, 13, Aij=-20.93000031)
        self.set_interaction(23, 16, Aij=-293.70001221)
        self.set_interaction(23, 18, Aij=18.97999954)
        self.set_interaction(23, 19, Aij=74.04000092)
        self.set_interaction(23, 20, Aij=504.20001221)
        self.set_interaction(23, 21, Aij=-157.1000061)
        self.set_interaction(23, 24, Aij=-30.10000038)
        self.set_interaction(23, 25, Aij=192.1000061)
        self.set_interaction(23, 28, Aij=116.59999847)
        self.set_interaction(23, 30, Aij=-64.37999725)
        self.set_interaction(23, 32, Aij=86.40000153)
        self.set_interaction(23, 35, Aij=363.70001221)
        self.set_interaction(23, 37, Aij=111.19999695)
        self.set_interaction(23, 41, Aij=255.80000305)
        self.set_interaction(23, 44, Aij=-35.68000031)
        self.set_interaction(23, 48, Aij=565.90002441)
        self.set_interaction(23, 52, Aij=36.70000076)
        self.set_interaction(23, 53, Aij=742.09997559)
        self.set_interaction(23, 54, Aij=504.20001221)
        self.set_interaction(23, 55, Aij=552.09997559)
        self.set_interaction(23, 56, Aij=36.70000076)
        self.set_interaction(23, 57, Aij=-20.93000031)
        self.set_interaction(23, 58, Aij=742.09997559)
        self.set_interaction(23, 59, Aij=36.70000076)
        self.set_interaction(23, 60, Aij=504.20001221)
        self.set_interaction(23, 61, Aij=742.09997559)
        self.set_interaction(23, 62, Aij=36.70000076)
        self.set_interaction(23, 63, Aij=504.20001221)
        self.set_interaction(23, 64, Aij=742.09997559)
        self.set_interaction(23, 65, Aij=504.20001221)
        self.set_interaction(23, 66, Aij=36.70000076)
        self.set_interaction(23, 67, Aij=36.70000076)
        self.set_interaction(23, 68, Aij=36.70000076)
        self.set_interaction(23, 69, Aij=742.09997559)
        self.set_interaction(24, 1, Aij=-78.44999695)
        self.set_interaction(24, 2, Aij=160.8999939)
        self.set_interaction(24, 3, Aij=-4.69999981)
        self.set_interaction(24, 4, Aij=134.69999695)
        self.set_interaction(24, 5, Aij=856.29998779)
        self.set_interaction(24, 6, Aij=709.59997559)
        self.set_interaction(24, 7, Aij=1201.0)
        self.set_interaction(24, 8, Aij=10000.0)
        self.set_interaction(24, 9, Aij=372.0)
        self.set_interaction(24, 11, Aij=129.5)
        self.set_interaction(24, 12, Aij=351.8999939)
        self.set_interaction(24, 13, Aij=113.90000153)
        self.set_interaction(24, 14, Aij=261.1000061)
        self.set_interaction(24, 15, Aij=91.12999725)
        self.set_interaction(24, 16, Aij=316.8999939)
        self.set_interaction(24, 17, Aij=898.20001221)
        self.set_interaction(24, 18, Aij=368.5)
        self.set_interaction(24, 19, Aij=492.0)
        self.set_interaction(24, 20, Aij=631.0)
        self.set_interaction(24, 21, Aij=11.80000019)
        self.set_interaction(24, 22, Aij=17.96999931)
        self.set_interaction(24, 23, Aij=51.90000153)
        self.set_interaction(24, 25, Aij=-75.97000122)
        self.set_interaction(24, 26, Aij=490.8999939)
        self.set_interaction(24, 27, Aij=534.70001221)
        self.set_interaction(24, 28, Aij=132.19999695)
        self.set_interaction(24, 30, Aij=546.70001221)
        self.set_interaction(24, 32, Aij=247.80000305)
        self.set_interaction(24, 33, Aij=146.6000061)
        self.set_interaction(24, 35, Aij=337.70001221)
        self.set_interaction(24, 36, Aij=369.5)
        self.set_interaction(24, 37, Aij=187.1000061)
        self.set_interaction(24, 38, Aij=215.19999695)
        self.set_interaction(24, 39, Aij=498.6000061)
        self.set_interaction(24, 41, Aij=256.5)
        self.set_interaction(24, 43, Aij=233.1000061)
        self.set_interaction(24, 47, Aij=423.1000061)
        self.set_interaction(24, 48, Aij=63.95000076)
        self.set_interaction(24, 50, Aij=108.5)
        self.set_interaction(24, 52, Aij=-78.44999695)
        self.set_interaction(24, 53, Aij=856.29998779)
        self.set_interaction(24, 54, Aij=631.0)
        self.set_interaction(24, 55, Aij=372.0)
        self.set_interaction(24, 56, Aij=-78.44999695)
        self.set_interaction(24, 57, Aij=113.90000153)
        self.set_interaction(24, 58, Aij=856.29998779)
        self.set_interaction(24, 59, Aij=-78.44999695)
        self.set_interaction(24, 60, Aij=631.0)
        self.set_interaction(24, 61, Aij=856.29998779)
        self.set_interaction(24, 62, Aij=-78.44999695)
        self.set_interaction(24, 63, Aij=631.0)
        self.set_interaction(24, 64, Aij=856.29998779)
        self.set_interaction(24, 65, Aij=631.0)
        self.set_interaction(24, 66, Aij=-78.44999695)
        self.set_interaction(24, 67, Aij=-78.44999695)
        self.set_interaction(24, 68, Aij=-78.44999695)
        self.set_interaction(24, 69, Aij=856.29998779)
        self.set_interaction(25, 1, Aij=106.80000305)
        self.set_interaction(25, 2, Aij=70.31999969)
        self.set_interaction(25, 3, Aij=-97.26999664)
        self.set_interaction(25, 4, Aij=402.5)
        self.set_interaction(25, 5, Aij=325.70001221)
        self.set_interaction(25, 6, Aij=612.79998779)
        self.set_interaction(25, 7, Aij=-274.5)
        self.set_interaction(25, 8, Aij=622.29998779)
        self.set_interaction(25, 9, Aij=518.40002441)
        self.set_interaction(25, 11, Aij=-171.1000061)
        self.set_interaction(25, 12, Aij=383.29998779)
        self.set_interaction(25, 13, Aij=-25.14999962)
        self.set_interaction(25, 14, Aij=108.5)
        self.set_interaction(25, 15, Aij=102.19999695)
        self.set_interaction(25, 16, Aij=2951.0)
        self.set_interaction(25, 17, Aij=334.8999939)
        self.set_interaction(25, 19, Aij=363.5)
        self.set_interaction(25, 20, Aij=993.40002441)
        self.set_interaction(25, 21, Aij=-129.69999695)
        self.set_interaction(25, 22, Aij=-8.30900002)
        self.set_interaction(25, 23, Aij=-0.22660001)
        self.set_interaction(25, 24, Aij=248.3999939)
        self.set_interaction(25, 26, Aij=132.69999695)
        self.set_interaction(25, 27, Aij=2213.0)
        self.set_interaction(25, 33, Aij=593.40002441)
        self.set_interaction(25, 41, Aij=-71.18000031)
        self.set_interaction(25, 44, Aij=-209.69999695)
        self.set_interaction(25, 47, Aij=434.1000061)
        self.set_interaction(25, 52, Aij=106.80000305)
        self.set_interaction(25, 53, Aij=325.70001221)
        self.set_interaction(25, 54, Aij=993.40002441)
        self.set_interaction(25, 55, Aij=518.40002441)
        self.set_interaction(25, 56, Aij=106.80000305)
        self.set_interaction(25, 57, Aij=-25.14999962)
        self.set_interaction(25, 58, Aij=325.70001221)
        self.set_interaction(25, 59, Aij=106.80000305)
        self.set_interaction(25, 60, Aij=993.40002441)
        self.set_interaction(25, 61, Aij=325.70001221)
        self.set_interaction(25, 62, Aij=106.80000305)
        self.set_interaction(25, 63, Aij=993.40002441)
        self.set_interaction(25, 64, Aij=325.70001221)
        self.set_interaction(25, 65, Aij=993.40002441)
        self.set_interaction(25, 66, Aij=106.80000305)
        self.set_interaction(25, 67, Aij=106.80000305)
        self.set_interaction(25, 68, Aij=106.80000305)
        self.set_interaction(25, 69, Aij=325.70001221)
        self.set_interaction(26, 1, Aij=-32.68999863)
        self.set_interaction(26, 2, Aij=-1.99600005)
        self.set_interaction(26, 3, Aij=10.38000011)
        self.set_interaction(26, 4, Aij=-97.05000305)
        self.set_interaction(26, 5, Aij=261.6000061)
        self.set_interaction(26, 6, Aij=252.6000061)
        self.set_interaction(26, 7, Aij=417.8999939)
        self.set_interaction(26, 9, Aij=-142.6000061)
        self.set_interaction(26, 11, Aij=129.30000305)
        self.set_interaction(26, 13, Aij=-94.48999786)
        self.set_interaction(26, 19, Aij=0.28299999)
        self.set_interaction(26, 21, Aij=113.0)
        self.set_interaction(26, 22, Aij=-9.63899994)
        self.set_interaction(26, 24, Aij=-34.68000031)
        self.set_interaction(26, 25, Aij=132.8999939)
        self.set_interaction(26, 27, Aij=533.20001221)
        self.set_interaction(26, 28, Aij=320.20001221)
        self.set_interaction(26, 31, Aij=139.80000305)
        self.set_interaction(26, 32, Aij=304.29998779)
        self.set_interaction(26, 33, Aij=10.17000008)
        self.set_interaction(26, 34, Aij=-27.70000076)
        self.set_interaction(26, 37, Aij=10.76000023)
        self.set_interaction(26, 39, Aij=-223.1000061)
        self.set_interaction(26, 41, Aij=248.3999939)
        self.set_interaction(26, 45, Aij=-218.8999939)
        self.set_interaction(26, 50, Aij=-4.56500006)
        self.set_interaction(26, 52, Aij=-32.68999863)
        self.set_interaction(26, 53, Aij=261.6000061)
        self.set_interaction(26, 55, Aij=-142.6000061)
        self.set_interaction(26, 56, Aij=-32.68999863)
        self.set_interaction(26, 57, Aij=-94.48999786)
        self.set_interaction(26, 58, Aij=261.6000061)
        self.set_interaction(26, 59, Aij=-32.68999863)
        self.set_interaction(26, 61, Aij=261.6000061)
        self.set_interaction(26, 62, Aij=-32.68999863)
        self.set_interaction(26, 64, Aij=261.6000061)
        self.set_interaction(26, 66, Aij=-32.68999863)
        self.set_interaction(26, 67, Aij=-32.68999863)
        self.set_interaction(26, 68, Aij=-32.68999863)
        self.set_interaction(26, 69, Aij=261.6000061)
        self.set_interaction(27, 1, Aij=5541.0)
        self.set_interaction(27, 3, Aij=1824.0)
        self.set_interaction(27, 4, Aij=-127.80000305)
        self.set_interaction(27, 5, Aij=561.59997559)
        self.set_interaction(27, 7, Aij=360.70001221)
        self.set_interaction(27, 9, Aij=-101.5)
        self.set_interaction(27, 17, Aij=134.8999939)
        self.set_interaction(27, 18, Aij=2475.0)
        self.set_interaction(27, 21, Aij=1971.0)
        self.set_interaction(27, 24, Aij=514.59997559)
        self.set_interaction(27, 25, Aij=-123.09999847)
        self.set_interaction(27, 26, Aij=-85.12000275)
        self.set_interaction(27, 32, Aij=2990.0)
        self.set_interaction(27, 33, Aij=-124.0)
        self.set_interaction(27, 52, Aij=5541.0)
        self.set_interaction(27, 53, Aij=561.59997559)
        self.set_interaction(27, 55, Aij=-101.5)
        self.set_interaction(27, 56, Aij=5541.0)
        self.set_interaction(27, 58, Aij=561.59997559)
        self.set_interaction(27, 59, Aij=5541.0)
        self.set_interaction(27, 61, Aij=561.59997559)
        self.set_interaction(27, 62, Aij=5541.0)
        self.set_interaction(27, 64, Aij=561.59997559)
        self.set_interaction(27, 66, Aij=5541.0)
        self.set_interaction(27, 67, Aij=5541.0)
        self.set_interaction(27, 68, Aij=5541.0)
        self.set_interaction(27, 69, Aij=561.59997559)
        self.set_interaction(28, 1, Aij=-52.65000153)
        self.set_interaction(28, 2, Aij=16.62000084)
        self.set_interaction(28, 3, Aij=21.5)
        self.set_interaction(28, 4, Aij=40.68000031)
        self.set_interaction(28, 5, Aij=609.79998779)
        self.set_interaction(28, 6, Aij=914.20001221)
        self.set_interaction(28, 7, Aij=1081.0)
        self.set_interaction(28, 8, Aij=1421.0)
        self.set_interaction(28, 9, Aij=303.70001221)
        self.set_interaction(28, 11, Aij=243.80000305)
        self.set_interaction(28, 13, Aij=112.40000153)
        self.set_interaction(28, 19, Aij=335.70001221)
        self.set_interaction(28, 21, Aij=-73.08999634)
        self.set_interaction(28, 23, Aij=-26.05999947)
        self.set_interaction(28, 24, Aij=-60.70999908)
        self.set_interaction(28, 26, Aij=277.79998779)
        self.set_interaction(28, 32, Aij=292.70001221)
        self.set_interaction(28, 37, Aij=-47.36999893)
        self.set_interaction(28, 41, Aij=469.79998779)
        self.set_interaction(28, 52, Aij=-52.65000153)
        self.set_interaction(28, 53, Aij=609.79998779)
        self.set_interaction(28, 55, Aij=303.70001221)
        self.set_interaction(28, 56, Aij=-52.65000153)
        self.set_interaction(28, 57, Aij=112.40000153)
        self.set_interaction(28, 58, Aij=609.79998779)
        self.set_interaction(28, 59, Aij=-52.65000153)
        self.set_interaction(28, 61, Aij=609.79998779)
        self.set_interaction(28, 62, Aij=-52.65000153)
        self.set_interaction(28, 64, Aij=609.79998779)
        self.set_interaction(28, 66, Aij=-52.65000153)
        self.set_interaction(28, 67, Aij=-52.65000153)
        self.set_interaction(28, 68, Aij=-52.65000153)
        self.set_interaction(28, 69, Aij=609.79998779)
        self.set_interaction(29, 1, Aij=-7.48099995)
        self.set_interaction(29, 3, Aij=28.40999985)
        self.set_interaction(29, 4, Aij=19.55999947)
        self.set_interaction(29, 5, Aij=461.6000061)
        self.set_interaction(29, 6, Aij=448.6000061)
        self.set_interaction(29, 9, Aij=160.6000061)
        self.set_interaction(29, 12, Aij=201.5)
        self.set_interaction(29, 13, Aij=63.70999908)
        self.set_interaction(29, 14, Aij=106.69999695)
        self.set_interaction(29, 19, Aij=161.0)
        self.set_interaction(29, 21, Aij=-27.94000053)
        self.set_interaction(29, 35, Aij=31.65999985)
        self.set_interaction(29, 39, Aij=78.91999817)
        self.set_interaction(29, 44, Aij=1004.0)
        self.set_interaction(29, 48, Aij=-18.27000046)
        self.set_interaction(29, 52, Aij=-7.48099995)
        self.set_interaction(29, 53, Aij=461.6000061)
        self.set_interaction(29, 55, Aij=160.6000061)
        self.set_interaction(29, 56, Aij=-7.48099995)
        self.set_interaction(29, 57, Aij=63.70999908)
        self.set_interaction(29, 58, Aij=461.6000061)
        self.set_interaction(29, 59, Aij=-7.48099995)
        self.set_interaction(29, 61, Aij=461.6000061)
        self.set_interaction(29, 62, Aij=-7.48099995)
        self.set_interaction(29, 64, Aij=461.6000061)
        self.set_interaction(29, 66, Aij=-7.48099995)
        self.set_interaction(29, 67, Aij=-7.48099995)
        self.set_interaction(29, 68, Aij=-7.48099995)
        self.set_interaction(29, 69, Aij=461.6000061)
        self.set_interaction(30, 1, Aij=-25.30999947)
        self.set_interaction(30, 2, Aij=82.63999939)
        self.set_interaction(30, 3, Aij=157.30000305)
        self.set_interaction(30, 4, Aij=128.80000305)
        self.set_interaction(30, 5, Aij=521.59997559)
        self.set_interaction(30, 7, Aij=23.47999954)
        self.set_interaction(30, 9, Aij=317.5)
        self.set_interaction(30, 11, Aij=-146.30000305)
        self.set_interaction(30, 13, Aij=-87.30999756)
        self.set_interaction(30, 20, Aij=570.59997559)
        self.set_interaction(30, 23, Aij=48.47999954)
        self.set_interaction(30, 24, Aij=-133.19999695)
        self.set_interaction(30, 41, Aij=43.36999893)
        self.set_interaction(30, 52, Aij=-25.30999947)
        self.set_interaction(30, 53, Aij=521.59997559)
        self.set_interaction(30, 54, Aij=570.59997559)
        self.set_interaction(30, 55, Aij=317.5)
        self.set_interaction(30, 56, Aij=-25.30999947)
        self.set_interaction(30, 57, Aij=-87.30999756)
        self.set_interaction(30, 58, Aij=521.59997559)
        self.set_interaction(30, 59, Aij=-25.30999947)
        self.set_interaction(30, 60, Aij=570.59997559)
        self.set_interaction(30, 61, Aij=521.59997559)
        self.set_interaction(30, 62, Aij=-25.30999947)
        self.set_interaction(30, 63, Aij=570.59997559)
        self.set_interaction(30, 64, Aij=521.59997559)
        self.set_interaction(30, 65, Aij=570.59997559)
        self.set_interaction(30, 66, Aij=-25.30999947)
        self.set_interaction(30, 67, Aij=-25.30999947)
        self.set_interaction(30, 68, Aij=-25.30999947)
        self.set_interaction(30, 69, Aij=521.59997559)
        self.set_interaction(31, 1, Aij=139.8999939)
        self.set_interaction(31, 3, Aij=221.3999939)
        self.set_interaction(31, 4, Aij=150.6000061)
        self.set_interaction(31, 5, Aij=267.6000061)
        self.set_interaction(31, 6, Aij=240.80000305)
        self.set_interaction(31, 7, Aij=-137.3999939)
        self.set_interaction(31, 8, Aij=838.40002441)
        self.set_interaction(31, 9, Aij=135.3999939)
        self.set_interaction(31, 11, Aij=152.0)
        self.set_interaction(31, 13, Aij=9.20699978)
        self.set_interaction(31, 17, Aij=192.30000305)
        self.set_interaction(31, 19, Aij=169.6000061)
        self.set_interaction(31, 26, Aij=481.29998779)
        self.set_interaction(31, 35, Aij=-417.20001221)
        self.set_interaction(31, 39, Aij=302.20001221)
        self.set_interaction(31, 41, Aij=347.79998779)
        self.set_interaction(31, 44, Aij=-262.0)
        self.set_interaction(31, 47, Aij=-353.5)
        self.set_interaction(31, 52, Aij=139.8999939)
        self.set_interaction(31, 53, Aij=267.6000061)
        self.set_interaction(31, 55, Aij=135.3999939)
        self.set_interaction(31, 56, Aij=139.8999939)
        self.set_interaction(31, 57, Aij=9.20699978)
        self.set_interaction(31, 58, Aij=267.6000061)
        self.set_interaction(31, 59, Aij=139.8999939)
        self.set_interaction(31, 61, Aij=267.6000061)
        self.set_interaction(31, 62, Aij=139.8999939)
        self.set_interaction(31, 64, Aij=267.6000061)
        self.set_interaction(31, 66, Aij=139.8999939)
        self.set_interaction(31, 67, Aij=139.8999939)
        self.set_interaction(31, 68, Aij=139.8999939)
        self.set_interaction(31, 69, Aij=267.6000061)
        self.set_interaction(32, 1, Aij=128.0)
        self.set_interaction(32, 3, Aij=58.68000031)
        self.set_interaction(32, 4, Aij=26.40999985)
        self.set_interaction(32, 5, Aij=501.29998779)
        self.set_interaction(32, 6, Aij=431.29998779)
        self.set_interaction(32, 9, Aij=138.0)
        self.set_interaction(32, 10, Aij=245.8999939)
        self.set_interaction(32, 11, Aij=21.92000008)
        self.set_interaction(32, 13, Aij=476.6000061)
        self.set_interaction(32, 20, Aij=616.59997559)
        self.set_interaction(32, 22, Aij=-40.81999969)
        self.set_interaction(32, 23, Aij=21.76000023)
        self.set_interaction(32, 24, Aij=48.49000168)
        self.set_interaction(32, 26, Aij=64.27999878)
        self.set_interaction(32, 27, Aij=2448.0)
        self.set_interaction(32, 28, Aij=-27.45000076)
        self.set_interaction(32, 41, Aij=68.55000305)
        self.set_interaction(32, 52, Aij=128.0)
        self.set_interaction(32, 53, Aij=501.29998779)
        self.set_interaction(32, 54, Aij=616.59997559)
        self.set_interaction(32, 55, Aij=138.0)
        self.set_interaction(32, 56, Aij=128.0)
        self.set_interaction(32, 57, Aij=476.6000061)
        self.set_interaction(32, 58, Aij=501.29998779)
        self.set_interaction(32, 59, Aij=128.0)
        self.set_interaction(32, 60, Aij=616.59997559)
        self.set_interaction(32, 61, Aij=501.29998779)
        self.set_interaction(32, 62, Aij=128.0)
        self.set_interaction(32, 63, Aij=616.59997559)
        self.set_interaction(32, 64, Aij=501.29998779)
        self.set_interaction(32, 65, Aij=616.59997559)
        self.set_interaction(32, 66, Aij=128.0)
        self.set_interaction(32, 67, Aij=128.0)
        self.set_interaction(32, 68, Aij=128.0)
        self.set_interaction(32, 69, Aij=501.29998779)
        self.set_interaction(33, 1, Aij=-31.52000046)
        self.set_interaction(33, 2, Aij=174.6000061)
        self.set_interaction(33, 3, Aij=-154.19999695)
        self.set_interaction(33, 4, Aij=1112.0)
        self.set_interaction(33, 5, Aij=524.90002441)
        self.set_interaction(33, 6, Aij=494.70001221)
        self.set_interaction(33, 9, Aij=-142.6000061)
        self.set_interaction(33, 11, Aij=24.37000084)
        self.set_interaction(33, 13, Aij=736.40002441)
        self.set_interaction(33, 18, Aij=-42.70999908)
        self.set_interaction(33, 19, Aij=136.8999939)
        self.set_interaction(33, 20, Aij=5256.0)
        self.set_interaction(33, 21, Aij=-262.29998779)
        self.set_interaction(33, 22, Aij=-174.5)
        self.set_interaction(33, 24, Aij=77.55000305)
        self.set_interaction(33, 25, Aij=-185.30000305)
        self.set_interaction(33, 26, Aij=125.30000305)
        self.set_interaction(33, 27, Aij=4288.0)
        self.set_interaction(33, 35, Aij=32.90000153)
        self.set_interaction(33, 41, Aij=-195.1000061)
        self.set_interaction(33, 52, Aij=-31.52000046)
        self.set_interaction(33, 53, Aij=524.90002441)
        self.set_interaction(33, 54, Aij=5256.0)
        self.set_interaction(33, 55, Aij=-142.6000061)
        self.set_interaction(33, 56, Aij=-31.52000046)
        self.set_interaction(33, 57, Aij=736.40002441)
        self.set_interaction(33, 58, Aij=524.90002441)
        self.set_interaction(33, 59, Aij=-31.52000046)
        self.set_interaction(33, 60, Aij=5256.0)
        self.set_interaction(33, 61, Aij=524.90002441)
        self.set_interaction(33, 62, Aij=-31.52000046)
        self.set_interaction(33, 63, Aij=5256.0)
        self.set_interaction(33, 64, Aij=524.90002441)
        self.set_interaction(33, 65, Aij=5256.0)
        self.set_interaction(33, 66, Aij=-31.52000046)
        self.set_interaction(33, 67, Aij=-31.52000046)
        self.set_interaction(33, 68, Aij=-31.52000046)
        self.set_interaction(33, 69, Aij=524.90002441)
        self.set_interaction(34, 1, Aij=-72.87999725)
        self.set_interaction(34, 2, Aij=41.38000107)
        self.set_interaction(34, 5, Aij=68.94999695)
        self.set_interaction(34, 9, Aij=443.6000061)
        self.set_interaction(34, 19, Aij=329.1000061)
        self.set_interaction(34, 26, Aij=174.3999939)
        self.set_interaction(34, 37, Aij=2073.0)
        self.set_interaction(34, 39, Aij=-119.80000305)
        self.set_interaction(34, 52, Aij=-72.87999725)
        self.set_interaction(34, 53, Aij=68.94999695)
        self.set_interaction(34, 55, Aij=443.6000061)
        self.set_interaction(34, 56, Aij=-72.87999725)
        self.set_interaction(34, 58, Aij=68.94999695)
        self.set_interaction(34, 59, Aij=-72.87999725)
        self.set_interaction(34, 61, Aij=68.94999695)
        self.set_interaction(34, 62, Aij=-72.87999725)
        self.set_interaction(34, 64, Aij=68.94999695)
        self.set_interaction(34, 66, Aij=-72.87999725)
        self.set_interaction(34, 67, Aij=-72.87999725)
        self.set_interaction(34, 68, Aij=-72.87999725)
        self.set_interaction(34, 69, Aij=68.94999695)
        self.set_interaction(35, 1, Aij=50.49000168)
        self.set_interaction(35, 2, Aij=64.06999969)
        self.set_interaction(35, 3, Aij=-2.50399995)
        self.set_interaction(35, 4, Aij=-143.19999695)
        self.set_interaction(35, 5, Aij=-25.87000084)
        self.set_interaction(35, 6, Aij=695.0)
        self.set_interaction(35, 7, Aij=-240.0)
        self.set_interaction(35, 9, Aij=110.40000153)
        self.set_interaction(35, 11, Aij=41.56999969)
        self.set_interaction(35, 13, Aij=-93.51000214)
        self.set_interaction(35, 16, Aij=-257.20001221)
        self.set_interaction(35, 20, Aij=-180.19999695)
        self.set_interaction(35, 22, Aij=-215.0)
        self.set_interaction(35, 23, Aij=-343.6000061)
        self.set_interaction(35, 24, Aij=-58.43000031)
        self.set_interaction(35, 29, Aij=85.69999695)
        self.set_interaction(35, 31, Aij=535.79998779)
        self.set_interaction(35, 33, Aij=-111.19999695)
        self.set_interaction(35, 39, Aij=-97.70999908)
        self.set_interaction(35, 41, Aij=153.69999695)
        self.set_interaction(35, 52, Aij=50.49000168)
        self.set_interaction(35, 53, Aij=-25.87000084)
        self.set_interaction(35, 54, Aij=-180.19999695)
        self.set_interaction(35, 55, Aij=110.40000153)
        self.set_interaction(35, 56, Aij=50.49000168)
        self.set_interaction(35, 57, Aij=-93.51000214)
        self.set_interaction(35, 58, Aij=-25.87000084)
        self.set_interaction(35, 59, Aij=50.49000168)
        self.set_interaction(35, 60, Aij=-180.19999695)
        self.set_interaction(35, 61, Aij=-25.87000084)
        self.set_interaction(35, 62, Aij=50.49000168)
        self.set_interaction(35, 63, Aij=-180.19999695)
        self.set_interaction(35, 64, Aij=-25.87000084)
        self.set_interaction(35, 65, Aij=-180.19999695)
        self.set_interaction(35, 66, Aij=50.49000168)
        self.set_interaction(35, 67, Aij=50.49000168)
        self.set_interaction(35, 68, Aij=50.49000168)
        self.set_interaction(35, 69, Aij=-25.87000084)
        self.set_interaction(36, 1, Aij=-165.8999939)
        self.set_interaction(36, 2, Aij=573.0)
        self.set_interaction(36, 3, Aij=-123.59999847)
        self.set_interaction(36, 4, Aij=397.3999939)
        self.set_interaction(36, 5, Aij=389.29998779)
        self.set_interaction(36, 6, Aij=218.80000305)
        self.set_interaction(36, 7, Aij=386.6000061)
        self.set_interaction(36, 10, Aij=354.0)
        self.set_interaction(36, 11, Aij=175.5)
        self.set_interaction(36, 19, Aij=-42.31000137)
        self.set_interaction(36, 24, Aij=-85.15000153)
        self.set_interaction(36, 37, Aij=-208.80000305)
        self.set_interaction(36, 39, Aij=-8.8039999)
        self.set_interaction(36, 41, Aij=423.3999939)
        self.set_interaction(36, 52, Aij=-165.8999939)
        self.set_interaction(36, 53, Aij=389.29998779)
        self.set_interaction(36, 56, Aij=-165.8999939)
        self.set_interaction(36, 58, Aij=389.29998779)
        self.set_interaction(36, 59, Aij=-165.8999939)
        self.set_interaction(36, 61, Aij=389.29998779)
        self.set_interaction(36, 62, Aij=-165.8999939)
        self.set_interaction(36, 64, Aij=389.29998779)
        self.set_interaction(36, 66, Aij=-165.8999939)
        self.set_interaction(36, 67, Aij=-165.8999939)
        self.set_interaction(36, 68, Aij=-165.8999939)
        self.set_interaction(36, 69, Aij=389.29998779)
        self.set_interaction(37, 1, Aij=47.40999985)
        self.set_interaction(37, 2, Aij=124.19999695)
        self.set_interaction(37, 3, Aij=395.79998779)
        self.set_interaction(37, 4, Aij=419.1000061)
        self.set_interaction(37, 5, Aij=738.90002441)
        self.set_interaction(37, 6, Aij=528.0)
        self.set_interaction(37, 9, Aij=-40.90000153)
        self.set_interaction(37, 10, Aij=183.80000305)
        self.set_interaction(37, 11, Aij=611.29998779)
        self.set_interaction(37, 12, Aij=134.5)
        self.set_interaction(37, 13, Aij=-217.8999939)
        self.set_interaction(37, 18, Aij=281.6000061)
        self.set_interaction(37, 19, Aij=335.20001221)
        self.set_interaction(37, 20, Aij=898.20001221)
        self.set_interaction(37, 21, Aij=383.20001221)
        self.set_interaction(37, 22, Aij=301.8999939)
        self.set_interaction(37, 23, Aij=-149.80000305)
        self.set_interaction(37, 24, Aij=-134.19999695)
        self.set_interaction(37, 26, Aij=379.3999939)
        self.set_interaction(37, 28, Aij=167.8999939)
        self.set_interaction(37, 34, Aij=631.5)
        self.set_interaction(37, 36, Aij=837.20001221)
        self.set_interaction(37, 39, Aij=255.0)
        self.set_interaction(37, 41, Aij=730.79998779)
        self.set_interaction(37, 44, Aij=26.35000038)
        self.set_interaction(37, 48, Aij=2429.0)
        self.set_interaction(37, 52, Aij=47.40999985)
        self.set_interaction(37, 53, Aij=738.90002441)
        self.set_interaction(37, 54, Aij=898.20001221)
        self.set_interaction(37, 55, Aij=-40.90000153)
        self.set_interaction(37, 56, Aij=47.40999985)
        self.set_interaction(37, 57, Aij=-217.8999939)
        self.set_interaction(37, 58, Aij=738.90002441)
        self.set_interaction(37, 59, Aij=47.40999985)
        self.set_interaction(37, 60, Aij=898.20001221)
        self.set_interaction(37, 61, Aij=738.90002441)
        self.set_interaction(37, 62, Aij=47.40999985)
        self.set_interaction(37, 63, Aij=898.20001221)
        self.set_interaction(37, 64, Aij=738.90002441)
        self.set_interaction(37, 65, Aij=898.20001221)
        self.set_interaction(37, 66, Aij=47.40999985)
        self.set_interaction(37, 67, Aij=47.40999985)
        self.set_interaction(37, 68, Aij=47.40999985)
        self.set_interaction(37, 69, Aij=738.90002441)
        self.set_interaction(38, 1, Aij=-5.13199997)
        self.set_interaction(38, 2, Aij=-131.69999695)
        self.set_interaction(38, 3, Aij=-237.19999695)
        self.set_interaction(38, 4, Aij=-157.30000305)
        self.set_interaction(38, 5, Aij=649.70001221)
        self.set_interaction(38, 6, Aij=645.90002441)
        self.set_interaction(38, 13, Aij=167.30000305)
        self.set_interaction(38, 15, Aij=-198.80000305)
        self.set_interaction(38, 16, Aij=116.5)
        self.set_interaction(38, 18, Aij=159.80000305)
        self.set_interaction(38, 24, Aij=-124.59999847)
        self.set_interaction(38, 40, Aij=-117.19999695)
        self.set_interaction(38, 52, Aij=-5.13199997)
        self.set_interaction(38, 53, Aij=649.70001221)
        self.set_interaction(38, 56, Aij=-5.13199997)
        self.set_interaction(38, 57, Aij=167.30000305)
        self.set_interaction(38, 58, Aij=649.70001221)
        self.set_interaction(38, 59, Aij=-5.13199997)
        self.set_interaction(38, 61, Aij=649.70001221)
        self.set_interaction(38, 62, Aij=-5.13199997)
        self.set_interaction(38, 64, Aij=649.70001221)
        self.set_interaction(38, 66, Aij=-5.13199997)
        self.set_interaction(38, 67, Aij=-5.13199997)
        self.set_interaction(38, 68, Aij=-5.13199997)
        self.set_interaction(38, 69, Aij=649.70001221)
        self.set_interaction(39, 1, Aij=-31.95000076)
        self.set_interaction(39, 2, Aij=249.0)
        self.set_interaction(39, 3, Aij=-133.8999939)
        self.set_interaction(39, 4, Aij=-240.19999695)
        self.set_interaction(39, 5, Aij=64.16000366)
        self.set_interaction(39, 6, Aij=172.19999695)
        self.set_interaction(39, 7, Aij=-287.1000061)
        self.set_interaction(39, 9, Aij=97.04000092)
        self.set_interaction(39, 10, Aij=13.89000034)
        self.set_interaction(39, 11, Aij=-82.12000275)
        self.set_interaction(39, 12, Aij=-116.69999695)
        self.set_interaction(39, 13, Aij=-158.19999695)
        self.set_interaction(39, 14, Aij=49.70000076)
        self.set_interaction(39, 16, Aij=-185.19999695)
        self.set_interaction(39, 17, Aij=343.70001221)
        self.set_interaction(39, 19, Aij=150.6000061)
        self.set_interaction(39, 20, Aij=-97.76999664)
        self.set_interaction(39, 24, Aij=-186.69999695)
        self.set_interaction(39, 26, Aij=223.6000061)
        self.set_interaction(39, 29, Aij=-71.0)
        self.set_interaction(39, 31, Aij=-191.69999695)
        self.set_interaction(39, 34, Aij=6.69899988)
        self.set_interaction(39, 35, Aij=136.6000061)
        self.set_interaction(39, 36, Aij=5.1500001)
        self.set_interaction(39, 37, Aij=-137.69999695)
        self.set_interaction(39, 40, Aij=-5.579)
        self.set_interaction(39, 41, Aij=72.30999756)
        self.set_interaction(39, 52, Aij=-31.95000076)
        self.set_interaction(39, 53, Aij=64.16000366)
        self.set_interaction(39, 54, Aij=-97.76999664)
        self.set_interaction(39, 55, Aij=97.04000092)
        self.set_interaction(39, 56, Aij=-31.95000076)
        self.set_interaction(39, 57, Aij=-158.19999695)
        self.set_interaction(39, 58, Aij=64.16000366)
        self.set_interaction(39, 59, Aij=-31.95000076)
        self.set_interaction(39, 60, Aij=-97.76999664)
        self.set_interaction(39, 61, Aij=64.16000366)
        self.set_interaction(39, 62, Aij=-31.95000076)
        self.set_interaction(39, 63, Aij=-97.76999664)
        self.set_interaction(39, 64, Aij=64.16000366)
        self.set_interaction(39, 65, Aij=-97.76999664)
        self.set_interaction(39, 66, Aij=-31.95000076)
        self.set_interaction(39, 67, Aij=-31.95000076)
        self.set_interaction(39, 68, Aij=-31.95000076)
        self.set_interaction(39, 69, Aij=64.16000366)
        self.set_interaction(40, 1, Aij=147.30000305)
        self.set_interaction(40, 2, Aij=62.40000153)
        self.set_interaction(40, 3, Aij=140.6000061)
        self.set_interaction(40, 38, Aij=185.6000061)
        self.set_interaction(40, 39, Aij=55.79999924)
        self.set_interaction(40, 45, Aij=111.80000305)
        self.set_interaction(40, 52, Aij=147.30000305)
        self.set_interaction(40, 56, Aij=147.30000305)
        self.set_interaction(40, 59, Aij=147.30000305)
        self.set_interaction(40, 62, Aij=147.30000305)
        self.set_interaction(40, 66, Aij=147.30000305)
        self.set_interaction(40, 67, Aij=147.30000305)
        self.set_interaction(40, 68, Aij=147.30000305)
        self.set_interaction(41, 1, Aij=529.0)
        self.set_interaction(41, 2, Aij=1397.0)
        self.set_interaction(41, 3, Aij=317.6000061)
        self.set_interaction(41, 4, Aij=615.79998779)
        self.set_interaction(41, 5, Aij=88.62999725)
        self.set_interaction(41, 6, Aij=171.0)
        self.set_interaction(41, 7, Aij=284.3999939)
        self.set_interaction(41, 8, Aij=-167.30000305)
        self.set_interaction(41, 9, Aij=123.40000153)
        self.set_interaction(41, 10, Aij=577.5)
        self.set_interaction(41, 11, Aij=-234.8999939)
        self.set_interaction(41, 12, Aij=145.3999939)
        self.set_interaction(41, 13, Aij=-247.80000305)
        self.set_interaction(41, 15, Aij=284.5)
        self.set_interaction(41, 17, Aij=-22.10000038)
        self.set_interaction(41, 19, Aij=-61.59999847)
        self.set_interaction(41, 20, Aij=1179.0)
        self.set_interaction(41, 21, Aij=182.19999695)
        self.set_interaction(41, 22, Aij=305.3999939)
        self.set_interaction(41, 23, Aij=-193.0)
        self.set_interaction(41, 24, Aij=335.70001221)
        self.set_interaction(41, 25, Aij=956.09997559)
        self.set_interaction(41, 26, Aij=-124.69999695)
        self.set_interaction(41, 28, Aij=885.5)
        self.set_interaction(41, 30, Aij=-64.27999878)
        self.set_interaction(41, 31, Aij=-264.29998779)
        self.set_interaction(41, 32, Aij=288.1000061)
        self.set_interaction(41, 33, Aij=627.70001221)
        self.set_interaction(41, 35, Aij=-29.34000015)
        self.set_interaction(41, 36, Aij=-53.90999985)
        self.set_interaction(41, 37, Aij=-198.0)
        self.set_interaction(41, 39, Aij=-28.64999962)
        self.set_interaction(41, 47, Aij=122.40000153)
        self.set_interaction(41, 52, Aij=529.0)
        self.set_interaction(41, 53, Aij=88.62999725)
        self.set_interaction(41, 54, Aij=1179.0)
        self.set_interaction(41, 55, Aij=123.40000153)
        self.set_interaction(41, 56, Aij=529.0)
        self.set_interaction(41, 57, Aij=-247.80000305)
        self.set_interaction(41, 58, Aij=88.62999725)
        self.set_interaction(41, 59, Aij=529.0)
        self.set_interaction(41, 60, Aij=1179.0)
        self.set_interaction(41, 61, Aij=88.62999725)
        self.set_interaction(41, 62, Aij=529.0)
        self.set_interaction(41, 63, Aij=1179.0)
        self.set_interaction(41, 64, Aij=88.62999725)
        self.set_interaction(41, 65, Aij=1179.0)
        self.set_interaction(41, 66, Aij=529.0)
        self.set_interaction(41, 67, Aij=529.0)
        self.set_interaction(41, 68, Aij=529.0)
        self.set_interaction(41, 69, Aij=88.62999725)
        self.set_interaction(42, 1, Aij=-34.36000061)
        self.set_interaction(42, 3, Aij=787.90002441)
        self.set_interaction(42, 5, Aij=1913.0)
        self.set_interaction(42, 7, Aij=180.19999695)
        self.set_interaction(42, 9, Aij=992.40002441)
        self.set_interaction(42, 13, Aij=448.5)
        self.set_interaction(42, 14, Aij=961.79998779)
        self.set_interaction(42, 15, Aij=1464.0)
        self.set_interaction(42, 43, Aij=-2166.0)
        self.set_interaction(42, 52, Aij=-34.36000061)
        self.set_interaction(42, 53, Aij=1913.0)
        self.set_interaction(42, 55, Aij=992.40002441)
        self.set_interaction(42, 56, Aij=-34.36000061)
        self.set_interaction(42, 57, Aij=448.5)
        self.set_interaction(42, 58, Aij=1913.0)
        self.set_interaction(42, 59, Aij=-34.36000061)
        self.set_interaction(42, 61, Aij=1913.0)
        self.set_interaction(42, 62, Aij=-34.36000061)
        self.set_interaction(42, 64, Aij=1913.0)
        self.set_interaction(42, 66, Aij=-34.36000061)
        self.set_interaction(42, 67, Aij=-34.36000061)
        self.set_interaction(42, 68, Aij=-34.36000061)
        self.set_interaction(42, 69, Aij=1913.0)
        self.set_interaction(43, 1, Aij=110.19999695)
        self.set_interaction(43, 3, Aij=234.3999939)
        self.set_interaction(43, 14, Aij=-125.19999695)
        self.set_interaction(43, 15, Aij=1604.0)
        self.set_interaction(43, 24, Aij=70.80999756)
        self.set_interaction(43, 42, Aij=745.29998779)
        self.set_interaction(43, 52, Aij=110.19999695)
        self.set_interaction(43, 56, Aij=110.19999695)
        self.set_interaction(43, 59, Aij=110.19999695)
        self.set_interaction(43, 62, Aij=110.19999695)
        self.set_interaction(43, 66, Aij=110.19999695)
        self.set_interaction(43, 67, Aij=110.19999695)
        self.set_interaction(43, 68, Aij=110.19999695)
        self.set_interaction(44, 1, Aij=13.89000034)
        self.set_interaction(44, 2, Aij=-16.11000061)
        self.set_interaction(44, 3, Aij=-23.87999916)
        self.set_interaction(44, 4, Aij=6.21400023)
        self.set_interaction(44, 5, Aij=796.90002441)
        self.set_interaction(44, 7, Aij=832.20001221)
        self.set_interaction(44, 8, Aij=-234.69999695)
        self.set_interaction(44, 23, Aij=-196.19999695)
        self.set_interaction(44, 25, Aij=161.5)
        self.set_interaction(44, 29, Aij=-274.1000061)
        self.set_interaction(44, 31, Aij=262.0)
        self.set_interaction(44, 37, Aij=-66.30999756)
        self.set_interaction(44, 52, Aij=13.89000034)
        self.set_interaction(44, 53, Aij=796.90002441)
        self.set_interaction(44, 56, Aij=13.89000034)
        self.set_interaction(44, 58, Aij=796.90002441)
        self.set_interaction(44, 59, Aij=13.89000034)
        self.set_interaction(44, 61, Aij=796.90002441)
        self.set_interaction(44, 62, Aij=13.89000034)
        self.set_interaction(44, 64, Aij=796.90002441)
        self.set_interaction(44, 66, Aij=13.89000034)
        self.set_interaction(44, 67, Aij=13.89000034)
        self.set_interaction(44, 68, Aij=13.89000034)
        self.set_interaction(44, 69, Aij=796.90002441)
        self.set_interaction(45, 1, Aij=30.73999977)
        self.set_interaction(45, 3, Aij=167.8999939)
        self.set_interaction(45, 5, Aij=794.40002441)
        self.set_interaction(45, 6, Aij=762.70001221)
        self.set_interaction(45, 26, Aij=844.0)
        self.set_interaction(45, 40, Aij=-32.16999817)
        self.set_interaction(45, 52, Aij=30.73999977)
        self.set_interaction(45, 53, Aij=794.40002441)
        self.set_interaction(45, 56, Aij=30.73999977)
        self.set_interaction(45, 58, Aij=794.40002441)
        self.set_interaction(45, 59, Aij=30.73999977)
        self.set_interaction(45, 61, Aij=794.40002441)
        self.set_interaction(45, 62, Aij=30.73999977)
        self.set_interaction(45, 64, Aij=794.40002441)
        self.set_interaction(45, 66, Aij=30.73999977)
        self.set_interaction(45, 67, Aij=30.73999977)
        self.set_interaction(45, 68, Aij=30.73999977)
        self.set_interaction(45, 69, Aij=794.40002441)
        self.set_interaction(46, 1, Aij=27.96999931)
        self.set_interaction(46, 2, Aij=9.75500011)
        self.set_interaction(46, 5, Aij=394.79998779)
        self.set_interaction(46, 7, Aij=-509.29998779)
        self.set_interaction(46, 20, Aij=-70.25)
        self.set_interaction(46, 52, Aij=27.96999931)
        self.set_interaction(46, 53, Aij=394.79998779)
        self.set_interaction(46, 54, Aij=-70.25)
        self.set_interaction(46, 56, Aij=27.96999931)
        self.set_interaction(46, 58, Aij=394.79998779)
        self.set_interaction(46, 59, Aij=27.96999931)
        self.set_interaction(46, 60, Aij=-70.25)
        self.set_interaction(46, 61, Aij=394.79998779)
        self.set_interaction(46, 62, Aij=27.96999931)
        self.set_interaction(46, 63, Aij=-70.25)
        self.set_interaction(46, 64, Aij=394.79998779)
        self.set_interaction(46, 65, Aij=-70.25)
        self.set_interaction(46, 66, Aij=27.96999931)
        self.set_interaction(46, 67, Aij=27.96999931)
        self.set_interaction(46, 68, Aij=27.96999931)
        self.set_interaction(46, 69, Aij=394.79998779)
        self.set_interaction(47, 1, Aij=-11.92000008)
        self.set_interaction(47, 2, Aij=132.3999939)
        self.set_interaction(47, 3, Aij=-86.87999725)
        self.set_interaction(47, 4, Aij=-19.45000076)
        self.set_interaction(47, 5, Aij=517.5)
        self.set_interaction(47, 7, Aij=-205.69999695)
        self.set_interaction(47, 9, Aij=156.3999939)
        self.set_interaction(47, 11, Aij=-3.44400001)
        self.set_interaction(47, 19, Aij=119.19999695)
        self.set_interaction(47, 22, Aij=-194.69999695)
        self.set_interaction(47, 24, Aij=3.16300011)
        self.set_interaction(47, 25, Aij=7.08199978)
        self.set_interaction(47, 31, Aij=515.79998779)
        self.set_interaction(47, 41, Aij=101.19999695)
        self.set_interaction(47, 52, Aij=-11.92000008)
        self.set_interaction(47, 53, Aij=517.5)
        self.set_interaction(47, 55, Aij=156.3999939)
        self.set_interaction(47, 56, Aij=-11.92000008)
        self.set_interaction(47, 58, Aij=517.5)
        self.set_interaction(47, 59, Aij=-11.92000008)
        self.set_interaction(47, 61, Aij=517.5)
        self.set_interaction(47, 62, Aij=-11.92000008)
        self.set_interaction(47, 64, Aij=517.5)
        self.set_interaction(47, 66, Aij=-11.92000008)
        self.set_interaction(47, 67, Aij=-11.92000008)
        self.set_interaction(47, 68, Aij=-11.92000008)
        self.set_interaction(47, 69, Aij=517.5)
        self.set_interaction(48, 1, Aij=39.93000031)
        self.set_interaction(48, 2, Aij=543.59997559)
        self.set_interaction(48, 6, Aij=420.0)
        self.set_interaction(48, 23, Aij=-363.1000061)
        self.set_interaction(48, 24, Aij=-11.30000019)
        self.set_interaction(48, 29, Aij=6.97100019)
        self.set_interaction(48, 37, Aij=148.8999939)
        self.set_interaction(48, 52, Aij=39.93000031)
        self.set_interaction(48, 56, Aij=39.93000031)
        self.set_interaction(48, 59, Aij=39.93000031)
        self.set_interaction(48, 62, Aij=39.93000031)
        self.set_interaction(48, 66, Aij=39.93000031)
        self.set_interaction(48, 67, Aij=39.93000031)
        self.set_interaction(48, 68, Aij=39.93000031)
        self.set_interaction(49, 1, Aij=-23.61000061)
        self.set_interaction(49, 2, Aij=161.1000061)
        self.set_interaction(49, 3, Aij=142.8999939)
        self.set_interaction(49, 4, Aij=274.1000061)
        self.set_interaction(49, 5, Aij=-61.20000076)
        self.set_interaction(49, 6, Aij=-89.23999786)
        self.set_interaction(49, 7, Aij=-384.29998779)
        self.set_interaction(49, 52, Aij=-23.61000061)
        self.set_interaction(49, 53, Aij=-61.20000076)
        self.set_interaction(49, 56, Aij=-23.61000061)
        self.set_interaction(49, 58, Aij=-61.20000076)
        self.set_interaction(49, 59, Aij=-23.61000061)
        self.set_interaction(49, 61, Aij=-61.20000076)
        self.set_interaction(49, 62, Aij=-23.61000061)
        self.set_interaction(49, 64, Aij=-61.20000076)
        self.set_interaction(49, 66, Aij=-23.61000061)
        self.set_interaction(49, 67, Aij=-23.61000061)
        self.set_interaction(49, 68, Aij=-23.61000061)
        self.set_interaction(49, 69, Aij=-61.20000076)
        self.set_interaction(50, 1, Aij=-8.47900009)
        self.set_interaction(50, 3, Aij=23.93000031)
        self.set_interaction(50, 4, Aij=2.84500003)
        self.set_interaction(50, 5, Aij=682.5)
        self.set_interaction(50, 6, Aij=597.79998779)
        self.set_interaction(50, 8, Aij=810.5)
        self.set_interaction(50, 9, Aij=278.79998779)
        self.set_interaction(50, 18, Aij=221.3999939)
        self.set_interaction(50, 24, Aij=-79.33999634)
        self.set_interaction(50, 26, Aij=176.30000305)
        self.set_interaction(50, 52, Aij=-8.47900009)
        self.set_interaction(50, 53, Aij=682.5)
        self.set_interaction(50, 55, Aij=278.79998779)
        self.set_interaction(50, 56, Aij=-8.47900009)
        self.set_interaction(50, 58, Aij=682.5)
        self.set_interaction(50, 59, Aij=-8.47900009)
        self.set_interaction(50, 61, Aij=682.5)
        self.set_interaction(50, 62, Aij=-8.47900009)
        self.set_interaction(50, 64, Aij=682.5)
        self.set_interaction(50, 66, Aij=-8.47900009)
        self.set_interaction(50, 67, Aij=-8.47900009)
        self.set_interaction(50, 68, Aij=-8.47900009)
        self.set_interaction(50, 69, Aij=682.5)
        self.set_interaction(52, 2, Aij=86.01999664)
        self.set_interaction(52, 3, Aij=61.13000107)
        self.set_interaction(52, 4, Aij=76.5)
        self.set_interaction(52, 5, Aij=986.5)
        self.set_interaction(52, 6, Aij=697.20001221)
        self.set_interaction(52, 7, Aij=1318.0)
        self.set_interaction(52, 8, Aij=1333.0)
        self.set_interaction(52, 9, Aij=476.3999939)
        self.set_interaction(52, 10, Aij=677.0)
        self.set_interaction(52, 11, Aij=232.1000061)
        self.set_interaction(52, 12, Aij=507.0)
        self.set_interaction(52, 13, Aij=251.5)
        self.set_interaction(52, 14, Aij=391.5)
        self.set_interaction(52, 15, Aij=255.69999695)
        self.set_interaction(52, 16, Aij=206.6000061)
        self.set_interaction(52, 17, Aij=920.70001221)
        self.set_interaction(52, 18, Aij=287.79998779)
        self.set_interaction(52, 19, Aij=597.0)
        self.set_interaction(52, 20, Aij=663.5)
        self.set_interaction(52, 21, Aij=35.93000031)
        self.set_interaction(52, 22, Aij=53.75999832)
        self.set_interaction(52, 23, Aij=24.89999962)
        self.set_interaction(52, 24, Aij=104.30000305)
        self.set_interaction(52, 25, Aij=11.43999958)
        self.set_interaction(52, 26, Aij=661.5)
        self.set_interaction(52, 27, Aij=543.0)
        self.set_interaction(52, 28, Aij=153.6000061)
        self.set_interaction(52, 29, Aij=184.3999939)
        self.set_interaction(52, 30, Aij=354.6000061)
        self.set_interaction(52, 31, Aij=3025.0)
        self.set_interaction(52, 32, Aij=335.79998779)
        self.set_interaction(52, 33, Aij=479.5)
        self.set_interaction(52, 34, Aij=298.8999939)
        self.set_interaction(52, 35, Aij=526.5)
        self.set_interaction(52, 36, Aij=689.0)
        self.set_interaction(52, 37, Aij=-4.18900013)
        self.set_interaction(52, 38, Aij=125.80000305)
        self.set_interaction(52, 39, Aij=485.29998779)
        self.set_interaction(52, 40, Aij=-2.85899997)
        self.set_interaction(52, 41, Aij=387.1000061)
        self.set_interaction(52, 42, Aij=-450.3999939)
        self.set_interaction(52, 43, Aij=252.69999695)
        self.set_interaction(52, 44, Aij=220.30000305)
        self.set_interaction(52, 45, Aij=-5.86899996)
        self.set_interaction(52, 46, Aij=390.8999939)
        self.set_interaction(52, 47, Aij=553.29998779)
        self.set_interaction(52, 48, Aij=187.0)
        self.set_interaction(52, 49, Aij=216.1000061)
        self.set_interaction(52, 50, Aij=92.98999786)
        self.set_interaction(52, 53, Aij=986.5)
        self.set_interaction(52, 54, Aij=663.5)
        self.set_interaction(52, 55, Aij=476.3999939)
        self.set_interaction(52, 57, Aij=251.5)
        self.set_interaction(52, 58, Aij=986.5)
        self.set_interaction(52, 60, Aij=663.5)
        self.set_interaction(52, 61, Aij=986.5)
        self.set_interaction(52, 63, Aij=663.5)
        self.set_interaction(52, 64, Aij=986.5)
        self.set_interaction(52, 65, Aij=663.5)
        self.set_interaction(52, 69, Aij=986.5)
        self.set_interaction(52, 71, Aij=500.95)
        self.set_interaction(52, 72, Aij=977.56)
        self.set_interaction(52, 73, Aij=1331.0)
        self.set_interaction(52, 74, Aij=297.24)
        self.set_interaction(52, 75, Aij=528.5)
        self.set_interaction(53, 1, Aij=156.3999939)
        self.set_interaction(53, 2, Aij=457.0)
        self.set_interaction(53, 3, Aij=89.59999847)
        self.set_interaction(53, 4, Aij=25.81999969)
        self.set_interaction(53, 6, Aij=-137.1000061)
        self.set_interaction(53, 7, Aij=1000.0)
        self.set_interaction(53, 8, Aij=-259.70001221)
        self.set_interaction(53, 9, Aij=84.0)
        self.set_interaction(53, 10, Aij=-203.6000061)
        self.set_interaction(53, 11, Aij=101.09999847)
        self.set_interaction(53, 12, Aij=267.79998779)
        self.set_interaction(53, 13, Aij=28.05999947)
        self.set_interaction(53, 14, Aij=8.6420002)
        self.set_interaction(53, 15, Aij=42.70000076)
        self.set_interaction(53, 16, Aij=-323.0)
        self.set_interaction(53, 17, Aij=-52.38999939)
        self.set_interaction(53, 18, Aij=170.0)
        self.set_interaction(53, 19, Aij=6.71199989)
        self.set_interaction(53, 20, Aij=199.0)
        self.set_interaction(53, 21, Aij=75.62000275)
        self.set_interaction(53, 22, Aij=65.27999878)
        self.set_interaction(53, 23, Aij=-98.12000275)
        self.set_interaction(53, 24, Aij=143.1000061)
        self.set_interaction(53, 25, Aij=123.5)
        self.set_interaction(53, 26, Aij=256.5)
        self.set_interaction(53, 27, Aij=157.1000061)
        self.set_interaction(53, 28, Aij=488.8999939)
        self.set_interaction(53, 29, Aij=147.5)
        self.set_interaction(53, 30, Aij=-120.5)
        self.set_interaction(53, 31, Aij=-318.8999939)
        self.set_interaction(53, 32, Aij=313.5)
        self.set_interaction(53, 33, Aij=202.1000061)
        self.set_interaction(53, 34, Aij=727.79998779)
        self.set_interaction(53, 35, Aij=-202.1000061)
        self.set_interaction(53, 36, Aij=74.26999664)
        self.set_interaction(53, 37, Aij=225.80000305)
        self.set_interaction(53, 38, Aij=44.77999878)
        self.set_interaction(53, 39, Aij=-143.8999939)
        self.set_interaction(53, 41, Aij=190.30000305)
        self.set_interaction(53, 42, Aij=-817.70001221)
        self.set_interaction(53, 44, Aij=-504.20001221)
        self.set_interaction(53, 45, Aij=72.95999908)
        self.set_interaction(53, 46, Aij=-382.70001221)
        self.set_interaction(53, 47, Aij=-248.30000305)
        self.set_interaction(53, 49, Aij=104.69999695)
        self.set_interaction(53, 50, Aij=57.65000153)
        self.set_interaction(53, 52, Aij=156.3999939)
        self.set_interaction(53, 54, Aij=199.0)
        self.set_interaction(53, 55, Aij=84.0)
        self.set_interaction(53, 56, Aij=156.3999939)
        self.set_interaction(53, 57, Aij=28.05999947)
        self.set_interaction(53, 59, Aij=156.3999939)
        self.set_interaction(53, 60, Aij=199.0)
        self.set_interaction(53, 62, Aij=156.3999939)
        self.set_interaction(53, 63, Aij=199.0)
        self.set_interaction(53, 65, Aij=199.0)
        self.set_interaction(53, 66, Aij=156.3999939)
        self.set_interaction(53, 67, Aij=156.3999939)
        self.set_interaction(53, 68, Aij=156.3999939)
        self.set_interaction(53, 70, Aij=-843.09002686)
        self.set_interaction(53, 71, Aij=37.631)
        self.set_interaction(53, 72, Aij=-330.28)
        self.set_interaction(53, 73, Aij=1789.0)
        self.set_interaction(53, 74, Aij=221.38)
        self.set_interaction(53, 75, Aij=-77.526)
        self.set_interaction(54, 1, Aij=315.29998779)
        self.set_interaction(54, 2, Aij=1264.0)
        self.set_interaction(54, 3, Aij=62.31999969)
        self.set_interaction(54, 4, Aij=89.86000061)
        self.set_interaction(54, 5, Aij=-151.0)
        self.set_interaction(54, 6, Aij=339.79998779)
        self.set_interaction(54, 7, Aij=1000.0)
        self.set_interaction(54, 8, Aij=-11.0)
        self.set_interaction(54, 9, Aij=-297.79998779)
        self.set_interaction(54, 10, Aij=-165.5)
        self.set_interaction(54, 11, Aij=-256.29998779)
        self.set_interaction(54, 12, Aij=193.8999939)
        self.set_interaction(54, 13, Aij=-338.5)
        self.set_interaction(54, 17, Aij=493.79998779)
        self.set_interaction(54, 18, Aij=-313.5)
        self.set_interaction(54, 21, Aij=13.40999985)
        self.set_interaction(54, 22, Aij=-44.70000076)
        self.set_interaction(54, 23, Aij=39.63000107)
        self.set_interaction(54, 24, Aij=183.3999939)
        self.set_interaction(54, 25, Aij=-79.08000183)
        self.set_interaction(54, 30, Aij=-208.8999939)
        self.set_interaction(54, 32, Aij=228.3999939)
        self.set_interaction(54, 33, Aij=-95.0)
        self.set_interaction(54, 35, Aij=-463.6000061)
        self.set_interaction(54, 37, Aij=-11.15999985)
        self.set_interaction(54, 39, Aij=-228.0)
        self.set_interaction(54, 41, Aij=-337.0)
        self.set_interaction(54, 46, Aij=-322.29998779)
        self.set_interaction(54, 52, Aij=315.29998779)
        self.set_interaction(54, 53, Aij=-151.0)
        self.set_interaction(54, 55, Aij=-297.79998779)
        self.set_interaction(54, 56, Aij=315.29998779)
        self.set_interaction(54, 57, Aij=-338.5)
        self.set_interaction(54, 58, Aij=-151.0)
        self.set_interaction(54, 59, Aij=315.29998779)
        self.set_interaction(54, 61, Aij=-151.0)
        self.set_interaction(54, 62, Aij=315.29998779)
        self.set_interaction(54, 64, Aij=-151.0)
        self.set_interaction(54, 66, Aij=315.29998779)
        self.set_interaction(54, 67, Aij=315.29998779)
        self.set_interaction(54, 68, Aij=315.29998779)
        self.set_interaction(54, 69, Aij=-151.0)
        self.set_interaction(54, 71, Aij=-100.17)
        self.set_interaction(54, 72, Aij=-501.23)
        self.set_interaction(54, 73, Aij=-579.8)
        self.set_interaction(54, 74, Aij=79.052)
        self.set_interaction(54, 75, Aij=-340.95)
        self.set_interaction(55, 1, Aij=26.76000023)
        self.set_interaction(55, 2, Aij=42.91999817)
        self.set_interaction(55, 3, Aij=140.1000061)
        self.set_interaction(55, 4, Aij=365.79998779)
        self.set_interaction(55, 5, Aij=164.5)
        self.set_interaction(55, 6, Aij=108.69999695)
        self.set_interaction(55, 7, Aij=1000.0)
        self.set_interaction(55, 8, Aij=-133.1000061)
        self.set_interaction(55, 10, Aij=-37.36000061)
        self.set_interaction(55, 11, Aij=-213.69999695)
        self.set_interaction(55, 12, Aij=-190.3999939)
        self.set_interaction(55, 13, Aij=-103.59999847)
        self.set_interaction(55, 15, Aij=-174.19999695)
        self.set_interaction(55, 16, Aij=-169.0)
        self.set_interaction(55, 17, Aij=6201.0)
        self.set_interaction(55, 18, Aij=7.34100008)
        self.set_interaction(55, 19, Aij=481.70001221)
        self.set_interaction(55, 20, Aij=669.40002441)
        self.set_interaction(55, 21, Aij=-191.69999695)
        self.set_interaction(55, 22, Aij=-130.30000305)
        self.set_interaction(55, 23, Aij=-354.6000061)
        self.set_interaction(55, 24, Aij=-39.20000076)
        self.set_interaction(55, 25, Aij=-119.80000305)
        self.set_interaction(55, 26, Aij=137.5)
        self.set_interaction(55, 27, Aij=548.5)
        self.set_interaction(55, 28, Aij=216.1000061)
        self.set_interaction(55, 29, Aij=-46.27999878)
        self.set_interaction(55, 30, Aij=-163.69999695)
        self.set_interaction(55, 31, Aij=71.45999908)
        self.set_interaction(55, 32, Aij=53.59000015)
        self.set_interaction(55, 33, Aij=245.19999695)
        self.set_interaction(55, 34, Aij=-246.6000061)
        self.set_interaction(55, 35, Aij=-44.58000183)
        self.set_interaction(55, 37, Aij=-34.56999969)
        self.set_interaction(55, 39, Aij=-61.70000076)
        self.set_interaction(55, 41, Aij=-18.79999924)
        self.set_interaction(55, 42, Aij=-588.90002441)
        self.set_interaction(55, 47, Aij=37.54000092)
        self.set_interaction(55, 50, Aij=-162.6000061)
        self.set_interaction(55, 52, Aij=26.76000023)
        self.set_interaction(55, 53, Aij=164.5)
        self.set_interaction(55, 54, Aij=669.40002441)
        self.set_interaction(55, 56, Aij=26.76000023)
        self.set_interaction(55, 57, Aij=-103.59999847)
        self.set_interaction(55, 58, Aij=164.5)
        self.set_interaction(55, 59, Aij=26.76000023)
        self.set_interaction(55, 60, Aij=669.40002441)
        self.set_interaction(55, 61, Aij=164.5)
        self.set_interaction(55, 62, Aij=26.76000023)
        self.set_interaction(55, 63, Aij=669.40002441)
        self.set_interaction(55, 64, Aij=164.5)
        self.set_interaction(55, 65, Aij=669.40002441)
        self.set_interaction(55, 66, Aij=26.76000023)
        self.set_interaction(55, 67, Aij=26.76000023)
        self.set_interaction(55, 68, Aij=26.76000023)
        self.set_interaction(55, 69, Aij=164.5)
        self.set_interaction(56, 2, Aij=86.01999664)
        self.set_interaction(56, 3, Aij=61.13000107)
        self.set_interaction(56, 4, Aij=76.5)
        self.set_interaction(56, 5, Aij=986.5)
        self.set_interaction(56, 6, Aij=697.20001221)
        self.set_interaction(56, 7, Aij=1318.0)
        self.set_interaction(56, 8, Aij=1333.0)
        self.set_interaction(56, 9, Aij=476.3999939)
        self.set_interaction(56, 10, Aij=677.0)
        self.set_interaction(56, 11, Aij=232.1000061)
        self.set_interaction(56, 12, Aij=507.0)
        self.set_interaction(56, 13, Aij=251.5)
        self.set_interaction(56, 14, Aij=391.5)
        self.set_interaction(56, 15, Aij=255.69999695)
        self.set_interaction(56, 16, Aij=206.6000061)
        self.set_interaction(56, 17, Aij=920.70001221)
        self.set_interaction(56, 18, Aij=287.79998779)
        self.set_interaction(56, 19, Aij=597.0)
        self.set_interaction(56, 20, Aij=663.5)
        self.set_interaction(56, 21, Aij=35.93000031)
        self.set_interaction(56, 22, Aij=53.75999832)
        self.set_interaction(56, 23, Aij=24.89999962)
        self.set_interaction(56, 24, Aij=104.30000305)
        self.set_interaction(56, 25, Aij=11.43999958)
        self.set_interaction(56, 26, Aij=661.5)
        self.set_interaction(56, 27, Aij=543.0)
        self.set_interaction(56, 28, Aij=153.6000061)
        self.set_interaction(56, 29, Aij=184.3999939)
        self.set_interaction(56, 30, Aij=354.6000061)
        self.set_interaction(56, 31, Aij=3025.0)
        self.set_interaction(56, 32, Aij=335.79998779)
        self.set_interaction(56, 33, Aij=479.5)
        self.set_interaction(56, 34, Aij=298.8999939)
        self.set_interaction(56, 35, Aij=526.5)
        self.set_interaction(56, 36, Aij=689.0)
        self.set_interaction(56, 37, Aij=-4.18900013)
        self.set_interaction(56, 38, Aij=125.80000305)
        self.set_interaction(56, 39, Aij=485.29998779)
        self.set_interaction(56, 40, Aij=-2.85899997)
        self.set_interaction(56, 41, Aij=387.1000061)
        self.set_interaction(56, 42, Aij=-450.3999939)
        self.set_interaction(56, 43, Aij=252.69999695)
        self.set_interaction(56, 44, Aij=220.30000305)
        self.set_interaction(56, 45, Aij=-5.86899996)
        self.set_interaction(56, 46, Aij=390.8999939)
        self.set_interaction(56, 47, Aij=553.29998779)
        self.set_interaction(56, 48, Aij=187.0)
        self.set_interaction(56, 49, Aij=216.1000061)
        self.set_interaction(56, 50, Aij=92.98999786)
        self.set_interaction(56, 53, Aij=986.5)
        self.set_interaction(56, 54, Aij=663.5)
        self.set_interaction(56, 55, Aij=476.3999939)
        self.set_interaction(56, 57, Aij=251.5)
        self.set_interaction(56, 58, Aij=986.5)
        self.set_interaction(56, 60, Aij=663.5)
        self.set_interaction(56, 61, Aij=986.5)
        self.set_interaction(56, 63, Aij=663.5)
        self.set_interaction(56, 64, Aij=986.5)
        self.set_interaction(56, 65, Aij=663.5)
        self.set_interaction(56, 69, Aij=986.5)
        self.set_interaction(56, 71, Aij=500.95)
        self.set_interaction(56, 72, Aij=977.56)
        self.set_interaction(56, 73, Aij=1331.0)
        self.set_interaction(56, 74, Aij=297.24)
        self.set_interaction(56, 75, Aij=528.5)
        self.set_interaction(57, 1, Aij=83.36000061)
        self.set_interaction(57, 2, Aij=26.51000023)
        self.set_interaction(57, 3, Aij=52.13000107)
        self.set_interaction(57, 4, Aij=65.69000244)
        self.set_interaction(57, 5, Aij=237.69999695)
        self.set_interaction(57, 6, Aij=238.3999939)
        self.set_interaction(57, 7, Aij=2007.0)
        self.set_interaction(57, 8, Aij=-178.5)
        self.set_interaction(57, 9, Aij=191.1000061)
        self.set_interaction(57, 10, Aij=-7.83799982)
        self.set_interaction(57, 11, Aij=461.29998779)
        self.set_interaction(57, 12, Aij=457.29998779)
        self.set_interaction(57, 14, Aij=-78.36000061)
        self.set_interaction(57, 15, Aij=251.5)
        self.set_interaction(57, 16, Aij=5422.0)
        self.set_interaction(57, 18, Aij=213.19999695)
        self.set_interaction(57, 19, Aij=-18.51000023)
        self.set_interaction(57, 20, Aij=664.59997559)
        self.set_interaction(57, 21, Aij=301.1000061)
        self.set_interaction(57, 22, Aij=137.80000305)
        self.set_interaction(57, 23, Aij=-154.30000305)
        self.set_interaction(57, 24, Aij=47.66999817)
        self.set_interaction(57, 25, Aij=134.80000305)
        self.set_interaction(57, 26, Aij=95.18000031)
        self.set_interaction(57, 28, Aij=140.8999939)
        self.set_interaction(57, 29, Aij=-8.53800011)
        self.set_interaction(57, 30, Aij=170.1000061)
        self.set_interaction(57, 31, Aij=-20.11000061)
        self.set_interaction(57, 32, Aij=-149.5)
        self.set_interaction(57, 33, Aij=-202.30000305)
        self.set_interaction(57, 35, Aij=128.80000305)
        self.set_interaction(57, 37, Aij=240.19999695)
        self.set_interaction(57, 38, Aij=-274.0)
        self.set_interaction(57, 39, Aij=254.80000305)
        self.set_interaction(57, 41, Aij=417.0)
        self.set_interaction(57, 42, Aij=1338.0)
        self.set_interaction(57, 52, Aij=83.36000061)
        self.set_interaction(57, 53, Aij=237.69999695)
        self.set_interaction(57, 54, Aij=664.59997559)
        self.set_interaction(57, 55, Aij=191.1000061)
        self.set_interaction(57, 56, Aij=83.36000061)
        self.set_interaction(57, 58, Aij=237.69999695)
        self.set_interaction(57, 59, Aij=83.36000061)
        self.set_interaction(57, 60, Aij=664.59997559)
        self.set_interaction(57, 61, Aij=237.69999695)
        self.set_interaction(57, 62, Aij=83.36000061)
        self.set_interaction(57, 63, Aij=664.59997559)
        self.set_interaction(57, 64, Aij=237.69999695)
        self.set_interaction(57, 65, Aij=664.59997559)
        self.set_interaction(57, 66, Aij=83.36000061)
        self.set_interaction(57, 67, Aij=83.36000061)
        self.set_interaction(57, 68, Aij=83.36000061)
        self.set_interaction(57, 69, Aij=237.69999695)
        self.set_interaction(58, 1, Aij=156.3999939)
        self.set_interaction(58, 2, Aij=457.0)
        self.set_interaction(58, 3, Aij=89.59999847)
        self.set_interaction(58, 4, Aij=25.81999969)
        self.set_interaction(58, 6, Aij=-137.1000061)
        self.set_interaction(58, 7, Aij=-189.69999695)
        self.set_interaction(58, 8, Aij=-259.70001221)
        self.set_interaction(58, 9, Aij=84.0)
        self.set_interaction(58, 10, Aij=-203.6000061)
        self.set_interaction(58, 11, Aij=101.09999847)
        self.set_interaction(58, 12, Aij=267.79998779)
        self.set_interaction(58, 13, Aij=28.05999947)
        self.set_interaction(58, 14, Aij=8.6420002)
        self.set_interaction(58, 15, Aij=42.70000076)
        self.set_interaction(58, 16, Aij=-323.0)
        self.set_interaction(58, 17, Aij=-52.38999939)
        self.set_interaction(58, 18, Aij=170.0)
        self.set_interaction(58, 19, Aij=6.71199989)
        self.set_interaction(58, 20, Aij=199.0)
        self.set_interaction(58, 21, Aij=75.62000275)
        self.set_interaction(58, 22, Aij=65.27999878)
        self.set_interaction(58, 23, Aij=-98.12000275)
        self.set_interaction(58, 24, Aij=143.1000061)
        self.set_interaction(58, 25, Aij=123.5)
        self.set_interaction(58, 26, Aij=256.5)
        self.set_interaction(58, 27, Aij=157.1000061)
        self.set_interaction(58, 28, Aij=488.8999939)
        self.set_interaction(58, 29, Aij=147.5)
        self.set_interaction(58, 30, Aij=-120.5)
        self.set_interaction(58, 31, Aij=-318.8999939)
        self.set_interaction(58, 32, Aij=313.5)
        self.set_interaction(58, 33, Aij=202.1000061)
        self.set_interaction(58, 34, Aij=727.79998779)
        self.set_interaction(58, 35, Aij=-202.1000061)
        self.set_interaction(58, 36, Aij=74.26999664)
        self.set_interaction(58, 37, Aij=225.80000305)
        self.set_interaction(58, 38, Aij=44.77999878)
        self.set_interaction(58, 39, Aij=-143.8999939)
        self.set_interaction(58, 41, Aij=190.30000305)
        self.set_interaction(58, 42, Aij=-817.70001221)
        self.set_interaction(58, 44, Aij=-504.20001221)
        self.set_interaction(58, 45, Aij=72.95999908)
        self.set_interaction(58, 46, Aij=-382.70001221)
        self.set_interaction(58, 47, Aij=-248.30000305)
        self.set_interaction(58, 49, Aij=104.69999695)
        self.set_interaction(58, 50, Aij=57.65000153)
        self.set_interaction(58, 52, Aij=156.3999939)
        self.set_interaction(58, 54, Aij=199.0)
        self.set_interaction(58, 55, Aij=84.0)
        self.set_interaction(58, 56, Aij=156.3999939)
        self.set_interaction(58, 57, Aij=28.05999947)
        self.set_interaction(58, 59, Aij=156.3999939)
        self.set_interaction(58, 60, Aij=199.0)
        self.set_interaction(58, 62, Aij=156.3999939)
        self.set_interaction(58, 63, Aij=199.0)
        self.set_interaction(58, 65, Aij=199.0)
        self.set_interaction(58, 66, Aij=156.3999939)
        self.set_interaction(58, 67, Aij=156.3999939)
        self.set_interaction(58, 68, Aij=156.3999939)
        self.set_interaction(58, 70, Aij=-843.09002686)
        self.set_interaction(58, 71, Aij=37.631)
        self.set_interaction(58, 72, Aij=-330.28)
        self.set_interaction(58, 73, Aij=1789.0)
        self.set_interaction(58, 74, Aij=221.38)
        self.set_interaction(58, 75, Aij=-77.526)
        self.set_interaction(59, 2, Aij=86.01999664)
        self.set_interaction(59, 3, Aij=61.13000107)
        self.set_interaction(59, 4, Aij=76.5)
        self.set_interaction(59, 5, Aij=986.5)
        self.set_interaction(59, 6, Aij=697.20001221)
        self.set_interaction(59, 7, Aij=1318.0)
        self.set_interaction(59, 8, Aij=1333.0)
        self.set_interaction(59, 9, Aij=476.3999939)
        self.set_interaction(59, 10, Aij=677.0)
        self.set_interaction(59, 11, Aij=232.1000061)
        self.set_interaction(59, 12, Aij=507.0)
        self.set_interaction(59, 13, Aij=251.5)
        self.set_interaction(59, 14, Aij=391.5)
        self.set_interaction(59, 15, Aij=255.69999695)
        self.set_interaction(59, 16, Aij=206.6000061)
        self.set_interaction(59, 17, Aij=920.70001221)
        self.set_interaction(59, 18, Aij=287.79998779)
        self.set_interaction(59, 19, Aij=597.0)
        self.set_interaction(59, 20, Aij=663.5)
        self.set_interaction(59, 21, Aij=35.93000031)
        self.set_interaction(59, 22, Aij=53.75999832)
        self.set_interaction(59, 23, Aij=24.89999962)
        self.set_interaction(59, 24, Aij=104.30000305)
        self.set_interaction(59, 25, Aij=11.43999958)
        self.set_interaction(59, 26, Aij=661.5)
        self.set_interaction(59, 27, Aij=543.0)
        self.set_interaction(59, 28, Aij=153.6000061)
        self.set_interaction(59, 29, Aij=184.3999939)
        self.set_interaction(59, 30, Aij=354.6000061)
        self.set_interaction(59, 31, Aij=3025.0)
        self.set_interaction(59, 32, Aij=335.79998779)
        self.set_interaction(59, 33, Aij=479.5)
        self.set_interaction(59, 34, Aij=298.8999939)
        self.set_interaction(59, 35, Aij=526.5)
        self.set_interaction(59, 36, Aij=689.0)
        self.set_interaction(59, 37, Aij=-4.18900013)
        self.set_interaction(59, 38, Aij=125.80000305)
        self.set_interaction(59, 39, Aij=485.29998779)
        self.set_interaction(59, 40, Aij=-2.85899997)
        self.set_interaction(59, 41, Aij=387.1000061)
        self.set_interaction(59, 42, Aij=-450.3999939)
        self.set_interaction(59, 43, Aij=252.69999695)
        self.set_interaction(59, 44, Aij=220.30000305)
        self.set_interaction(59, 45, Aij=-5.86899996)
        self.set_interaction(59, 46, Aij=390.8999939)
        self.set_interaction(59, 47, Aij=553.29998779)
        self.set_interaction(59, 48, Aij=187.0)
        self.set_interaction(59, 49, Aij=216.1000061)
        self.set_interaction(59, 50, Aij=92.98999786)
        self.set_interaction(59, 53, Aij=986.5)
        self.set_interaction(59, 54, Aij=663.5)
        self.set_interaction(59, 55, Aij=476.3999939)
        self.set_interaction(59, 57, Aij=251.5)
        self.set_interaction(59, 58, Aij=986.5)
        self.set_interaction(59, 60, Aij=663.5)
        self.set_interaction(59, 61, Aij=986.5)
        self.set_interaction(59, 63, Aij=663.5)
        self.set_interaction(59, 64, Aij=986.5)
        self.set_interaction(59, 65, Aij=663.5)
        self.set_interaction(59, 69, Aij=986.5)
        self.set_interaction(60, 1, Aij=315.29998779)
        self.set_interaction(60, 2, Aij=1264.0)
        self.set_interaction(60, 3, Aij=62.31999969)
        self.set_interaction(60, 4, Aij=89.86000061)
        self.set_interaction(60, 5, Aij=-151.0)
        self.set_interaction(60, 6, Aij=339.79998779)
        self.set_interaction(60, 7, Aij=-163.30000305)
        self.set_interaction(60, 8, Aij=-11.0)
        self.set_interaction(60, 9, Aij=-297.79998779)
        self.set_interaction(60, 10, Aij=-165.5)
        self.set_interaction(60, 11, Aij=-256.29998779)
        self.set_interaction(60, 12, Aij=193.8999939)
        self.set_interaction(60, 13, Aij=-338.5)
        self.set_interaction(60, 17, Aij=493.79998779)
        self.set_interaction(60, 18, Aij=-313.5)
        self.set_interaction(60, 21, Aij=13.40999985)
        self.set_interaction(60, 22, Aij=-44.70000076)
        self.set_interaction(60, 23, Aij=39.63000107)
        self.set_interaction(60, 24, Aij=183.3999939)
        self.set_interaction(60, 25, Aij=-79.08000183)
        self.set_interaction(60, 30, Aij=-208.8999939)
        self.set_interaction(60, 32, Aij=228.3999939)
        self.set_interaction(60, 33, Aij=-95.0)
        self.set_interaction(60, 35, Aij=-463.6000061)
        self.set_interaction(60, 37, Aij=-11.15999985)
        self.set_interaction(60, 39, Aij=-228.0)
        self.set_interaction(60, 41, Aij=-337.0)
        self.set_interaction(60, 46, Aij=-322.29998779)
        self.set_interaction(60, 52, Aij=315.29998779)
        self.set_interaction(60, 53, Aij=-151.0)
        self.set_interaction(60, 55, Aij=-297.79998779)
        self.set_interaction(60, 56, Aij=315.29998779)
        self.set_interaction(60, 57, Aij=-338.5)
        self.set_interaction(60, 58, Aij=-151.0)
        self.set_interaction(60, 59, Aij=315.29998779)
        self.set_interaction(60, 61, Aij=-151.0)
        self.set_interaction(60, 62, Aij=315.29998779)
        self.set_interaction(60, 64, Aij=-151.0)
        self.set_interaction(60, 66, Aij=315.29998779)
        self.set_interaction(60, 67, Aij=315.29998779)
        self.set_interaction(60, 68, Aij=315.29998779)
        self.set_interaction(60, 69, Aij=-151.0)
        self.set_interaction(60, 71, Aij=-100.17)
        self.set_interaction(60, 72, Aij=-501.23)
        self.set_interaction(60, 73, Aij=-579.8)
        self.set_interaction(60, 74, Aij=79.052)
        self.set_interaction(60, 75, Aij=-340.95)
        self.set_interaction(61, 1, Aij=156.3999939)
        self.set_interaction(61, 2, Aij=457.0)
        self.set_interaction(61, 3, Aij=89.59999847)
        self.set_interaction(61, 4, Aij=25.81999969)
        self.set_interaction(61, 6, Aij=-137.1000061)
        self.set_interaction(61, 7, Aij=-92.30000305)
        self.set_interaction(61, 8, Aij=-259.70001221)
        self.set_interaction(61, 9, Aij=84.0)
        self.set_interaction(61, 10, Aij=-203.6000061)
        self.set_interaction(61, 11, Aij=101.09999847)
        self.set_interaction(61, 12, Aij=267.79998779)
        self.set_interaction(61, 13, Aij=28.05999947)
        self.set_interaction(61, 14, Aij=8.6420002)
        self.set_interaction(61, 15, Aij=42.70000076)
        self.set_interaction(61, 16, Aij=-323.0)
        self.set_interaction(61, 17, Aij=-52.38999939)
        self.set_interaction(61, 18, Aij=170.0)
        self.set_interaction(61, 19, Aij=6.71199989)
        self.set_interaction(61, 20, Aij=199.0)
        self.set_interaction(61, 21, Aij=75.62000275)
        self.set_interaction(61, 22, Aij=65.27999878)
        self.set_interaction(61, 23, Aij=-98.12000275)
        self.set_interaction(61, 24, Aij=143.1000061)
        self.set_interaction(61, 25, Aij=123.5)
        self.set_interaction(61, 26, Aij=256.5)
        self.set_interaction(61, 27, Aij=157.1000061)
        self.set_interaction(61, 28, Aij=488.8999939)
        self.set_interaction(61, 29, Aij=147.5)
        self.set_interaction(61, 30, Aij=-120.5)
        self.set_interaction(61, 31, Aij=-318.8999939)
        self.set_interaction(61, 32, Aij=313.5)
        self.set_interaction(61, 33, Aij=202.1000061)
        self.set_interaction(61, 34, Aij=727.79998779)
        self.set_interaction(61, 35, Aij=-202.1000061)
        self.set_interaction(61, 36, Aij=74.26999664)
        self.set_interaction(61, 37, Aij=225.80000305)
        self.set_interaction(61, 38, Aij=44.77999878)
        self.set_interaction(61, 39, Aij=-143.8999939)
        self.set_interaction(61, 41, Aij=190.30000305)
        self.set_interaction(61, 42, Aij=-817.70001221)
        self.set_interaction(61, 44, Aij=-504.20001221)
        self.set_interaction(61, 45, Aij=72.95999908)
        self.set_interaction(61, 46, Aij=-382.70001221)
        self.set_interaction(61, 47, Aij=-248.30000305)
        self.set_interaction(61, 49, Aij=104.69999695)
        self.set_interaction(61, 50, Aij=57.65000153)
        self.set_interaction(61, 52, Aij=156.3999939)
        self.set_interaction(61, 54, Aij=199.0)
        self.set_interaction(61, 55, Aij=84.0)
        self.set_interaction(61, 56, Aij=156.3999939)
        self.set_interaction(61, 57, Aij=28.05999947)
        self.set_interaction(61, 59, Aij=156.3999939)
        self.set_interaction(61, 60, Aij=199.0)
        self.set_interaction(61, 62, Aij=156.3999939)
        self.set_interaction(61, 63, Aij=199.0)
        self.set_interaction(61, 65, Aij=199.0)
        self.set_interaction(61, 66, Aij=156.3999939)
        self.set_interaction(61, 67, Aij=156.3999939)
        self.set_interaction(61, 68, Aij=156.3999939)
        self.set_interaction(61, 70, Aij=-843.09002686)
        self.set_interaction(61, 71, Aij=37.631)
        self.set_interaction(61, 72, Aij=-330.28)
        self.set_interaction(61, 73, Aij=1789.0)
        self.set_interaction(61, 74, Aij=221.38)
        self.set_interaction(61, 75, Aij=-77.526)
        self.set_interaction(62, 2, Aij=86.01999664)
        self.set_interaction(62, 3, Aij=61.13000107)
        self.set_interaction(62, 4, Aij=76.5)
        self.set_interaction(62, 5, Aij=986.5)
        self.set_interaction(62, 6, Aij=697.20001221)
        self.set_interaction(62, 7, Aij=4694.0)
        self.set_interaction(62, 8, Aij=1333.0)
        self.set_interaction(62, 9, Aij=476.3999939)
        self.set_interaction(62, 10, Aij=677.0)
        self.set_interaction(62, 11, Aij=232.1000061)
        self.set_interaction(62, 12, Aij=507.0)
        self.set_interaction(62, 13, Aij=251.5)
        self.set_interaction(62, 14, Aij=391.5)
        self.set_interaction(62, 15, Aij=255.69999695)
        self.set_interaction(62, 16, Aij=206.6000061)
        self.set_interaction(62, 17, Aij=920.70001221)
        self.set_interaction(62, 18, Aij=287.79998779)
        self.set_interaction(62, 19, Aij=597.0)
        self.set_interaction(62, 20, Aij=663.5)
        self.set_interaction(62, 21, Aij=35.93000031)
        self.set_interaction(62, 22, Aij=53.75999832)
        self.set_interaction(62, 23, Aij=24.89999962)
        self.set_interaction(62, 24, Aij=104.30000305)
        self.set_interaction(62, 25, Aij=11.43999958)
        self.set_interaction(62, 26, Aij=661.5)
        self.set_interaction(62, 27, Aij=543.0)
        self.set_interaction(62, 28, Aij=153.6000061)
        self.set_interaction(62, 29, Aij=184.3999939)
        self.set_interaction(62, 30, Aij=354.6000061)
        self.set_interaction(62, 31, Aij=3025.0)
        self.set_interaction(62, 32, Aij=335.79998779)
        self.set_interaction(62, 33, Aij=479.5)
        self.set_interaction(62, 34, Aij=298.8999939)
        self.set_interaction(62, 35, Aij=526.5)
        self.set_interaction(62, 36, Aij=689.0)
        self.set_interaction(62, 37, Aij=-4.18900013)
        self.set_interaction(62, 38, Aij=125.80000305)
        self.set_interaction(62, 39, Aij=485.29998779)
        self.set_interaction(62, 40, Aij=-2.85899997)
        self.set_interaction(62, 41, Aij=387.1000061)
        self.set_interaction(62, 42, Aij=-450.3999939)
        self.set_interaction(62, 43, Aij=252.69999695)
        self.set_interaction(62, 44, Aij=220.30000305)
        self.set_interaction(62, 45, Aij=-5.86899996)
        self.set_interaction(62, 46, Aij=390.8999939)
        self.set_interaction(62, 47, Aij=553.29998779)
        self.set_interaction(62, 48, Aij=187.0)
        self.set_interaction(62, 49, Aij=216.1000061)
        self.set_interaction(62, 50, Aij=92.98999786)
        self.set_interaction(62, 53, Aij=986.5)
        self.set_interaction(62, 54, Aij=663.5)
        self.set_interaction(62, 55, Aij=476.3999939)
        self.set_interaction(62, 57, Aij=251.5)
        self.set_interaction(62, 58, Aij=986.5)
        self.set_interaction(62, 60, Aij=663.5)
        self.set_interaction(62, 61, Aij=986.5)
        self.set_interaction(62, 63, Aij=663.5)
        self.set_interaction(62, 64, Aij=986.5)
        self.set_interaction(62, 65, Aij=663.5)
        self.set_interaction(62, 69, Aij=986.5)
        self.set_interaction(63, 1, Aij=315.29998779)
        self.set_interaction(63, 2, Aij=1264.0)
        self.set_interaction(63, 3, Aij=62.31999969)
        self.set_interaction(63, 4, Aij=89.86000061)
        self.set_interaction(63, 5, Aij=-151.0)
        self.set_interaction(63, 6, Aij=339.79998779)
        self.set_interaction(63, 7, Aij=-186.30000305)
        self.set_interaction(63, 8, Aij=-11.0)
        self.set_interaction(63, 9, Aij=-297.79998779)
        self.set_interaction(63, 10, Aij=-165.5)
        self.set_interaction(63, 11, Aij=-256.29998779)
        self.set_interaction(63, 12, Aij=193.8999939)
        self.set_interaction(63, 13, Aij=-338.5)
        self.set_interaction(63, 17, Aij=493.79998779)
        self.set_interaction(63, 18, Aij=-313.5)
        self.set_interaction(63, 21, Aij=13.40999985)
        self.set_interaction(63, 22, Aij=-44.70000076)
        self.set_interaction(63, 23, Aij=39.63000107)
        self.set_interaction(63, 24, Aij=183.3999939)
        self.set_interaction(63, 25, Aij=-79.08000183)
        self.set_interaction(63, 30, Aij=-208.8999939)
        self.set_interaction(63, 32, Aij=228.3999939)
        self.set_interaction(63, 33, Aij=-95.0)
        self.set_interaction(63, 35, Aij=-463.6000061)
        self.set_interaction(63, 37, Aij=-11.15999985)
        self.set_interaction(63, 39, Aij=-228.0)
        self.set_interaction(63, 41, Aij=-337.0)
        self.set_interaction(63, 46, Aij=-322.29998779)
        self.set_interaction(63, 52, Aij=315.29998779)
        self.set_interaction(63, 53, Aij=-151.0)
        self.set_interaction(63, 55, Aij=-297.79998779)
        self.set_interaction(63, 56, Aij=315.29998779)
        self.set_interaction(63, 57, Aij=-338.5)
        self.set_interaction(63, 58, Aij=-151.0)
        self.set_interaction(63, 59, Aij=315.29998779)
        self.set_interaction(63, 61, Aij=-151.0)
        self.set_interaction(63, 62, Aij=315.29998779)
        self.set_interaction(63, 64, Aij=-151.0)
        self.set_interaction(63, 66, Aij=315.29998779)
        self.set_interaction(63, 67, Aij=315.29998779)
        self.set_interaction(63, 68, Aij=315.29998779)
        self.set_interaction(63, 69, Aij=-151.0)
        self.set_interaction(63, 71, Aij=-100.17)
        self.set_interaction(63, 72, Aij=-501.23)
        self.set_interaction(63, 73, Aij=-579.8)
        self.set_interaction(63, 74, Aij=79.052)
        self.set_interaction(63, 75, Aij=-340.95)
        self.set_interaction(64, 1, Aij=156.3999939)
        self.set_interaction(64, 2, Aij=457.0)
        self.set_interaction(64, 3, Aij=89.59999847)
        self.set_interaction(64, 4, Aij=25.81999969)
        self.set_interaction(64, 6, Aij=-137.1000061)
        self.set_interaction(64, 7, Aij=265.86999512)
        self.set_interaction(64, 8, Aij=-259.70001221)
        self.set_interaction(64, 9, Aij=84.0)
        self.set_interaction(64, 10, Aij=-203.6000061)
        self.set_interaction(64, 11, Aij=101.09999847)
        self.set_interaction(64, 12, Aij=267.79998779)
        self.set_interaction(64, 13, Aij=28.05999947)
        self.set_interaction(64, 14, Aij=8.6420002)
        self.set_interaction(64, 15, Aij=42.70000076)
        self.set_interaction(64, 16, Aij=-323.0)
        self.set_interaction(64, 17, Aij=-52.38999939)
        self.set_interaction(64, 18, Aij=170.0)
        self.set_interaction(64, 19, Aij=6.71199989)
        self.set_interaction(64, 20, Aij=224.38999939)
        self.set_interaction(64, 21, Aij=75.62000275)
        self.set_interaction(64, 22, Aij=65.27999878)
        self.set_interaction(64, 23, Aij=-98.12000275)
        self.set_interaction(64, 24, Aij=143.1000061)
        self.set_interaction(64, 25, Aij=123.5)
        self.set_interaction(64, 26, Aij=256.5)
        self.set_interaction(64, 27, Aij=157.1000061)
        self.set_interaction(64, 28, Aij=488.8999939)
        self.set_interaction(64, 29, Aij=147.5)
        self.set_interaction(64, 30, Aij=-120.5)
        self.set_interaction(64, 31, Aij=-318.8999939)
        self.set_interaction(64, 32, Aij=313.5)
        self.set_interaction(64, 33, Aij=202.1000061)
        self.set_interaction(64, 34, Aij=727.79998779)
        self.set_interaction(64, 35, Aij=-202.1000061)
        self.set_interaction(64, 36, Aij=74.26999664)
        self.set_interaction(64, 37, Aij=225.80000305)
        self.set_interaction(64, 38, Aij=44.77999878)
        self.set_interaction(64, 39, Aij=-143.8999939)
        self.set_interaction(64, 41, Aij=190.30000305)
        self.set_interaction(64, 42, Aij=-817.70001221)
        self.set_interaction(64, 44, Aij=-504.20001221)
        self.set_interaction(64, 45, Aij=72.95999908)
        self.set_interaction(64, 46, Aij=-382.70001221)
        self.set_interaction(64, 47, Aij=-248.30000305)
        self.set_interaction(64, 49, Aij=104.69999695)
        self.set_interaction(64, 50, Aij=57.65000153)
        self.set_interaction(64, 52, Aij=156.3999939)
        self.set_interaction(64, 54, Aij=199.0)
        self.set_interaction(64, 55, Aij=84.0)
        self.set_interaction(64, 56, Aij=156.3999939)
        self.set_interaction(64, 57, Aij=28.05999947)
        self.set_interaction(64, 59, Aij=156.3999939)
        self.set_interaction(64, 60, Aij=199.0)
        self.set_interaction(64, 62, Aij=156.3999939)
        self.set_interaction(64, 63, Aij=199.0)
        self.set_interaction(64, 65, Aij=224.39)
        self.set_interaction(64, 66, Aij=156.3999939)
        self.set_interaction(64, 67, Aij=156.3999939)
        self.set_interaction(64, 68, Aij=156.3999939)
        self.set_interaction(64, 70, Aij=-843.09002686)
        self.set_interaction(64, 71, Aij=37.631)
        self.set_interaction(64, 72, Aij=-330.28)
        self.set_interaction(64, 73, Aij=1789.0)
        self.set_interaction(64, 74, Aij=221.38)
        self.set_interaction(64, 75, Aij=-77.526)
        self.set_interaction(65, 1, Aij=315.29998779)
        self.set_interaction(65, 2, Aij=1264.0)
        self.set_interaction(65, 3, Aij=62.31999969)
        self.set_interaction(65, 4, Aij=89.86000061)
        self.set_interaction(65, 5, Aij=-103.02999878)
        self.set_interaction(65, 6, Aij=339.79998779)
        self.set_interaction(65, 7, Aij=-145.88000488)
        self.set_interaction(65, 8, Aij=-11.0)
        self.set_interaction(65, 9, Aij=-297.79998779)
        self.set_interaction(65, 10, Aij=-165.5)
        self.set_interaction(65, 11, Aij=-256.29998779)
        self.set_interaction(65, 12, Aij=193.8999939)
        self.set_interaction(65, 13, Aij=-338.5)
        self.set_interaction(65, 17, Aij=493.79998779)
        self.set_interaction(65, 18, Aij=-313.5)
        self.set_interaction(65, 21, Aij=13.40999985)
        self.set_interaction(65, 22, Aij=-44.70000076)
        self.set_interaction(65, 23, Aij=39.63000107)
        self.set_interaction(65, 24, Aij=183.3999939)
        self.set_interaction(65, 25, Aij=-79.08000183)
        self.set_interaction(65, 30, Aij=-208.8999939)
        self.set_interaction(65, 32, Aij=228.3999939)
        self.set_interaction(65, 33, Aij=-95.0)
        self.set_interaction(65, 35, Aij=-463.6000061)
        self.set_interaction(65, 37, Aij=-11.15999985)
        self.set_interaction(65, 39, Aij=-228.0)
        self.set_interaction(65, 41, Aij=-337.0)
        self.set_interaction(65, 46, Aij=-322.29998779)
        self.set_interaction(65, 52, Aij=315.29998779)
        self.set_interaction(65, 53, Aij=-151.0)
        self.set_interaction(65, 55, Aij=-297.79998779)
        self.set_interaction(65, 56, Aij=315.29998779)
        self.set_interaction(65, 57, Aij=-338.5)
        self.set_interaction(65, 58, Aij=-151.0)
        self.set_interaction(65, 59, Aij=315.29998779)
        self.set_interaction(65, 61, Aij=-151.0)
        self.set_interaction(65, 62, Aij=315.29998779)
        self.set_interaction(65, 64, Aij=-103.03)
        self.set_interaction(65, 66, Aij=315.29998779)
        self.set_interaction(65, 67, Aij=315.29998779)
        self.set_interaction(65, 68, Aij=315.29998779)
        self.set_interaction(65, 69, Aij=-103.02999878)
        self.set_interaction(65, 71, Aij=-100.17)
        self.set_interaction(65, 72, Aij=-501.23)
        self.set_interaction(65, 73, Aij=-579.8)
        self.set_interaction(65, 74, Aij=79.052)
        self.set_interaction(65, 75, Aij=-340.95)
        self.set_interaction(66, 2, Aij=86.01999664)
        self.set_interaction(66, 3, Aij=61.13000107)
        self.set_interaction(66, 4, Aij=76.5)
        self.set_interaction(66, 5, Aij=986.5)
        self.set_interaction(66, 6, Aij=697.20001221)
        self.set_interaction(66, 7, Aij=1890.0)
        self.set_interaction(66, 8, Aij=1333.0)
        self.set_interaction(66, 9, Aij=476.3999939)
        self.set_interaction(66, 10, Aij=677.0)
        self.set_interaction(66, 11, Aij=232.1000061)
        self.set_interaction(66, 12, Aij=507.0)
        self.set_interaction(66, 13, Aij=251.5)
        self.set_interaction(66, 14, Aij=391.5)
        self.set_interaction(66, 15, Aij=255.69999695)
        self.set_interaction(66, 16, Aij=206.6000061)
        self.set_interaction(66, 17, Aij=920.70001221)
        self.set_interaction(66, 18, Aij=287.79998779)
        self.set_interaction(66, 19, Aij=597.0)
        self.set_interaction(66, 20, Aij=663.5)
        self.set_interaction(66, 21, Aij=35.93000031)
        self.set_interaction(66, 22, Aij=53.75999832)
        self.set_interaction(66, 23, Aij=24.89999962)
        self.set_interaction(66, 24, Aij=104.30000305)
        self.set_interaction(66, 25, Aij=11.43999958)
        self.set_interaction(66, 26, Aij=661.5)
        self.set_interaction(66, 27, Aij=543.0)
        self.set_interaction(66, 28, Aij=153.6000061)
        self.set_interaction(66, 29, Aij=184.3999939)
        self.set_interaction(66, 30, Aij=354.6000061)
        self.set_interaction(66, 31, Aij=3025.0)
        self.set_interaction(66, 32, Aij=335.79998779)
        self.set_interaction(66, 33, Aij=479.5)
        self.set_interaction(66, 34, Aij=298.8999939)
        self.set_interaction(66, 35, Aij=526.5)
        self.set_interaction(66, 36, Aij=689.0)
        self.set_interaction(66, 37, Aij=-4.18900013)
        self.set_interaction(66, 38, Aij=125.80000305)
        self.set_interaction(66, 39, Aij=485.29998779)
        self.set_interaction(66, 40, Aij=-2.85899997)
        self.set_interaction(66, 41, Aij=387.1000061)
        self.set_interaction(66, 42, Aij=-450.3999939)
        self.set_interaction(66, 43, Aij=252.69999695)
        self.set_interaction(66, 44, Aij=220.30000305)
        self.set_interaction(66, 45, Aij=-5.86899996)
        self.set_interaction(66, 46, Aij=390.8999939)
        self.set_interaction(66, 47, Aij=553.29998779)
        self.set_interaction(66, 48, Aij=187.0)
        self.set_interaction(66, 49, Aij=216.1000061)
        self.set_interaction(66, 50, Aij=92.98999786)
        self.set_interaction(66, 53, Aij=986.5)
        self.set_interaction(66, 54, Aij=663.5)
        self.set_interaction(66, 55, Aij=476.3999939)
        self.set_interaction(66, 57, Aij=251.5)
        self.set_interaction(66, 58, Aij=986.5)
        self.set_interaction(66, 60, Aij=663.5)
        self.set_interaction(66, 61, Aij=986.5)
        self.set_interaction(66, 63, Aij=663.5)
        self.set_interaction(66, 64, Aij=986.5)
        self.set_interaction(66, 65, Aij=663.5)
        self.set_interaction(66, 69, Aij=986.5)
        self.set_interaction(66, 70, Aij=66.12200165)
        self.set_interaction(66, 71, Aij=500.95)
        self.set_interaction(66, 72, Aij=977.56)
        self.set_interaction(66, 73, Aij=1331.0)
        self.set_interaction(66, 74, Aij=297.24)
        self.set_interaction(66, 75, Aij=528.5)
        self.set_interaction(67, 2, Aij=86.01999664)
        self.set_interaction(67, 3, Aij=61.13000107)
        self.set_interaction(67, 4, Aij=76.5)
        self.set_interaction(67, 5, Aij=986.5)
        self.set_interaction(67, 6, Aij=697.20001221)
        self.set_interaction(67, 7, Aij=1325.0)
        self.set_interaction(67, 8, Aij=1333.0)
        self.set_interaction(67, 9, Aij=476.3999939)
        self.set_interaction(67, 10, Aij=677.0)
        self.set_interaction(67, 11, Aij=232.1000061)
        self.set_interaction(67, 12, Aij=507.0)
        self.set_interaction(67, 13, Aij=251.5)
        self.set_interaction(67, 14, Aij=391.5)
        self.set_interaction(67, 15, Aij=255.69999695)
        self.set_interaction(67, 16, Aij=206.6000061)
        self.set_interaction(67, 17, Aij=920.70001221)
        self.set_interaction(67, 18, Aij=287.79998779)
        self.set_interaction(67, 19, Aij=597.0)
        self.set_interaction(67, 20, Aij=663.5)
        self.set_interaction(67, 21, Aij=35.93000031)
        self.set_interaction(67, 22, Aij=53.75999832)
        self.set_interaction(67, 23, Aij=24.89999962)
        self.set_interaction(67, 24, Aij=104.30000305)
        self.set_interaction(67, 25, Aij=11.43999958)
        self.set_interaction(67, 26, Aij=661.5)
        self.set_interaction(67, 27, Aij=543.0)
        self.set_interaction(67, 28, Aij=153.6000061)
        self.set_interaction(67, 29, Aij=184.3999939)
        self.set_interaction(67, 30, Aij=354.6000061)
        self.set_interaction(67, 31, Aij=3025.0)
        self.set_interaction(67, 32, Aij=335.79998779)
        self.set_interaction(67, 33, Aij=479.5)
        self.set_interaction(67, 34, Aij=298.8999939)
        self.set_interaction(67, 35, Aij=526.5)
        self.set_interaction(67, 36, Aij=689.0)
        self.set_interaction(67, 37, Aij=-4.18900013)
        self.set_interaction(67, 38, Aij=125.80000305)
        self.set_interaction(67, 39, Aij=485.29998779)
        self.set_interaction(67, 40, Aij=-2.85899997)
        self.set_interaction(67, 41, Aij=387.1000061)
        self.set_interaction(67, 42, Aij=-450.3999939)
        self.set_interaction(67, 43, Aij=252.69999695)
        self.set_interaction(67, 44, Aij=220.30000305)
        self.set_interaction(67, 45, Aij=-5.86899996)
        self.set_interaction(67, 46, Aij=390.8999939)
        self.set_interaction(67, 47, Aij=553.29998779)
        self.set_interaction(67, 48, Aij=187.0)
        self.set_interaction(67, 49, Aij=216.1000061)
        self.set_interaction(67, 50, Aij=92.98999786)
        self.set_interaction(67, 53, Aij=986.5)
        self.set_interaction(67, 54, Aij=663.5)
        self.set_interaction(67, 55, Aij=476.3999939)
        self.set_interaction(67, 57, Aij=251.5)
        self.set_interaction(67, 58, Aij=986.5)
        self.set_interaction(67, 60, Aij=663.5)
        self.set_interaction(67, 61, Aij=986.5)
        self.set_interaction(67, 63, Aij=663.5)
        self.set_interaction(67, 64, Aij=986.5)
        self.set_interaction(67, 65, Aij=663.5)
        self.set_interaction(67, 69, Aij=986.5)
        self.set_interaction(67, 70, Aij=66.12200165)
        self.set_interaction(67, 71, Aij=500.95)
        self.set_interaction(67, 72, Aij=977.56)
        self.set_interaction(67, 73, Aij=1331.0)
        self.set_interaction(67, 74, Aij=297.24)
        self.set_interaction(67, 75, Aij=528.5)
        self.set_interaction(68, 2, Aij=86.01999664)
        self.set_interaction(68, 3, Aij=61.13000107)
        self.set_interaction(68, 4, Aij=76.5)
        self.set_interaction(68, 5, Aij=986.5)
        self.set_interaction(68, 6, Aij=697.20001221)
        self.set_interaction(68, 7, Aij=2314.0)
        self.set_interaction(68, 8, Aij=1333.0)
        self.set_interaction(68, 9, Aij=476.3999939)
        self.set_interaction(68, 10, Aij=677.0)
        self.set_interaction(68, 11, Aij=232.1000061)
        self.set_interaction(68, 12, Aij=507.0)
        self.set_interaction(68, 13, Aij=251.5)
        self.set_interaction(68, 14, Aij=391.5)
        self.set_interaction(68, 15, Aij=255.69999695)
        self.set_interaction(68, 16, Aij=206.6000061)
        self.set_interaction(68, 17, Aij=920.70001221)
        self.set_interaction(68, 18, Aij=287.79998779)
        self.set_interaction(68, 19, Aij=597.0)
        self.set_interaction(68, 20, Aij=663.5)
        self.set_interaction(68, 21, Aij=35.93000031)
        self.set_interaction(68, 22, Aij=53.75999832)
        self.set_interaction(68, 23, Aij=24.89999962)
        self.set_interaction(68, 24, Aij=104.30000305)
        self.set_interaction(68, 25, Aij=11.43999958)
        self.set_interaction(68, 26, Aij=661.5)
        self.set_interaction(68, 27, Aij=543.0)
        self.set_interaction(68, 28, Aij=153.6000061)
        self.set_interaction(68, 29, Aij=184.3999939)
        self.set_interaction(68, 30, Aij=354.6000061)
        self.set_interaction(68, 31, Aij=3025.0)
        self.set_interaction(68, 32, Aij=335.79998779)
        self.set_interaction(68, 33, Aij=479.5)
        self.set_interaction(68, 34, Aij=298.8999939)
        self.set_interaction(68, 35, Aij=526.5)
        self.set_interaction(68, 36, Aij=689.0)
        self.set_interaction(68, 37, Aij=-4.18900013)
        self.set_interaction(68, 38, Aij=125.80000305)
        self.set_interaction(68, 39, Aij=485.29998779)
        self.set_interaction(68, 40, Aij=-2.85899997)
        self.set_interaction(68, 41, Aij=387.1000061)
        self.set_interaction(68, 42, Aij=-450.3999939)
        self.set_interaction(68, 43, Aij=252.69999695)
        self.set_interaction(68, 44, Aij=220.30000305)
        self.set_interaction(68, 45, Aij=-5.86899996)
        self.set_interaction(68, 46, Aij=390.8999939)
        self.set_interaction(68, 47, Aij=553.29998779)
        self.set_interaction(68, 48, Aij=187.0)
        self.set_interaction(68, 49, Aij=216.1000061)
        self.set_interaction(68, 50, Aij=92.98999786)
        self.set_interaction(68, 53, Aij=986.5)
        self.set_interaction(68, 54, Aij=663.5)
        self.set_interaction(68, 55, Aij=476.3999939)
        self.set_interaction(68, 57, Aij=251.5)
        self.set_interaction(68, 58, Aij=986.5)
        self.set_interaction(68, 60, Aij=663.5)
        self.set_interaction(68, 61, Aij=986.5)
        self.set_interaction(68, 63, Aij=663.5)
        self.set_interaction(68, 64, Aij=986.5)
        self.set_interaction(68, 65, Aij=663.5)
        self.set_interaction(68, 69, Aij=986.5)
        self.set_interaction(68, 70, Aij=66.12200165)
        self.set_interaction(68, 71, Aij=500.95)
        self.set_interaction(68, 72, Aij=977.56)
        self.set_interaction(68, 73, Aij=1331.0)
        self.set_interaction(68, 74, Aij=297.24)
        self.set_interaction(68, 75, Aij=528.5)
        self.set_interaction(69, 1, Aij=156.3999939)
        self.set_interaction(69, 2, Aij=457.0)
        self.set_interaction(69, 3, Aij=89.59999847)
        self.set_interaction(69, 4, Aij=25.81999969)
        self.set_interaction(69, 6, Aij=-137.1000061)
        self.set_interaction(69, 7, Aij=276.3999939)
        self.set_interaction(69, 8, Aij=-259.70001221)
        self.set_interaction(69, 9, Aij=84.0)
        self.set_interaction(69, 10, Aij=-203.6000061)
        self.set_interaction(69, 11, Aij=101.09999847)
        self.set_interaction(69, 12, Aij=267.79998779)
        self.set_interaction(69, 13, Aij=28.05999947)
        self.set_interaction(69, 14, Aij=8.6420002)
        self.set_interaction(69, 15, Aij=42.70000076)
        self.set_interaction(69, 16, Aij=-323.0)
        self.set_interaction(69, 17, Aij=-52.38999939)
        self.set_interaction(69, 18, Aij=170.0)
        self.set_interaction(69, 19, Aij=6.71199989)
        self.set_interaction(69, 20, Aij=199.0)
        self.set_interaction(69, 21, Aij=75.62000275)
        self.set_interaction(69, 22, Aij=65.27999878)
        self.set_interaction(69, 23, Aij=-98.12000275)
        self.set_interaction(69, 24, Aij=143.1000061)
        self.set_interaction(69, 25, Aij=123.5)
        self.set_interaction(69, 26, Aij=256.5)
        self.set_interaction(69, 27, Aij=157.1000061)
        self.set_interaction(69, 28, Aij=488.8999939)
        self.set_interaction(69, 29, Aij=147.5)
        self.set_interaction(69, 30, Aij=-120.5)
        self.set_interaction(69, 31, Aij=-318.8999939)
        self.set_interaction(69, 32, Aij=313.5)
        self.set_interaction(69, 33, Aij=202.1000061)
        self.set_interaction(69, 34, Aij=727.79998779)
        self.set_interaction(69, 35, Aij=-202.1000061)
        self.set_interaction(69, 36, Aij=74.26999664)
        self.set_interaction(69, 37, Aij=225.80000305)
        self.set_interaction(69, 38, Aij=44.77999878)
        self.set_interaction(69, 39, Aij=-143.8999939)
        self.set_interaction(69, 41, Aij=190.30000305)
        self.set_interaction(69, 42, Aij=-817.70001221)
        self.set_interaction(69, 44, Aij=-504.20001221)
        self.set_interaction(69, 45, Aij=72.95999908)
        self.set_interaction(69, 46, Aij=-382.70001221)
        self.set_interaction(69, 47, Aij=-248.30000305)
        self.set_interaction(69, 49, Aij=104.69999695)
        self.set_interaction(69, 50, Aij=57.65000153)
        self.set_interaction(69, 52, Aij=156.3999939)
        self.set_interaction(69, 54, Aij=199.0)
        self.set_interaction(69, 55, Aij=84.0)
        self.set_interaction(69, 56, Aij=156.3999939)
        self.set_interaction(69, 57, Aij=28.05999947)
        self.set_interaction(69, 59, Aij=156.3999939)
        self.set_interaction(69, 60, Aij=199.0)
        self.set_interaction(69, 62, Aij=156.3999939)
        self.set_interaction(69, 63, Aij=199.0)
        self.set_interaction(69, 65, Aij=224.39)
        self.set_interaction(69, 66, Aij=156.3999939)
        self.set_interaction(69, 67, Aij=156.3999939)
        self.set_interaction(69, 68, Aij=156.3999939)
        self.set_interaction(69, 70, Aij=-843.09002686)
        self.set_interaction(69, 71, Aij=37.631)
        self.set_interaction(69, 72, Aij=-330.28)
        self.set_interaction(69, 73, Aij=1789.0)
        self.set_interaction(69, 74, Aij=221.38)
        self.set_interaction(69, 75, Aij=-77.526)
        self.set_interaction(70, 1, Aij=163.47999573)
        self.set_interaction(70, 5, Aij=74.44999695)
        self.set_interaction(70, 7, Aij=579.80999756)
        self.set_interaction(70, 64, Aij=74.44999695)
        self.set_interaction(70, 66, Aij=163.47999573)
        self.set_interaction(70, 67, Aij=163.47999573)
        self.set_interaction(70, 68, Aij=163.47999573)
        self.set_interaction(70, 69, Aij=74.44999695)
        self.set_interaction(71, 1, Aij=-75.71800232)
        self.set_interaction(71, 2, Aij=-294.42999268)
        self.set_interaction(71, 5, Aij=818.9699707)
        self.set_interaction(71, 7, Aij=681.7800293)
        self.set_interaction(71, 9, Aij=188.72000122)
        self.set_interaction(71, 10, Aij=-179.38000488)
        self.set_interaction(71, 11, Aij=-356.25)
        self.set_interaction(71, 13, Aij=-289.80999756)
        self.set_interaction(71, 20, Aij=1173.30004883)
        self.set_interaction(71, 53, Aij=818.9699707)
        self.set_interaction(71, 54, Aij=1173.30004883)
        self.set_interaction(71, 58, Aij=818.9699707)
        self.set_interaction(71, 60, Aij=1173.30004883)
        self.set_interaction(71, 61, Aij=818.9699707)
        self.set_interaction(71, 63, Aij=1173.30004883)
        self.set_interaction(71, 64, Aij=818.9699707)
        self.set_interaction(71, 65, Aij=1173.30004883)
        self.set_interaction(71, 66, Aij=-75.71800232)
        self.set_interaction(71, 67, Aij=-75.71800232)
        self.set_interaction(71, 68, Aij=-75.71800232)
        self.set_interaction(71, 69, Aij=818.9699707)
        self.set_interaction(71, 72, Aij=545.65997314)
        self.set_interaction(71, 73, Aij=551.95001221)
        self.set_interaction(71, 74, Aij=-308.16000366)
        self.set_interaction(71, 75, Aij=-239.6499939)
        self.set_interaction(72, 1, Aij=-23.2329998)
        self.set_interaction(72, 2, Aij=-57.94900131)
        self.set_interaction(72, 5, Aij=342.92001343)
        self.set_interaction(72, 7, Aij=795.54998779)
        self.set_interaction(72, 9, Aij=380.94000244)
        self.set_interaction(72, 10, Aij=408.88000488)
        self.set_interaction(72, 11, Aij=-355.0)
        self.set_interaction(72, 13, Aij=490.35998535)
        self.set_interaction(72, 20, Aij=1479.0)
        self.set_interaction(72, 52, Aij=-23.2329998)
        self.set_interaction(72, 53, Aij=342.92001343)
        self.set_interaction(72, 54, Aij=1479.0)
        self.set_interaction(72, 56, Aij=-23.2329998)
        self.set_interaction(72, 58, Aij=342.92001343)
        self.set_interaction(72, 60, Aij=1479.0)
        self.set_interaction(72, 61, Aij=342.92001343)
        self.set_interaction(72, 63, Aij=1479.0)
        self.set_interaction(72, 64, Aij=342.92001343)
        self.set_interaction(72, 65, Aij=1479.0)
        self.set_interaction(72, 66, Aij=-23.2329998)
        self.set_interaction(72, 67, Aij=-23.2329998)
        self.set_interaction(72, 68, Aij=-23.2329998)
        self.set_interaction(72, 69, Aij=342.92001343)
        self.set_interaction(72, 71, Aij=-86.29699707)
        self.set_interaction(72, 73, Aij=202.91000366)
        self.set_interaction(72, 74, Aij=-395.80999756)
        self.set_interaction(72, 75, Aij=-147.47000122)
        self.set_interaction(73, 1, Aij=5853.10009766)
        self.set_interaction(73, 2, Aij=883.7800293)
        self.set_interaction(73, 5, Aij=-457.92999268)
        self.set_interaction(73, 7, Aij=670.32000732)
        self.set_interaction(73, 9, Aij=-98.44999695)
        self.set_interaction(73, 10, Aij=-520.90002441)
        self.set_interaction(73, 11, Aij=131.1499939)
        self.set_interaction(73, 13, Aij=-471.67001343)
        self.set_interaction(73, 20, Aij=1896.09997559)
        self.set_interaction(73, 52, Aij=5853.10009766)
        self.set_interaction(73, 53, Aij=-457.92999268)
        self.set_interaction(73, 54, Aij=1896.09997559)
        self.set_interaction(73, 56, Aij=5853.10009766)
        self.set_interaction(73, 58, Aij=-457.92999268)
        self.set_interaction(73, 60, Aij=1896.09997559)
        self.set_interaction(73, 61, Aij=-457.92999268)
        self.set_interaction(73, 63, Aij=1896.09997559)
        self.set_interaction(73, 64, Aij=-457.92999268)
        self.set_interaction(73, 65, Aij=1896.09997559)
        self.set_interaction(73, 66, Aij=5853.10009766)
        self.set_interaction(73, 67, Aij=5853.10009766)
        self.set_interaction(73, 68, Aij=5853.10009766)
        self.set_interaction(73, 69, Aij=-457.92999268)
        self.set_interaction(73, 71, Aij=221.82000732)
        self.set_interaction(73, 72, Aij=-62.01670074)
        self.set_interaction(73, 74, Aij=210.57000732)
        self.set_interaction(73, 75, Aij=395.32998657)
        self.set_interaction(74, 1, Aij=-151.61000061)
        self.set_interaction(74, 2, Aij=-237.61000061)
        self.set_interaction(74, 5, Aij=820.85998535)
        self.set_interaction(74, 7, Aij=483.55300903)
        self.set_interaction(74, 9, Aij=587.21002197)
        self.set_interaction(74, 10, Aij=509.17001343)
        self.set_interaction(74, 11, Aij=449.04000854)
        self.set_interaction(74, 13, Aij=142.6499939)
        self.set_interaction(74, 20, Aij=1043.90002441)
        self.set_interaction(74, 52, Aij=-151.61000061)
        self.set_interaction(74, 53, Aij=820.85998535)
        self.set_interaction(74, 54, Aij=1043.90002441)
        self.set_interaction(74, 55, Aij=587.21002197)
        self.set_interaction(74, 56, Aij=-151.61000061)
        self.set_interaction(74, 58, Aij=820.85998535)
        self.set_interaction(74, 60, Aij=1043.90002441)
        self.set_interaction(74, 61, Aij=820.85998535)
        self.set_interaction(74, 63, Aij=1043.90002441)
        self.set_interaction(74, 64, Aij=820.85998535)
        self.set_interaction(74, 65, Aij=1043.90002441)
        self.set_interaction(74, 66, Aij=-151.61000061)
        self.set_interaction(74, 67, Aij=-151.61000061)
        self.set_interaction(74, 68, Aij=-151.61000061)
        self.set_interaction(74, 69, Aij=820.85998535)
        self.set_interaction(74, 71, Aij=676.61999512)
        self.set_interaction(74, 72, Aij=1088.80004883)
        self.set_interaction(74, 73, Aij=537.70001221)
        self.set_interaction(74, 75, Aij=-2.07949996)
        self.set_interaction(75, 1, Aij=333.07000732)
        self.set_interaction(75, 2, Aij=86.30699921)
        self.set_interaction(75, 5, Aij=612.04998779)
        self.set_interaction(75, 7, Aij=319.98999023)
        self.set_interaction(75, 9, Aij=111.76000214)
        self.set_interaction(75, 10, Aij=-187.02000427)
        self.set_interaction(75, 11, Aij=-157.63999939)
        self.set_interaction(75, 13, Aij=-208.91000366)
        self.set_interaction(75, 20, Aij=1207.69995117)
        self.set_interaction(75, 52, Aij=333.07000732)
        self.set_interaction(75, 53, Aij=612.04998779)
        self.set_interaction(75, 54, Aij=1207.69995117)
        self.set_interaction(75, 55, Aij=111.76000214)
        self.set_interaction(75, 56, Aij=333.07000732)
        self.set_interaction(75, 58, Aij=612.04998779)
        self.set_interaction(75, 60, Aij=1207.69995117)
        self.set_interaction(75, 61, Aij=612.04998779)
        self.set_interaction(75, 63, Aij=1207.69995117)
        self.set_interaction(75, 64, Aij=612.04998779)
        self.set_interaction(75, 65, Aij=1207.69995117)
        self.set_interaction(75, 66, Aij=333.07000732)
        self.set_interaction(75, 67, Aij=333.07000732)
        self.set_interaction(75, 68, Aij=333.07000732)
        self.set_interaction(75, 69, Aij=612.04998779)
        self.set_interaction(75, 71, Aij=474.47000122)
        self.set_interaction(75, 72, Aij=392.54000854)
        self.set_interaction(75, 73, Aij=-80.54299927)
        self.set_interaction(75, 74, Aij=339.79998779)
